/**
 * TrueReality - N64/controller.c
 * Copyright (C) 1998, 1999 Niki W. Waibel
 *
 * This program is free software; you can redistribute it and/
 * or modify it under the terms of the GNU General Public Li-
 * cence as published by the Free Software Foundation; either
 * version 2 of the Licence, or any later version.
 *
 * This program is distributed in the hope that it will be use-
 * ful, but WITHOUT ANY WARRANTY; without even the implied war-
 * ranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public Licence for more details.
 *
 * You should have received a copy of the GNU General Public
 * Licence along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 * USA.
 *
 * Information about me (the author):
 *   Niki W. Waibel, Reichenau 20, 6890 Lustenau, Austria - EUROPE
 *   niki.waibel@gmx.net
**/

/**
 *
 * Most of the knowledge in this file is based on a document i got from
 * LaC (the document has been been written by him).
 * LaC - thank you very much for that!
 *
**/



#include <stdio.h>

#include <type_sizes.h>
#include <memory_extern.h>
#include <dispatch.h>
#include <parser_extern.h>

#include "../input/keyboard/keyboard.h"
#include "controller.h"





#define GET_STATUS      0x00
#define READ_CONTROLLER 0x01
#define READ_MEMPACK    0x02
#define WRITE_MEMPACK   0x03
#define READ_EEPROM     0x04
#define WRITE_EEPROM    0x05
#define RESET           0xff
#define STATUS_EEPROM   0x100   /* hack - normally this is an impossible value */

#define ABS             0x01000000
#define REL             0x02000000
#define JOY             0x04000000
#define EEPROM          0x00800000
#define MEMPACK         0x00000100
#define NOMEMPACK       0x00000200

#define EEPROM_SIZE     1024





static uBYTE eeprom[EEPROM_SIZE];










void controller_check()
{
    /* major thanks go to LaC */

        int     i;      /* offset to origin of pif ram */
        int     code;
        int     c1;     /* command (main) */
        int     c2;     /* command */
        int     re;     /* bytes to read */
        int     wr;     /* bytes to write */
        int     curr;   /* current controller */
        uWORD    *pif = (uWORD *)mem.pi_ram;



        i = 0;
        curr = 1;

        do
        {
                code = pif[i];
#ifdef DEBUG
                if(prefs.debug & DBG_CONTR_PIF)
                        printf("PIF: code: $%08x [%d] -", code, i>>1);
#endif

                if(code == 0)
                {
                    /* let's read or write a eeprom ;-) */
                        i++;   /* we have to jump to the next word to get the info */
                        code = pif[i];
#ifdef DEBUG
                        if(prefs.debug & DBG_CONTR_PIF)
                                printf(" $%08x -", code);
#endif

                        wr = (code >> 24) & 0xff;   /* length+2 in bytes of eeprom read/write */
                        re = (code >> 16) & 0xff;   /* write: always 1? read: read re bytes   */
                        c2 = (code >>  8) & 0xff;   /* write: 5; read: 4                      */
                        c1 = (code      ) & 0xff;   /* offset in DWORDs                       */
                        if(wr == 0xff)   /* hack */
                        {
                            /* this way we can get the status of the eeprom ;-) */
                                c2 |= STATUS_EEPROM;
                        }
                        c1 <<= 3;   /* c1 = c1 * 8; so we get the offset in bytes */
                }
                else
                {
                    /* let's read or write the controller ;-) */
                        c1 = (code >> 24) & 0xff;
                        wr = (code >> 16) & 0xff;
                        re = (code >>  8) & 0xff;
                        c2 = (code      ) & 0xff;
                }

                if(c1 == 0xfe)   /* indicates the end */
                {
#ifdef DEBUG
                        if(prefs.debug & DBG_CONTR_PIF)
                                putchar('\n');
#endif
                        break;   /* exit this do{}while; */
                }

                switch(c2)
                {
                    case GET_STATUS:
#ifdef DEBUG
                        if(prefs.debug & DBG_CONTR_PIF)
                                fputs(" GET_STATUS", stdout);
#endif
                        if(curr & prefs.plugged_controller)   /* connected */
                        {
                                if(wr == 1 && re <= 4)
                                {
                                        i++;
                                        pif[i] &= (re==4) ? 0x00000000 :
                                                  (re==3) ? 0x000000ff :
                                                  (re==2) ? 0x0000ffff :
                                                  (re==1) ? 0x00ffffff : 0xffffffff;
                                        pif[i] |= ABS|JOY|NOMEMPACK;
#ifdef DEBUG
                                        if(prefs.debug & DBG_CONTR_PIF)
                                                printf(" wrote $%08lx\n", pif[i]);
#endif
                                }
                                else   /* error in transfer */
                                {
                                        pif[i] |= 0x4000;
#ifdef DEBUG
                                        if(prefs.debug & DBG_CONTR_PIF)
                                                printf(" (transfer error) wrote $%08lx\n", pif[i]);
#endif
                                        i++;
                                }
                        }
                        else   /* not connected */
                        {
                                pif[i] |= 0x8000;
#ifdef DEBUG
                                if(prefs.debug & DBG_CONTR_PIF)
                                        printf(" (not connected) wrote $%08lx\n", pif[i]);
#endif
                                i++;
                        }
                        break;

                    case READ_CONTROLLER:
#ifdef DEBUG
                        if(prefs.debug & DBG_CONTR_PIF)
                                fputs(" READ_CONTROLLER", stdout);
#endif
                        if(curr & prefs.plugged_controller)   /* connected */
                        {
                                if(wr == 1 && re <= 4)
                                {
                                        i++;
                                        pif[i] &= (re==4) ? 0x00000000 :
                                                  (re==3) ? 0x000000ff :
                                                  (re==2) ? 0x0000ffff :
                                                  (re==1) ? 0x00ffffff : 0xffffffff;
                                        pif[i] |= GetControllerData(i>>1);
#ifdef DEBUG
                                        if(prefs.debug & DBG_CONTR_PIF)
                                                printf(" wrote $%08lx\n", pif[i]);
#endif
                                }
                                else   /* error in transfer */
                                {
                                        pif[i] |= 0x4000;
#ifdef DEBUG
                                        if(prefs.debug & DBG_CONTR_PIF)
                                                printf(" (transfer error) wrote $%08lx\n", pif[i]);
#endif
                                        i++;
                                }
                        }
                        else   /* not connected */
                        { 
                                pif[i] |= 0x8000;
#ifdef DEBUG
                                if(prefs.debug & DBG_CONTR_PIF)
                                        printf(" (not connected) wrote $%08lx\n", pif[i]);
#endif
                                i++;
                        }
                        break;

                    case READ_MEMPACK:
                        pif[i] |= 0x8000;
                        printf(" WARNING - READ_MEMPACK: wrote $%08lx\n", pif[i]);
                        i++;
                        break;

                    case WRITE_MEMPACK:
                        pif[i] |= 0x8000;
                        printf(" WARNING - WRITE_MEMPACK: wrote $%08lx\n", pif[i]);
                        break;

                    case STATUS_EEPROM | 0x00:
                    case STATUS_EEPROM | 0x01:
                    case STATUS_EEPROM | 0x02:
                    case STATUS_EEPROM | 0x03:
                    case STATUS_EEPROM | 0x04:
#ifdef DEBUG
                        if(prefs.debug & DBG_CONTR_PIF)
                                fputs(" STATUS_EEPROM", stdout);
#endif
                        if(prefs.eeprom)   /* eeprom */
                        {
                                wr = c2 & 0xff;   /* hack */
                                if(re == 1 && wr <= 4)
                                {
                                        i++;
                                        pif[i] &= (wr==4) ? 0x00000000 :
                                                  (wr==3) ? 0x000000ff :
                                                  (wr==2) ? 0x0000ffff :
                                                  (wr==1) ? 0x00ffffff : 0xffffffff;
                                        pif[i] |= EEPROM;
#ifdef DEBUG
                                        if(prefs.debug & DBG_CONTR_PIF)
                                                printf(" wrote $%08lx\n", pif[i]);
#endif
                                }
                                else   /* error in transfer */
                                {
                                        pif[i] |= 0x4000;
#ifdef DEBUG
                                        if(prefs.debug & DBG_CONTR_PIF)
                                                printf(" (transfer error) wrote $%08lx\n", pif[i]);
#endif
                                        i++;
                                }
                        }
                        else   /* no eeprom */
                        {
                                pif[i] |= 0x8000;
#ifdef DEBUG
                                if(prefs.debug & DBG_CONTR_PIF)
                                        printf(" (no eeprom) wrote $%08lx\n", pif[i]);
#endif
                                i++;
                        }
                        break;

                    case READ_EEPROM:
#ifdef DEBUG
                        if(prefs.debug & DBG_CONTR_PIF)
                                fputs(" READ_EEPROM", stdout);
#endif
                        if(prefs.eeprom)   /* eeprom */
                        {
                                if(wr == 2)
                                {
                                        int j;

#ifdef DEBUG
                                        if(prefs.debug & DBG_CONTR_PIF)
                                                fputs(" wrote $", stdout);
#endif
                                        i++;
                                        for(j=0; j<re; j++)
                                        {
                                                ((uBYTE *)pif)[(i<<2)+j] = eeprom[j+c1];
#ifdef DEBUG
                                                if(prefs.debug & DBG_CONTR_PIF)
                                                        printf("%02x", ((unsigned char *)eeprom)[j+c1]);
#endif
                                        }
                                        i = 16;   /* just one eeprom command can appear */
#ifdef DEBUG
                                        if(prefs.debug & DBG_CONTR_PIF)
                                                putchar('\n');
#endif
                                }
                                else   /* error in transfer */
                                {
                                        pif[i] |= 0x4000;
#ifdef DEBUG
                                        if(prefs.debug & DBG_CONTR_PIF)
                                                printf(" (transfer error) wrote $%08lx\n", pif[i]);
#endif
                                        i = 16;
                                }
                        }
                        else   /* no eeprom */
                        {
                                pif[i] |= 0x8000;
#ifdef DEBUG
                                if(prefs.debug & DBG_CONTR_PIF)
                                        printf(" (no eeprom) wrote $%08lx\n", pif[i]);
#endif
                                i = 16;
                        }
                        break;

                    case WRITE_EEPROM:
#ifdef DEBUG
                        if(prefs.debug & DBG_CONTR_PIF)
                                fputs(" WRITE_EEPROM", stdout);
#endif
                        if(prefs.eeprom)   /* eeprom */
                        {
                                if(re == 1 && wr >= 2)
                                {
                                        int j;

#ifdef DEBUG
                                        if(prefs.debug & DBG_CONTR_PIF)
                                                fputs(" wrote $", stdout);
#endif
                                        i++;
                                        wr -= 2;
                                        for(j=0; j<wr; j++)
                                        {
                                                eeprom[j+c1] = ((uBYTE *)pif)[(i<<2)+j];
#ifdef DEBUG
                                                if(prefs.debug & DBG_CONTR_PIF)
                                                        printf("%02x", ((unsigned char *)eeprom)[j+c1]);
#endif
                                        }
                                        i = 16;   /* just one eeprom command can appear */
#ifdef DEBUG
                                        if(prefs.debug & DBG_CONTR_PIF)
                                                putchar('\n');
#endif
                                }
                                else   /* error in transfer */
                                {
                                        pif[i] |= 0x4000;
#ifdef DEBUG
                                        if(prefs.debug & DBG_CONTR_PIF)
                                                printf(" (transfer error) wrote $%08lx\n", pif[i]);
#endif
                                        i = 16;
                                }
                        }
                        else   /* no eeprom */
                        {
                                pif[i] |= 0x8000;
#ifdef DEBUG
                                if(prefs.debug & DBG_CONTR_PIF)
                                        printf(" (no eeprom) wrote $%08lx\n", pif[i]);
#endif
                                i = 16;
                        }
                        break;

                    case RESET:
#ifdef DEBUG
                        if(prefs.debug & DBG_CONTR_PIF)
                                puts(" WARNING - RESET: ignored");
#endif
                        i++;
                        break;

                    default:
                        pif[i] |= 0x8000;
                        printf(" WARNING - unknown command: wrote $%08lx\n", pif[i]);
                        i=16;   /* exit */

                } /* switch(c2) */

                i++;
                curr <<= 1;   /* switch to next controller */

        } /* do */
        while(i<=16);   /* we have to pay attension that we don't read beond pif ram */

        fflush(stdout);


        ((uBYTE *)pif)[63] = 0;   /* command processed */

} /* void controller_check() */





void InitController(int num)
{
        int i;



        prefs.plugged_controller = num;
        for(i=0; i<EEPROM_SIZE; i++)
                eeprom[i] = 0;

} /* void InitController(int num) */






