/**
 * TrueReality - N64/dma.c
 * Copyright (C) 1998, 1999 Niki W. Waibel
 *
 * This program is free software; you can redistribute it and/
 * or modify it under the terms of the GNU General Public Li-
 * cence as published by the Free Software Foundation; either
 * version 2 of the Licence, or any later version.
 *
 * This program is distributed in the hope that it will be use-
 * ful, but WITHOUT ANY WARRANTY; without even the implied war-
 * ranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public Licence for more details.
 *
 * You should have received a copy of the GNU General Public
 * Licence along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 * USA.
 *
 * Information about me (the author):
 *   Niki W. Waibel, Reichenau 20, 6890 Lustenau, Austria - EUROPE
 *   niki.waibel@gmx.net
**/





/**
 * Version information:
 *
 * 1999032000:
 *      Fixed endian stuff.
 *      Fixed bug in dma_idmem2dram() and dma_dram2idmem():
 *      One byte was missing!
 *
**/




#include <stdio.h>
#include <string.h>

#include <type_sizes.h>
#include <file_loader_extern.h>
#include <memory_extern.h>

#include "interrupts.h"
#include "controller.h"
#include "registers.h"
#include "dma.h"
#include "rdp.h"
#include "rsp_registers.h"





/* the following routines should use real dma! */
/* ... but that is impossible under unix ... maybe we could use a thread? */
/* currently this works :-) */





/*
   write happened to:
   SP_WR_LEN_REG   (0x404000c)
*/

void dma_idmem2dram()
{
        #define DMEM 0x0000
        #define IMEM 0x1000

        int     i;

        uWORD    idmem_offset = ((uWORD *)mem.sp_reg)[0] & 0xfff;
        int     imem_or_dmem = (((uWORD *)mem.sp_reg)[0] & 0x1000);
        
        uWORD    dram_offset  = ((uWORD *)mem.sp_reg)[1] & 0xffffff;

        uWORD    dma_length   = (((uWORD *)mem.sp_reg)[3] & 0x00000fff);
#warning The following warnings are because 'count' and 'skip' are not implemented right now (Niki)
        uWORD    dma_count    = (((uWORD *)mem.sp_reg)[3] & 0x000ff000) >> 12;   /* BUG: not impl */
        uWORD    dma_skip     = (((uWORD *)mem.sp_reg)[3] & 0xfff00000) >> 20;   /* BUG: not impl */



//        puts("idmem -> dram");
//        printf("%08lx -> %08lx\n", ((uWORD *)mem.sp_reg)[0], ((uWORD *)mem.sp_reg)[1]);

        for(i=0; i<=dma_length; i++)
        {
                if(imem_or_dmem & IMEM)
                {
#ifdef ENDIAN_IS_LITTLE
                        mem.rd_ram[(i+dram_offset)^3] = mem.sp_imem[(i+idmem_offset)^3];
#endif
#ifdef ENDIAN_IS_BIG
                        mem.rd_ram[i+dram_offset] = mem.sp_imem[i+idmem_offset];
#endif
                }
                else
                {
#ifdef ENDIAN_IS_LITTLE
                        mem.rd_ram[(i+dram_offset)^3] = mem.sp_dmem[(i+idmem_offset)^3];
#endif
#ifdef ENDIAN_IS_BIG
                        mem.rd_ram[i+dram_offset] = mem.sp_dmem[i+idmem_offset];
#endif
                }

        }

        reg.do_or_check_sthg |= RS4300I_DO_SP_INTERRUPT;



        #undef IMEM
        #undef DMEM
        
} /* void dma_idmem2dram() */





/*
   write happened to:
   SP_RD_LEN_REG   (0x4040008)
*/

void dma_dram2idmem()
{
        #define DMEM 0x0000
        #define IMEM 0x1000

        int     i;
        
        uWORD    idmem_offset = ((uWORD *)mem.sp_reg)[0] & 0xfff;
        int     imem_or_dmem = (((uWORD *)mem.sp_reg)[0] & 0x1000);
        
        uWORD    dram_offset  = ((uWORD *)mem.sp_reg)[1] & 0xffffff;

        uWORD    dma_length   = (((uWORD *)mem.sp_reg)[2] & 0x00000fff);
        uWORD    dma_count    = (((uWORD *)mem.sp_reg)[2] & 0x000ff000) >> 12;   /* BUG: not impl */
        uWORD    dma_skip     = (((uWORD *)mem.sp_reg)[2] & 0xfff00000) >> 20;   /* BUG: not impl */



        //puts("dram -> idmem");
        //printf("%08lx -> %08lx\n", ((uWORD *)mem.sp_reg)[1], ((uWORD *)mem.sp_reg)[0]);
        
        for(i=0; i<=dma_length; i++)
        {
                if(imem_or_dmem & IMEM)
                {
#ifdef ENDIAN_IS_LITTLE
                        mem.sp_imem[(i+idmem_offset)^3] = mem.rd_ram[(i+dram_offset)^3];
#endif
#ifdef ENDIAN_IS_BIG
                        mem.sp_imem[i+idmem_offset] = mem.rd_ram[i+dram_offset];
#endif
                        //putchar('.');
                }
                else
                {
#ifdef ENDIAN_IS_LITTLE
                        mem.sp_dmem[(i+idmem_offset)^3] = mem.rd_ram[(i+dram_offset)^3];
#endif
#ifdef ENDIAN_IS_BIG
                        mem.sp_dmem[i+idmem_offset] = mem.rd_ram[i+dram_offset];
#endif
                        //putchar('.');
                }
        }

        //putchar('\n');

        reg.do_or_check_sthg |= RS4300I_DO_SP_INTERRUPT;


#ifdef HLE
    /**
    * The following checks if we have a graphics or audio task loaded.
    * This just should happen if the RSP is stopped.
    *
    * We also check if we have written to IMEM last time.
    * This might be a BUG!
    **/
        if(rsp_reg.halt == 1)
                if(imem_or_dmem & IMEM)
                        rdp_dl_check();
#endif

        //reg.halt = 1;

        #undef IMEM
        #undef DMEM
        
} /* void dma_dram2idmem() */






/*
   write happened to:
   SI_PIF_ADDR_RD64B_REG   (0x4600004)
*/

void dma_pif2dram()
{
        int   i;
        uWORD  offset = ((uWORD *)mem.si_reg)[0];



//        puts("pif -> dram");
        
        controller_check();
        
        for(i=0; i<64; i++)
        {
                mem.rd_ram[i+offset] = mem.pi_ram[i];
        }

        reg.do_or_check_sthg |= RS4300I_DO_SI_INTERRUPT;


} /* void dma_pif2dram() */





/*
   write happened to:
   SI_PIF_ADDR_WR64B_REG   (0x4600010)
*/

void dma_dram2pif()
{
        int   i;
        uWORD  offset = ((uWORD *)mem.si_reg)[0];



//        puts("idram -> pif");
        
        for(i=0; i<64; i++)
        {
                mem.pi_ram[i] = mem.pi_ram_w[i] = mem.rd_ram[i+offset];
        }

        //controller_check();

        reg.do_or_check_sthg |= RS4300I_DO_SI_INTERRUPT;

} /* void dma_dram2pif() */





void dma_pi_copy()
{
        int i;
        
        uWORD source = ((uWORD *)mem.pi_reg)[1];
        uWORD dest   = ((uWORD *)mem.pi_reg)[0];
        uWORD len    = ((uWORD *)mem.pi_reg)[3];



//        puts("piram -> dram");
        
        /* BUG:? horrible hack */
        source &= 0x03ffffff;   /* 64MB ... assumed as max rom length */

        if((source+len) > rom.length)
                len = rom.length - source;
        
        
        
        //fputs("dma_pi_copy:\n", stderr);

        for(i=0; i<=len; i++)
        {
#ifdef ENDIAN_IS_LITTLE
                mem.rd_ram[(i+dest)^3] = rom.image[(i+source)^3];
#endif
#ifdef ENDIAN_IS_BIG
                mem.rd_ram[i+dest] = rom.image[i+source];
#endif
                //fprintf(stderr, "%02x\n", mem.rd_ram[i+dest]);
        }

        reg.do_or_check_sthg |= RS4300I_DO_PI_INTERRUPT;
        
} /* void dma_pi_copy() */










