/**
 * TrueReality - N64/interrupts.c
 * Copyright (C) 1998, 1999 Niki W. Waibel
 *
 * This program is free software; you can redistribute it and/
 * or modify it under the terms of the GNU General Public Li-
 * cence as published by the Free Software Foundation; either
 * version 2 of the Licence, or any later version.
 *
 * This program is distributed in the hope that it will be use-
 * ful, but WITHOUT ANY WARRANTY; without even the implied war-
 * ranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public Licence for more details.
 *
 * You should have received a copy of the GNU General Public
 * Licence along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 * USA.
 *
 * Information about me (the author):
 *   Niki W. Waibel, Reichenau 20, 6890 Lustenau, Austria - EUROPE
 *   niki.waibel@gmx.net
**/










#include <stdio.h>
#include <stdlib.h>

#include <config.h>
#include <type_sizes.h>
#include <memory_extern.h>
#include <parser_extern.h>

#include "cpu.h"
#ifndef PUR_VI_INTERRUPT
        #include "dispatch.h"
#endif
#include "../input/keyboard/keyboard.h"
#include "registers.h"
#include "exceptions.h"
#include "interrupts.h"

#ifdef DEBUG
        #include <debug.h>
#endif





#define VERTICAL_RETRACE_TIME 0x98968   /* I hope that this value ok - pur uses it (625000). */
    /* Maybe it should be 0x2faf08 (instr in 1/30sec by 93.75MHz) */
/* #define VERTICAL_RETRACE_TIME 0x2faf08 */
/* #define VERTICAL_RETRACE_TIME 0x17d784 */

#ifdef DEBUG
        #define PRINT_INTR(flag, interrupt)                                                     \
                if(prefs.debug & (flag))                                                        \
                {                                                                               \
                        printf("INTERRUPT (%s) @ %8lx started\n", interrupt, (uWORD)reg.cpr[0][COUNT]);  \
                        fflush(stdout);                                                         \
                }

        #define PRINT_INTR_INFO(flag, interrupt)                                                \
                if(prefs.debug & (flag))                                                        \
                {                                                                               \
                        printf("INTERRUPT (%s) @ %8lx called\n", interrupt, (uWORD)reg.cpr[0][COUNT]);  \
                        fflush(stdout);                                                         \
                }
#else
        #define PRINT_INTR(flag, interrupt)
        #define PRINT_INTR_INFO(flag, interrupt)
#endif
                        





uWORD next_interrupt_count;





static uWORD next_vi_interrupt;

static void rs4300i_MI_Intr_VI();
static void rs4300i_CompareInterrupt();






void rs4300i_InitInterrupts()
{
        next_interrupt_count    = 0;   /* to let the cpu know */
        next_vi_interrupt       = VERTICAL_RETRACE_TIME;
        /* next_cp_interrupt       = reg.cpr[0][COMPARE]; */   /* We use the reg.cpr[0] reg. */

} /* void rs4300i_InitInterrupts() */









void rs4300i_VI_or_CP_calc()
{
        uWORD vi_diff;
        uWORD cp_diff;

        vi_diff = next_vi_interrupt - ((uWORD)reg.cpr[0][COUNT]);
        cp_diff = ((uWORD)reg.cpr[0][COMPARE]) - ((uWORD)reg.cpr[0][COUNT]);
//printf("dvi: %lx dcp: %lx @ %lx\n", vi_diff, cp_diff, ((uWORD)reg.cpr[0][COUNT]));
//printf("vi: %lx cp: %lx @ %lx\n", next_vi_interrupt, ((uWORD)reg.cpr[0][COMPARE]), ((uWORD)reg.cpr[0][COUNT]));
//fflush(stdout);
        if(cp_diff < vi_diff)
        {
                next_interrupt_count = ((uWORD)reg.cpr[0][COMPARE]);
        }
        else
        {
                next_interrupt_count = next_vi_interrupt;
        }

} /* void rs4300i_VI_or_CP_calc() */





void rs4300i_VI_or_CP_interrupt()
{
//printf("nic: %lx nvi: %lx nci: %lx\n", next_interrupt_count, next_vi_interrupt, ((uWORD)reg.cpr[0][COMPARE]));
//fflush(stdout);
        if(next_interrupt_count == next_vi_interrupt)
        {
                next_vi_interrupt += VERTICAL_RETRACE_TIME;
                rs4300i_MI_Intr_VI();
        }

        if(next_interrupt_count == ((uWORD)reg.cpr[0][COMPARE]))
        {
                rs4300i_CompareInterrupt();
        }

        rs4300i_VI_or_CP_calc();

} /* void rs4300i_VI_or_CP_interrupt() */





void rs4300i_MI_Intr_SP()
{
        PRINT_INTR_INFO(DBG_INTR_SP, "SP")

        /* add MI_INTR_SP (0x01) in MI INTR reg */
        ((uWORD *)mem.mi_reg)[2] |= MI_INTR_SP;


   
        /* if MI_INTR_SP (0x01) in MI INTR MASK reg is disabled: return */
        if(!(((uWORD *)mem.mi_reg)[3] & MI_INTR_SP))
                return;



        /* set IP2 in CAUSE reg */
        reg.cpr[0][CAUSE] |= 0x0400;
   

   
        /* if IP2 or IE bits in STATUS reg are clr: return */
        if((reg.cpr[0][STATUS] & 0x0401) != 0x0401)
                return;

      
   
        /* if EXL or ERL bits in STATUS reg are set: return */
        if(reg.cpr[0][STATUS] & 0x0006)
                return;



        PRINT_INTR(DBG_INTR_SP, "SP")


   
        /* set ExcCode to Int (0) */
        reg.cpr[0][CAUSE] &= ~EXC_CODE__MASK;
        


        jump_to_common_exception_vector();

} /* void rs4300i_MI_Intr_SP() */





void rs4300i_MI_Intr_SI()
{
        PRINT_INTR_INFO(DBG_INTR_SI, "SI")

        /* add bit # 12 in SI_STATUS_REG */
        ((uWORD *)mem.si_reg)[6] = 0x00001000;


   
        /* add MI_INTR_SI (0x02) in MI INTR reg */
        ((uWORD *)mem.mi_reg)[2] |= MI_INTR_SI;


   
        /* if MI_INTR_SI (0x02) in MI INTR MASK reg is disabled: return */
        if(!(((uWORD *)mem.mi_reg)[3] & MI_INTR_SI))
                return;



        /* set IP2 in CAUSE reg */
        reg.cpr[0][CAUSE] |= 0x0400;
   

   
        /* if IP2 or IE bits in STATUS reg are clr: return */
        if((reg.cpr[0][STATUS] & 0x0401) != 0x0401)
                return;

      
   
        /* if EXL or ERL bits in STATUS reg are set: return */
        if(reg.cpr[0][STATUS] & 0x0006)
                return;



        PRINT_INTR(DBG_INTR_SI, "SI")


   
        /* set ExcCode to Int (0) */
        reg.cpr[0][CAUSE] &= ~EXC_CODE__MASK;
        


        jump_to_common_exception_vector();

} /* void rs4300i_MI_Intr_SI() */





void rs4300i_MI_Intr_AI()
{
        PRINT_INTR_INFO(DBG_INTR_AI, "AI")

        /* add MI_INTR_AI (0x04) in MI INTR reg */
        ((uWORD *)mem.mi_reg)[2] |= MI_INTR_AI;


   
        /* if MI_INTR_AI (0x00) in MI INTR MASK reg is disabled: return */
        if(!(((uWORD *)mem.mi_reg)[3] & MI_INTR_AI))
                return;



        /* set IP2 in CAUSE reg */
        reg.cpr[0][CAUSE] |= 0x0400;
   

   
        /* if IP2 or IE bits in STATUS reg are clr: return */
        if((reg.cpr[0][STATUS] & 0x0401) != 0x0401)
                return;

      
   
        /* if EXL or ERL bits in STATUS reg are set: return */
        if(reg.cpr[0][STATUS] & 0x0006)
                return;



        PRINT_INTR(DBG_INTR_AI, "AI")


   
        /* set ExcCode to Int (0) */
        reg.cpr[0][CAUSE] &= ~EXC_CODE__MASK;
        


        jump_to_common_exception_vector();

} /* void rs4300i_MI_Intr_AI() */





static void rs4300i_MI_Intr_VI()
{
        PRINT_INTR_INFO(DBG_INTR_VI, "VI")

        CheckKeyboardSystem();
#ifndef PUR_VI_INTERRUPT
        dispatch(DISPLAY_REFRESH, 0);   /* pur did not do this! */
#endif

        /* add MI_INTR_VI (0x08) in MI INTR reg */
        ((uWORD *)mem.mi_reg)[2] |= MI_INTR_VI;


   
        /* if MI_INTR_VI (0x08) in MI INTR MASK reg is disabled: return */
        if(!(((uWORD *)mem.mi_reg)[3] & MI_INTR_VI))
                return;



        /* set IP2 in CAUSE reg */
        reg.cpr[0][13] |= 0x0400;
   

   
        /* if IP2 or IE bits in STATUS reg are clr: return */
        if((reg.cpr[0][12] & 0x0401) != 0x0401)
                return;

      
   
        /* if EXL or ERL bits in STATUS reg are set: return */
        if(reg.cpr[0][12] & 0x0006)
                return;



        PRINT_INTR(DBG_INTR_VI, "VI")


   
        /* set ExcCode to Int (0) */
        reg.cpr[0][CAUSE] &= ~EXC_CODE__MASK;



        jump_to_common_exception_vector();

} /* static void rs4300i_MI_Intr_VI() */





void rs4300i_MI_Intr_PI()
{
        PRINT_INTR_INFO(DBG_INTR_PI, "PI")

        /* add MI_INTR_PI (0x10) in MI INTR reg */
        ((uWORD *)mem.mi_reg)[2] |= MI_INTR_PI;


   
        /* if MI_INTR_PI (0x10) in MI INTR MASK reg is disabled: return */
        if(!(((uWORD *)mem.mi_reg)[3] & MI_INTR_PI))
                return;



        /* set IP2 in CAUSE reg */
        reg.cpr[0][CAUSE] |= 0x0400;
   

   
        /* if IP2 or IE bits in STATUS reg are clr: return */
        if((reg.cpr[0][STATUS] & 0x0401) != 0x0401)
                return;

      
   
        /* if EXL or ERL bits in STATUS reg are set: return */
        if(reg.cpr[0][STATUS] & 0x0006)
                return;



        PRINT_INTR(DBG_INTR_PI, "PI")


   
        /* set ExcCode to Int (0) */
        reg.cpr[0][CAUSE] &= ~EXC_CODE__MASK;



        jump_to_common_exception_vector();

} /* void rs4300i_MI_Intr_PI() */





void rs4300i_MI_Intr_DP()
{
        PRINT_INTR_INFO(DBG_INTR_DP, "DP")

        /* add MI_INTR_DP (0x20) in MI INTR reg */
        ((uWORD *)mem.mi_reg)[2] |= MI_INTR_DP;


   
        /* if MI_INTR_DP (0x20) in MI INTR MASK reg is disabled: return */
        if(!(((uWORD *)mem.mi_reg)[3] & MI_INTR_DP))
                return;
//puts("H A L L O");



        /* set IP2 in CAUSE reg */
        reg.cpr[0][CAUSE] |= 0x0400;
//puts("H A L L O");
   

   
        /* if IP2 or IE bits in STATUS reg are clr: return */
        if((reg.cpr[0][STATUS] & 0x0401) != 0x0401)
                return;

      
//puts("H A L L O");   
        /* if EXL or ERL bits in STATUS reg are set: return */
        if(reg.cpr[0][STATUS] & 0x0006)
                return;


        PRINT_INTR(DBG_INTR_DP, "DP")


   
        /* set ExcCode to Int (0) */
        reg.cpr[0][CAUSE] &= ~EXC_CODE__MASK;
 


        jump_to_common_exception_vector();

} /* void rs4300i_MI_Intr_DP() */





/* IP7 */

static void rs4300i_CompareInterrupt()
{
        PRINT_INTR_INFO(DBG_INTR_COMP, "CP")

        /* set IP7 in CAUSE reg */
        reg.cpr[0][CAUSE] |= 0x8000;
   

   
        /* if IP7 or IE bits in STATUS reg are clr: return */
        if((reg.cpr[0][STATUS] & 0x8001) != 0x8001)
                return;

      
   
        /* if EXL or ERL bits in status reg are set: return */
        if(reg.cpr[0][12] & 0x0006)
                return;



        PRINT_INTR(DBG_INTR_COMP, "CP")


 
        /* set ExcCode to Int (0) */
        reg.cpr[0][CAUSE] &= ~EXC_CODE__MASK;
 


        jump_to_common_exception_vector();
   
} /* static void rs4300i_CompareInterrupt() */





void rs4300i_Check_MI_Interrupts()
{
        static uWORD     mi_old_intr_mask_reg = MI_INTR_NO;



        if(mi_old_intr_mask_reg != ((uWORD *)mem.mi_reg)[3])
        {
                mi_old_intr_mask_reg = ((uWORD *)mem.mi_reg)[3];

                
                
                if( ((uWORD *)mem.mi_reg)[3] & ((uWORD *)mem.mi_reg)[2] & MI_INTR_VI )
                {
                        rs4300i_MI_Intr_VI();
                        return;
                }

                if( ((uWORD *)mem.mi_reg)[3] & ((uWORD *)mem.mi_reg)[2] & MI_INTR_SI )
                {
                        rs4300i_MI_Intr_SI();
                        return;
                }

                if( ((uWORD *)mem.mi_reg)[3] & ((uWORD *)mem.mi_reg)[2] & MI_INTR_SP )
                {
                        rs4300i_MI_Intr_SP();
                        return;
                }

                if( ((uWORD *)mem.mi_reg)[3] & ((uWORD *)mem.mi_reg)[2] & MI_INTR_DP )
                {
                        rs4300i_MI_Intr_DP();
                        return;
                }

                if( ((uWORD *)mem.mi_reg)[3] & ((uWORD *)mem.mi_reg)[2] & MI_INTR_AI )
                {
                        rs4300i_MI_Intr_AI();
                        return;
                }

                if( ((uWORD *)mem.mi_reg)[3] & ((uWORD *)mem.mi_reg)[2] & MI_INTR_PI )
                {
                        rs4300i_MI_Intr_PI();
                        return;
                }

        } /* if(mi_old_intr_mask_reg != ((uWORD *)mem.mi_reg)[3]) */

} /* void rs4300i_Check_MI_Interrupts() */





void rs4300i_Check_CPU_Interrupts()
{
        if(reg.cpr[0][STATUS] & reg.cpr[0][CAUSE] & 0xff00)   /* if any interrupt within mask is pending */
        {
                /* if IE bit in STATUS reg is set */
                if(reg.cpr[0][STATUS] & 0x0001)
                {
                        /* if EXL and ERL bits in STATUS reg are clr */
                        if((reg.cpr[0][STATUS] & 0x0006) == 0x0000)
                        {                                      
#ifdef DEBUG
                                if(prefs.debug & DBG_INTR_VI)
                                        if((reg.cpr[0][CAUSE] & 0x0400) && (((uWORD *)mem.mi_reg)[2] & MI_INTR_VI))
                                                printf("Interrupt (VI) @ %8lx - started pending\n", (uWORD)reg.cpr[0][COUNT]);
                                if(prefs.debug & DBG_INTR_SI)
                                        if((reg.cpr[0][CAUSE] & 0x0400) && (((uWORD *)mem.mi_reg)[2] & MI_INTR_SI))
                                                printf("Interrupt (SI) @ %8lx - started pending\n", (uWORD)reg.cpr[0][COUNT]);
                                if(prefs.debug & DBG_INTR_SP)
                                        if((reg.cpr[0][CAUSE] & 0x0400) && (((uWORD *)mem.mi_reg)[2] & MI_INTR_SP))
                                                printf("Interrupt (SP) @ %8lx - started pending\n", (uWORD)reg.cpr[0][COUNT]);
                                if(prefs.debug & DBG_INTR_DP)
                                        if((reg.cpr[0][CAUSE] & 0x0400) && (((uWORD *)mem.mi_reg)[2] & MI_INTR_DP))
                                                printf("Interrupt (DP) @ %8lx - started pending\n", (uWORD)reg.cpr[0][COUNT]);
                                if(prefs.debug & DBG_INTR_PI)
                                        if((reg.cpr[0][CAUSE] & 0x0400) && (((uWORD *)mem.mi_reg)[2] & MI_INTR_PI))
                                                printf("Interrupt (PI) @ %8lx - started pending\n", (uWORD)reg.cpr[0][COUNT]);
                                if(prefs.debug & DBG_INTR_AI)
                                        if((reg.cpr[0][CAUSE] & 0x0400) && (((uWORD *)mem.mi_reg)[2] & MI_INTR_AI))
                                                printf("Interrupt (AI) @ %8lx - started pending\n", (uWORD)reg.cpr[0][COUNT]);

// maybe the compare interrupt should check COMPARE reg
// it's possible that this is a BUG

                                if(prefs.debug & DBG_INTR_COMP)
                                        if((reg.cpr[0][CAUSE] & 0x8000))
                                                printf("Interrupt (CP) @ %8lx - started pending\n", (uWORD)reg.cpr[0][COUNT]);
#endif

                                fflush(stdout);
 
                                if( 0xff00 & reg.cpr[0][STATUS] & reg.cpr[0][CAUSE] )
                                {
                                        //puts("YEAH!!!");
                                        jump_to_common_exception_vector();
                                }
                  
                        } /* if((reg.cpr[0][STATUS] & 0x0006) == 0x0000) */

                } /* if(reg.cpr[0][STATUS] & 0x0001) */

        } /* if(reg.cpr[0][STATUS] & reg.cpr[0][CAUSE] & 0xff00) */

} /* void rs4300i_Check_CPU_Interrupts() */







