#
# TrueReality - N64/cpu.c
# Copyright (C) 1998 Guenther Sohler
#
# This program is free software; you can redistribute it and/
# or modify it under the terms of the GNU General Public Li-
# cence as published by the Free Software Foundation; either
# version 2 of the Licence, or any later version.
#
# This program is distributed in the hope that it will be use-
# ful, but WITHOUT ANY WARRANTY; without even the implied war-
# ranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public Licence for more details.
#
# You should have received a copy of the GNU General Public
# Licence along with this program; if not, write to the Free
# Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
# USA.
#
# Information about me (the author):
#   Niki W. Waibel, Reichenau 20, 6890 Lustenau, Austria - EUROPE
#   niki.waibel@gmx.net
#
# Information the author of the assambler CPU core:
#   Guenther Sohler, Lerschen 1 6914 Hohenweiler - EUROPE
#   guenther.sohler@newlogic.at
#

#define	INDEX		0
#define RANDOM		8
#define ENTRYLO0	16
#define	ENTRYLO1	24
#define CONTEXT		32
#definee PAGEMASK	40
#define	WIRED		48
#define	BADVADDR	56
#define COUNT		72
#define ENTRYHI		80
#define COMPARE		88
#define	STATUS		96
#define	CAUSE		104
#define EPC		112
#define PRID		120
#define CONFIG		128
#define LLADDR		136
#define WATCHLO		144
#define WATCHHI		152
#define XCONTEXT	160
#define ECC		208
#define	CACHEERR	216
#define TAGLO		224
#define TAGHI		232
#define	ERROREPC	240	

#define	RS4300I_DO_OR_CHECK_NOTHING	0
#define RS4300I_CHECK_CPU_INTERRUPTS	1
#define RS4300I_CHECK_MI_INTERUPTS	2
#define RS4300I_DO_SI_INTERRUPT		4
#define RS4300I_DO_SP_INTERRUPT		8
#define RS4300I_DO_PI_INTERRUPT		256
#define RS4300I_DO_INTEGER_OVERFLOW_EXCEPTION	16
#define	RS4300I_DO_FLOATING_POINT_EXCEPTION	32
#define RS4300I_DO_SYSCALL_EXCEPTION	64
#define RS4300I_DO_TRAP_EXCEPTION	128

#define	rom_length	
#define	rd_ram		mem+0
#define	sp_dmem		mem+4
#define	sp_imem		mem+8
#define	pi_rom		mem+12	
#define	pi_ram		mem+16
#define	pi_ram_w	mem+20
#define	rd_reg		mem+24
#define	sp_reg		mem+28
#define	dpc_reg		mem+32
#define	dps_reg		mem+36
#define	mi_reg		mem+40
#define	mi_reg_w	mem+44
#define	vi_reg		mem+48
#define	ai_reg		mem+52
#define	pi_reg		mem+56
#define pi_reg_w	mem+60
#define	ri_reg		mem+64
#define	si_reg		mem+68
#define null_mem	mem+72
#define	sp_pc_reg	mem+76
#define	sp_reg_ibist	mem+80
#define sp_reg_status_w	mem+84
#define	mi_reg_mode_w	mem+88
#define	mi_reg_intr_mask_w	mem+92
#define	pi_reg_status_w	mem+96

#define	romlen		rom+4
#define	romimage	rom+0

#define	REG_OFS		reg+0
#define	CPR0_OFS	reg+256
#define	CPR1_OFS	reg+512
#define	CPR2_OFS	reg+768
#define	CCR0_OFS	reg+1024
#define	CCR1_OFS	reg+1280
#define CCR2_OFS	reg+1536
#define	HI_OFS		reg+1792
#define	LO_OFS		reg+1800
#define PC		reg+1808
#define	PC_DELAY	reg+1812
#define	DELAY		reg+1816
#define DOSOMETHING	reg+1820
#define HALT		reg+1824
#define	CODE		reg+1828
#define	LL_BIT		reg+1832

.macro GOTONEXTCMD
	RET
.endm

.macro GETSRCREG
	MOVL	%EAX,%EBX		#extract SRCREG register
	SHRL	$18,%EBX
	ANDL	$248,%EBX
.endm

.macro GETINDEX
	MOVSWL	%AX,%EBX		#get relative offset
.endm

.macro GETTARGREG
	MOVL	%EAX,%EBX		#extract TARGREG register content
	SHRL	$13,%EBX
	ANDL	$248,%EBX
.endm

.macro GETDSTREG
	MOVL	%EAX,%EBX		#extract TARGREG register content
	SHRL	$8,%EBX
	ANDL	$248,%EBX
.endm

.macro GETSA
	MOVL	%EAX,%ECX		
	SHRL	$6,%ECX
	ANDL	$63,%ECX
.endm

.macro GETTARGET
	MOVL	%EAX,%EBX
	ANDL	$0x03ffffff,%EBX
.endm

.macro	GETFTARGREG
	MOVL	%EAX,%EBX		#extract SRCREG register
	SHRL	$13,%EBX
	ANDL	$248,%EBX
.endm

.macro	GETFSRCREG
	MOVL	%EAX,%EBX		#extract SRCREG register
	SHRL	$8,%EBX
	ANDL	$248,%EBX
.endm	

.macro	GETFDSTREG
	MOVL	%EAX,%EBX		#extract SRCREG register
	SHRL	$3,%EBX
	ANDL	$248,%EBX
.endm


.macro	STOREFTARGDWORD
	BTL	$26,CPR0_OFS+96
	JNC	STOREFTARGDWORD1\@
	MOVL	%EAX,%ECX		#extract SRCREG register
	SHRL	$13,%ECX
	ANDL	$248,%ECX
	MOVL	%EDX,CPR1_OFS(%ECX)
	MOVL	%EBX,CPR1_OFS+4(%ECX)
	JMP	STOREFTARGDWORD2\@
STOREFTARGDWORD1\@:
	MOVL	%EAX,%ECX		#extract SRCREG register
	SHRL	$13,%ECX
	ANDL	$248,%ECX
	MOVL	%EDX,CPR1_OFS(%ECX)
	SARL	$31,%EDX
	MOVL	%EDX,CPR1_OFS+4(%ECX)
	MOVL	%EBX,CPR1_OFS+8(%ECX)
	SARL	$31,%EBX
	MOVL	%EBX,CPR1_OFS+12(%ECX)
STOREFTARGDWORD2\@:
.endm

.macro	LOADFTARGDWORD
	BTL	$26,CPR0_OFS+96
	JNC	LOADFTARGDWORD1\@
	MOVL	%EAX,%ECX		#extract SRCREG register
	SHRL	$13,%ECX
	ANDL	$248,%ECX
	MOVL	CPR1_OFS(%ECX),%EDX
	MOVL	CPR1_OFS+4(%ECX),%EBX
	JMP	LOADFTARGDWORD2\@
LOADFTARGDWORD1\@:
	MOVL	%EAX,%ECX		#extract SRCREG register
	SHRL	$13,%ECX
	ANDL	$248,%ECX
	MOVL	CPR1_OFS(%ECX),%EDX
	MOVL	CPR1_OFS+8(%ECX),%EBX
LOADFTARGDWORD2\@:
.endm

.macro	LOADFSRCWORD
	BTL	$26,CPR0_OFS+96
	JNC	LOADFSRCWORD1\@
	MOVL	%EAX,%ECX		#extract SRCREG register
	SHRL	$8,%ECX
	ANDL	$248,%ECX
	MOVL	CPR1_OFS(%ECX),%EDX
	JMP	LOADFSRCWORD2\@
LOADFSRCWORD1\@:
	MOVL	%EAX,%ECX		#extract SRCREG register
	SHRL	$8,%ECX
	ANDL	$248,%ECX
	MOVL	CPR1_OFS(%ECX),%EDX
LOADFSRCWORD2\@:
.endm

.macro	LOADFSRCDWORD
	BTL	$26,CPR0_OFS+96
	JNC	LOADFSRCDWORD1\@
	MOVL	%EAX,%ECX		#extract SRCREG register
	SHRL	$8,%ECX
	ANDL	$248,%ECX
	MOVL	CPR1_OFS(%ECX),%EDX
	MOVL	CPR1_OFS+4(%ECX),%EBX
	JMP	LOADFSRCDWORD2\@
LOADFSRCDWORD1\@:
	MOVL	%EAX,%ECX		#extract SRCREG register
	SHRL	$8,%ECX
	ANDL	$248,%ECX
	MOVL	CPR1_OFS(%ECX),%EDX
	MOVL	CPR1_OFS+8(%ECX),%EBX
LOADFSRCDWORD2\@:
.endm


.macro	STOREFSRCWORD
	BTL	$26,CPR0_OFS+96
	JNC	STOREFSRCWORD1\@
	MOVL	%EAX,%ECX		#extract SRCREG register
	SHRL	$8,%ECX
	ANDL	$248,%ECX
	MOVL	%EDX,CPR1_OFS(%ECX)
	SARL	$31,%EDX
	MOVL	%EDX,CPR1_OFS+4(%ECX)
	JMP	STOREFSRCWORD2\@
STOREFSRCWORD1\@:
	MOVL	%EAX,%ECX		#extract SRCREG register
	SHRL	$8,%ECX
	ANDL	$248,%ECX
	MOVL	%EDX,CPR1_OFS(%ECX)
	SARL	$31,%EDX
	MOVL	%EDX,CPR1_OFS+4(%ECX)

STOREFSRCWORD2\@:
.endm

.macro	STOREFSRCDWORD
	BTL	$26,CPR0_OFS+96
	JNC	STOREFSRCDWORD1\@
	MOVL	%EAX,%ECX		#extract SRCREG register
	SHRL	$8,%ECX
	ANDL	$248,%ECX
	MOVL	%EDX,CPR1_OFS(%ECX)
	MOVL	%EBX,CPR1_OFS+4(%ECX)
	JMP	STOREFSRCDWORD2\@
STOREFSRCDWORD1\@:
	MOVL	%EAX,%ECX		#extract SRCREG register
	SHRL	$8,%ECX
	ANDL	$248,%ECX
	MOVL	%EDX,CPR1_OFS(%ECX)
	SARL	$31,%EDX
	MOVL	%EDX,CPR1_OFS+4(%ECX)
	MOVL	%EBX,CPR1_OFS+8(%ECX)
	SARL	$31,%EBX
	MOVL	%EBX,CPR1_OFS+12(%ECX)
STOREFSRCDWORD2\@:
.endm


.macro	LOADFSRCWORDFPU
	MOVL	%EAX,%ECX		#extract SRCREG register
	SHRL	$8,%ECX
	ANDL	$248,%ECX
	FLDS	CPR1_OFS(%ECX)
.endm

.macro	LOADFSRCDWORDFPU
	BTL	$26,CPR0_OFS+96
	JNC	LOADFSRCDWORDFPU1\@
	MOVL	%EAX,%ECX		#extract SRCREG register
	SHRL	$8,%ECX
	ANDL	$248,%ECX
	FLDL	CPR1_OFS(%ECX)
	JMP	LOADFSRCDWORDFPU2\@
LOADFSRCDWORDFPU1\@:
	MOVL	%EAX,%ECX		#extract SRCREG register
	SHRL	$8,%ECX
	ANDL	$248,%ECX
	PUSHL	CPR1_OFS+8(%ECX)
	PUSHL	CPR1_OFS(%ECX)
	FLDL	(%ESP)
	ADDL	$8,%ESP
LOADFSRCDWORDFPU2\@:
.endm


.macro	LOADFSRCIWORDFPU
	MOVL	%EAX,%ECX		#extract SRCIREG register
	SHRL	$8,%ECX
	ANDL	$248,%ECX
	FILDL	CPR1_OFS(%ECX)
.endm

.macro	LOADFSRCIDWORDFPU
	BTL	$26,CPR0_OFS+96
	JNC	LOADFSRCIDWORDFPU1\@
	MOVL	%EAX,%ECX		#extract SRCIREG register
	SHRL	$8,%ECX
	ANDL	$248,%ECX
	FILDLL	CPR1_OFS(%ECX)
	JMP	LOADFSRCIDWORDFPU2\@
LOADFSRCIDWORDFPU1\@:
	MOVL	%EAX,%ECX		#extract SRCIREG register
	SHRL	$8,%ECX
	ANDL	$248,%ECX
	PUSHL	CPR1_OFS+8(%ECX)
	PUSHL	CPR1_OFS(%ECX)
	FILDLL	(%ESP)
	ADDL	$8,%ESP
LOADFSRCIDWORDFPU2\@:
.endm



.macro	LOADFTARGWORDFPU
	MOVL	%EAX,%ECX		#extract TARGREG register
	SHRL	$13,%ECX
	ANDL	$248,%ECX
	FLDS	CPR1_OFS(%ECX)
.endm

.macro	LOADFTARGDWORDFPU
	BTL	$26,CPR0_OFS+96
	JNC	LOADFTARGDWORDFPU1\@
	MOVL	%EAX,%ECX		#extract TARGREG register
	SHRL	$13,%ECX
	ANDL	$248,%ECX
	FLDL	CPR1_OFS(%ECX)
	JMP	LOADFTARGDWORDFPU2\@
LOADFTARGDWORDFPU1\@:
	MOVL	%EAX,%ECX		#extract TARGREG register
	SHRL	$13,%ECX
	ANDL	$248,%ECX
	PUSHL	CPR1_OFS+8(%ECX)
	PUSHL	CPR1_OFS(%ECX)
	FLDL	(%ESP)
	ADDL	$8,%ESP
LOADFTARGDWORDFPU2\@:
.endm

.macro	STOREFDSTWORDFPU
	MOVL	%EAX,%ECX		#extract DSTREG register
	SHRL	$3,%ECX
	ANDL	$248,%ECX
	FSTPS	CPR1_OFS(%ECX)
.endm

.macro	STOREFDSTDWORDFPU
	BTL	$26,CPR0_OFS+96
	JNC	STOREFDSTDWORDFPU1\@
	MOVL	%EAX,%ECX		#extract DSTREG register
	SHRL	$3,%ECX
	ANDL	$248,%ECX
	FSTPL	CPR1_OFS(%ECX)
	JMP	STOREFDSTDWORDFPU2\@
STOREFDSTDWORDFPU1\@:
	MOVL	%EAX,%ECX		#extract DSTREG register
	SHRL	$3,%ECX
	ANDL	$248,%ECX
	SUBL	$8,%ESP
	FSTPL	(%ESP)
	POPL	%EDX
	POPL	%EBX
	MOVL	%EDX,CPR1_OFS(%ECX)
	SARL	$31,%EDX
	MOVL	$0,CPR1_OFS+4(%ECX)
	MOVL	%EBX,CPR1_OFS+8(%ECX)
	SARL	$31,%EBX
	MOVL	$0,CPR1_OFS+12(%ECX)
STOREFDSTDWORDFPU2\@:
.endm

.macro	STOREFDSTIWORDFPU
	MOVL	%EAX,%ECX		#extract DSTREG register
	SHRL	$3,%ECX
	ANDL	$248,%ECX
	FISTPL	CPR1_OFS(%ECX)
	MOVL	CPR1_OFS(%ECX),%EDX
	SARL	$31,%EDX
	MOVL	%EDX,CPR1_OFS+4(%ECX)
.endm

.macro	STOREFDSTIDWORDFPU
	BTL	$26,CPR0_OFS+96
	JNC	STOREFDSTIDWORDFPU1\@
	MOVL	%EAX,%ECX		#extract DSTREG register
	SHRL	$3,%ECX
	ANDL	$248,%ECX
	FISTPLL	CPR1_OFS(%ECX)
	JMP	STOREFDSTIDWORDFPU2\@
STOREFDSTIDWORDFPU1\@:
	MOVL	%EAX,%ECX		#extract DSTREG register
	SHRL	$3,%ECX
	ANDL	$248,%ECX
	SUBL	$8,%ESP
	FISTPLL	(%ESP)
	POPL	%EDX
	POPL	%EBX
	MOVL	%EDX,CPR1_OFS(%ECX)
	SARL	$31,%EDX
	MOVL	$0,CPR1_OFS+4(%ECX)
	MOVL	%EBX,CPR1_OFS+8(%ECX)
	SARL	$31,%EBX
	MOVL	$0,CPR1_OFS+12(%ECX)
STOREFDSTIDWORDFPU2\@:
.endm

