#include "macros.S"

.globl a_phys_read_addr
.globl a_phys_write_addr
.globl doReadMemByte
.globl doReadMemHalfWord
.globl doReadMemWord
.globl doReadMemDoubleWord
.globl doWriteMemByte
.globl doWriteMemHalfWord
.globl doWriteMemWord
.globl doWriteMemDoubleWord


.align 4
doReadMemByte:
	MOVL	4(%ESP),%EAX
	XORL	$3,%EAX
	CALL	a_phys_read_addr
	MOVB	(%EAX),%AL
	RET

doReadMemHalfWord:
	MOVL	4(%ESP),%EAX
	XORL	$2,%EAX
	CALL	a_phys_read_addr
	MOVW	(%EAX),%AX
	RET

doReadMemWord:
	MOVL	4(%ESP),%EAX
	CALL	a_phys_read_addr
	MOVL	(%EAX),%EAX
	RET

doReadMemDoubleWord:
	MOVL	4(%ESP),%EAX
	CALL	a_phys_read_addr
	MOVL	(%EAX),%EDX
	MOVL	4(%EAX),%EAX
	RET

doWriteMemByte:
	MOVL	8(%ESP),%EAX
	XORL	$3,%EAX
	CALL	a_phys_write_addr
	MOVB	4(%ESP),%BL
	MOVB	%BL,(%EAX)
	CMPL	$0,mem_todo
	JNE	finalcheck
	RET

doWriteMemHalfWord:
	MOVL	8(%ESP),%EAX
	XORL	$2,%EAX
	CALL	a_phys_write_addr
	MOVW	4(%ESP),%BX
	MOVW	%BX,(%EAX)
	CMPL	$0,mem_todo
	JNE	finalcheck
	RET


doWriteMemWord:
	MOVL	8(%ESP),%EAX
	CALL	a_phys_write_addr
	MOVL	4(%ESP),%EBX
	MOVL	%EBX,(%EAX)
	CMPL	$0,mem_todo
	JNE	finalcheck
	RET

doWriteMemDoubleWord:
	MOVL	12(%ESP),%EAX
	PUSHL	%EAX
	CALL	a_phys_write_addr
	POPL	%EDX
	MOVL	4(%ESP),%EDX
	MOVL	8(%ESP),%EBX
	MOVL	%EDX,4(%EAX)
	MOVL	%EBX,(%EAX)
	CMPL	$0,mem_todo
	JNE	finalcheck
	RET

finalcheck:
	MOVL	mem_todo,%ESI
	shrw	$1,%SI
	jnc	finalcheck1
	call	update_mi_reg
	NOP
finalcheck1:
	shrw	$1,%SI
	jnc	finalcheck2
	call	update_pi_reg
	NOP
finalcheck2:
	shrw	$1,%SI
	jnc	finalcheck3
	call	update_sp_reg
	NOP
finalcheck3:
	shrw	$1,%SI
	jnc	finalcheck4
	call	check_vi_reg
	NOP
finalcheck4:
	shrw	$1,%SI
	jnc	finalcheck5
	call	dma_pif2dram
	NOP
finalcheck5:
	shrw	$1,%SI
	jnc	finalcheck6
	call	dma_dram2pif
	NOP
finalcheck6:
	shrw	$1,%SI
	jnc	finalcheck7
	call	dma_pi_copy
	NOP
finalcheck7:
	shrw	$1,%SI
	jnc	finalcheck8
	call	dma_idmem2dram
	NOP
finalcheck8:
	shrw	$1,%SI
	jnc	finalcheck9
	call	dma_dram2idmem
	NOP
finalcheck9:
	shrw	$1,%SI
	jnc	finalcheck10
	call	do_display_refresh
	NOP
finalcheck10:
	shrw	$1,%SI
	jnc	finalcheck11
	call	update_vi_reg
	NOP
finalcheck11:
	shrw	$1,%SI
	jnc	finalcheck12
	call	check_ai_reg
	NOP
finalcheck12:
	MOVL	$0,mem_todo
	RET

a_phys_read_addr:
	ANDL	$0x1fffffff,%EAX
	CMPL	$0x00400000,%EAX
	JAE	a_phys_read_addr1
	ADDL	rd_ram,%EAX
	RET
a_phys_read_addr1:
	CMPL	$0x04000000,%EAX
	JAE	a_phys_read_addr2
	SUBL	$0x03f00000,%EAX
	JC	a_phys_read_addr_null_mem
	ADDL	rd_reg,%EAX
	RET
a_phys_read_addr2:
	SUBL	$0x05000000,%EAX
	JC	a_phys_read_addr5
	SUBL	$0x01000000,%EAX
	JC	a_phys_read_addr_null_mem
	SUBL	$0x02000000,%EAX
	JNC	a_phys_read_addr2a
	ANDL	$0x01ffffff,%EAX
	ADDL	romimage,%EAX
	RET
a_phys_read_addr2a:
	SUBL	$0x08000000,%EAX
	JNC	a_phys_read_addr2b
	ANDL	$0x07ffffff,%EAX
	ADDL	romimage,%EAX
	RET
a_phys_read_addr2b:		
	CMPL	$0x0fc00000,%EAX
	JB	a_phys_read_addr3
	CMPL	$0x0fc00800,%EAX
	JAE	a_phys_read_addr3
	SUBL	$0x0fc007c0,%EAX
	JB	a_phys_read_addr4
	ADDL	pi_ram,%EAX
	RET

a_phys_read_addr4:
	ANDL	$0x7ff,%EAX
	ADDL	pi_rom,%EAX
	RET

a_phys_read_addr3:
	CMPL	romlen,%EAX
	JAE	a_phys_read_addr_null_mem
	ADDL	romimage,%EAX
	RET


a_phys_read_addr5:
	ADDL	$0x00F80000,%EAX
	JC	a_phys_read_addr6
	ADDL	$0x00040000,%EAX
	JC	a_phys_read_addr7
	ADDL	$0x0003f000,%EAX
	JC	a_phys_read_addr8
	ADDL	sp_dmem,%EAX
	ADDL	$0x1000,%EAX
	RET

a_phys_read_addr8:
	CMPL	$0x1000,%EAX
	JAE	a_phys_read_addr_null_mem
	ADDL	sp_imem,%EAX
	RET

a_phys_read_addr7:
	CMPL	$20,%EAX
	JAE	a_phys_read_addr_null_mem
	ADDL	sp_reg,%EAX
	RET

a_phys_read_addr17:
	ADDL	$0x00080000,%EAX
	CMPL	$0x00000008,%EAX
	JAE	a_phys_read_addr_null_mem
	SUBL	$0x00000004,%EAX
	JNC	a_phys_read_addr9
	ADDL	$sp_pc_reg,%EAX
	ADDL	$4,%EAX
	RET

a_phys_read_addr9:
	MOVL	$sp_reg_ibist,%EAX
	RET

a_phys_read_addr6:
	SUBL	$0x00080000,%EAX
	JC	a_phys_read_addr17
	SUBL	$0x00100000,%EAX
	JNC	a_phys_read_addr10
	ADDL	$0x000fffe0,%EAX
	JC	a_phys_read_addr_null_mem
	ADDL	dpc_reg,%EAX
	ADDL	$0x20,%EAX
	RET

a_phys_read_addr10:
	SUBL	$0x00100000,%EAX
	JNC	a_phys_read_addr11
	ADDL	$0x000ffff0,%EAX
	JC	a_phys_read_addr_null_mem
	ADDL	dps_reg,%EAX
	ADDL	$0x10,%EAX
	RET

a_phys_read_addr11:
	SUBL	$0x00100000,%EAX
	JNC	a_phys_read_addr12
	ADDL	$0x000ffff0,%EAX
	JC	a_phys_read_addr_null_mem
	ADDL	mi_reg,%EAX
	ADDL	$0x10,%EAX
	RET

a_phys_read_addr12:
	SUBL	$0x00100000,%EAX
	JNC	a_phys_read_addr13
	ADDL	$0x000fffc8,%EAX
	JC	a_phys_read_addr_null_mem
	ADDL	$0x38,%EAX
	CMP	$10,%EAX
	JB	a_phys_read_addr121
	CMP	$0x13,%EAX
	JA	a_phys_read_addr121
	ADDL	$2,current_vertical_line
	CMPL	$0x202,current_vertical_line
	JB	a_phys_read_addr_122
	MOVL	$0,current_vertical_line
a_phys_read_addr_122:
	MOVL	$(current_vertical_line),%EAX
	RET
a_phys_read_addr121:
	ADDL	vi_reg,%EAX
	RET

a_phys_read_addr13:
	SUBL	$0x00100000,%EAX
	JNC	a_phys_read_addr14
	ADDL	$0x000fffe8,%EAX
	JNC	a_phys_read_addr_null_mem
	ADDL	$0x18,%EAX
	CMPL	$0x0c,%EAX
	JB	a_phys_read_addr13a
	CMPL	$0x0f,%EAX
	JG	a_phys_read_addr13a
	MOVL	$0x00000000,ai_reg+12
	CMPL	$2,ai_addr_len_buffer_full
	JNZ	a_phys_read_addr13b
	MOVL	$0x80000000,ai_reg+12
a_phys_read_addr13b:
	CMPL	$3,ai_addr_len_buffer_full
	jne	a_phys_read_addr13c
	MOVL	$0,ai_addr_len_buffer_full
a_phys_read_addr13c:
a_phys_read_addr13a:
	ADDL	ai_reg,%EAX
	RET

a_phys_read_addr14:
	SUBL	$0x00100000,%EAX
	JNC	a_phys_read_addr15
	ADDL	$0x000fffc0,%EAX
	JC	a_phys_read_addr_null_mem
	ADDL	pi_reg,%EAX
	ADDL	$0x40,%EAX
	RET

a_phys_read_addr15:
	SUBL	$0x00100000,%EAX
	JNC	a_phys_read_addr16
	ADDL	$0x000fffe0,%EAX
	JC	a_phys_read_addr_null_mem
	ADDL	ri_reg,%EAX
	ADDL	$0x20,%EAX
	RET

a_phys_read_addr16:
	SUBL	$0x00100000,%EAX
	JNC	a_phys_read_addr_null_mem
	ADDL	$0x000fffe0,%EAX
	JC	a_phys_read_addr_null_mem
	ADDL	si_reg,%EAX
	ADDL	$0x20,%EAX
	RET

a_phys_read_addr_null_mem:
	MOVL	null_mem,%EAX
	RET



a_phys_write_addr:
	ANDL	$0x1fffffff,%EAX
	CMPL	$0x00400000,%EAX
	JAE	a_phys_write_addr1 
	ADDL	rd_ram,%EAX
	RET
a_phys_write_addr1:
	CMPL	$0x04000000,%EAX
	JAE	a_phys_write_addr2
	SUBL	$0x03f00000,%EAX   
	JC	a_phys_write_addr_null_mem
	ADDL	rd_reg,%EAX
	RET
a_phys_write_addr2:
	SUBL	$0x05000000,%EAX
	JC	a_phys_write_addr5
	SUBL	$0x01000000,%EAX
	JC	a_phys_write_addr_null_mem
	SUBL	$0x02000000,%EAX
	JNC	a_phys_write_addr2a
	ANDL	$0x01ffffff,%EAX
	ADDL	romimage,%EAX
	RET
a_phys_write_addr2a:
	SUBL	$0x08000000,%EAX
	JNC	a_phys_write_addr2b
	ANDL	$0x07ffffff,%EAX
	ADDL	romimage,%EAX
	RET
a_phys_write_addr2b:		
	CMPL	$0x0fc00000,%EAX
	JB	a_phys_write_addr3
	CMPL	$0x0fc00800,%EAX
	JAE	a_phys_write_addr3
	SUBL	$0x0fc007c0,%EAX
	JB	a_phys_write_addr4
	ADDL	pi_ram,%EAX
	RET
a_phys_write_addr4:
	ANDL	$0x7ff,%EAX
	ADDL	pi_rom,%EAX
	RET
a_phys_write_addr3:
	CMPL	romlen,%EAX
	JAE	a_phys_write_addr_null_mem
	ADDL	romimage,%EAX
	RET
a_phys_write_addr5:
	ADDL	$0x00F80000,%EAX
	JC	a_phys_write_addr6
	ADDL	$0x00040000,%EAX
	JC	a_phys_write_addr7
	ADDL	$0x0003f000,%EAX
	JC	a_phys_write_addr8
	ADDL	sp_dmem,%EAX   
	ADDL	$0x1000,%EAX   
	RET
a_phys_write_addr8:
	CMPL	$0x1000,%EAX
	JAE	a_phys_write_addr_null_mem
	ADDL	sp_imem,%EAX
	RET
a_phys_write_addr7:
	CMPL	$20,%EAX
	JAE	a_phys_write_addr_null_mem
	MOVL	%EAX,%EBX
	ANDL	$0xfffffffc,%EBX
	ADDL	$sp_reg_tbl,%EBX
	JMP	(%EBX)
sp_reg_tbl:
	.long sp_reg_tbl_nothing
	.long sp_reg_tbl_nothing
	.long sp_reg_tbl_dram2idmem_dma
	.long sp_reg_tbl_idmem2dram_dma
	.long sp_reg_tbl_update_sp_reg
	.long sp_reg_tbl_null_mem
	.long sp_reg_tbl_null_mem
	.long sp_reg_tbl_nothing

sp_reg_tbl_dram2idmem_dma:
	ORL	$256,mem_todo
	JMP	sp_reg_tbl_nothing
sp_reg_tbl_idmem2dram_dma:
	ORL	$128,mem_todo
	JMP	sp_reg_tbl_nothing
sp_reg_tbl_update_sp_reg:
	ORL	$4,mem_todo
	ANDL	$3,%EAX
	ADDL	$sp_reg_status_w,%EAX
	RET	
sp_reg_tbl_null_mem:
	MOVL	null_mem,%EAX
	RET
sp_reg_tbl_nothing:
	ADDL	sp_reg,%EAX
	RET
a_phys_write_addr17:
	ADDL	$0x00080000,%EAX
	CMPL	$0x00000008,%EAX
	JAE	a_phys_write_addr_null_mem
	SUBL	$0x00000004,%EAX
	JNC	a_phys_write_addr9
	ADDL	$sp_pc_reg,%EAX
	ADDL	$4,%EAX
	RET
a_phys_write_addr9:
	ANDL	$3,%EAX
	ADDL	$sp_reg_ibist,%EAX
	RET
a_phys_write_addr6:
	SUBL	$0x00080000,%EAX
	JC	a_phys_write_addr17
	SUBL	$0x00100000,%EAX
	JNC	a_phys_write_addr10
	ADDL	$0x000fffe0,%EAX
	JC	a_phys_write_addr_null_mem
	ADDL	dpc_reg,%EAX
	ADDL	$0x20,%EAX
	RET
a_phys_write_addr10:
	SUBL	$0x00100000,%EAX
	JNC	a_phys_write_addr11
	ADDL	$0x000ffff0,%EAX
	JC	a_phys_write_addr_null_mem
	ADDL	dps_reg,%EAX
	ADDL	$0x10,%EAX
	RET

a_phys_write_addr11:
	SUBL	$0x00100000,%EAX
	JNC	a_phys_write_addr12
	ADDL	$0x000ffff0,%EAX
	JC	a_phys_write_addr_null_mem
	ADDL	$0x10,%EAX
	MOVL	%EAX,%EBX
	ANDL	$0xfffffffc,%EBX
	ADDL	$mi_reg_tbl,%EBX
	JMP	(%EBX)
mi_reg_tbl:
	.long mi_reg_tbl_update_mi_reg
	.long mi_reg_tbl_null_mem
	.long mi_reg_tbl_null_mem
	.long mi_reg_tbl_update_mireg1
mi_reg_tbl_update_mi_reg:
	ORL	$1,mem_todo
	ANDL	$3,%EAX
	ADDL	$mi_reg_mode_w,%EAX
	RET
mi_reg_tbl_null_mem:
	MOVL	null_mem,%EAX
	RET
mi_reg_tbl_update_mireg1:
	ORL	$1,mem_todo
	ORL	$2,DOSOMETHING
	ANDL	$3,%EAX
	ADDL	$mi_reg_intr_mask_w,%EAX
	RET
a_phys_write_addr12:
	SUBL	$0x00100000,%EAX
	JNC	a_phys_write_addr13
	ADDL	$0x000fffc8,%EAX
	JC	a_phys_write_addr_null_mem
	ADDL	$0x38,%EAX
	MOVL	%EAX,%EBX
	ANDL	$0xfffffffc,%EBX
	ADDL	$vi_reg_tbl,%EBX
	JMP	(%EBX)
vi_reg_tbl:
	.long vi_reg_tbl_check_vi
	.long vi_reg_tbl_display_refresh
	.long vi_reg_tbl_update_vi
	.long vi_reg_tbl_nothing
	.long vi_reg_tbl_clint
	.long vi_reg_tbl_nothing
	.long vi_reg_tbl_nothing
	.long vi_reg_tbl_nothing
	.long vi_reg_tbl_nothing
	.long vi_reg_tbl_nothing
	.long vi_reg_tbl_nothing
	.long vi_reg_tbl_nothing
	.long vi_reg_tbl_nothing
	.long vi_reg_tbl_nothing

vi_reg_tbl_check_vi:
	ORL	$8,mem_todo
	JMP	vi_reg_tbl_nothing

vi_reg_tbl_display_refresh:
	ORL	$1536,mem_todo
	JMP	vi_reg_tbl_nothing
vi_reg_tbl_update_vi:
	ORL	$1032,mem_todo
	JMP	vi_reg_tbl_nothing
vi_reg_tbl_clint:
	MOVL	mi_reg,%EBX
	ANDL	$0xfffffff7,8(%EBX)
	MOVL	null_mem,%EAX
	RET
vi_reg_tbl_nothing:
	ADDL	vi_reg,%EAX
	RET

a_phys_write_addr13:
	SUBL	$0x00100000,%EAX
	JNC	a_phys_write_addr14
	ADDL	$0x000fffe8,%EAX
	JNC	a_phys_write_addr_null_mem
	ADDL	$0x18,%EAX
	MOVL	%EAX,%EBX
	ANDL	$0xfffffffc,%EBX
	ADDL	$ai_reg_tbl,%EBX
	JMP	(%EBX)
ai_reg_tbl:
	.long ai_reg_tbl_1
	.long ai_reg_tbl_2
	.long ai_reg_tbl_3
	.long ai_reg_tbl_4
	.long ai_reg_tbl_nothing
	.long ai_reg_tbl_nothing

ai_reg_tbl_1:
	CMPL	$0,ai_addr_len_buffer_full
	JNE	ai_reg_tbl_nothing
	MOVL	$1,ai_addr_len_buffer_full
	JMP	ai_reg_tbl_nothing

ai_reg_tbl_2:
	CMPL	$1,ai_addr_len_buffer_full
	JNE	ai_reg_tbl_nothing
	MOVL	$2,ai_addr_len_buffer_full
	JMP	ai_reg_tbl_nothing

ai_reg_tbl_3:
	ORL	$2048,mem_todo
	CMPL	$2,ai_addr_len_buffer_full
	JNE	ai_reg_tbl_nothing
	MOVL	$3,ai_addr_len_buffer_full
	JMP	ai_reg_tbl_nothing
ai_reg_tbl_4:
	MOVL	mi_reg,%EBX
	ANDL	$0xfffffffb,8(%EBX)
	MOVL	null_mem,%EAX
	RET
ai_reg_tbl_nothing:
	ADDL	ai_reg,%EAX
	RET
a_phys_write_addr14:
	SUBL	$0x00100000,%EAX
	JNC	a_phys_write_addr15
	ADDL	$0x000fffc0,%EAX
	JC	a_phys_write_addr_null_mem
	ADDL	$0x40,%EAX
	MOVL	%EAX,%EBX
	ANDL	$0xfffffffc,%EBX
	ADDL	$pi_reg_tbl,%EBX
	JMP	(%EBX)
pi_reg_tbl:
	.long pi_reg_tbl_nothing
	.long pi_reg_tbl_nothing
	.long pi_reg_tbl_nothing
	.long pi_reg_tbl_pi_dma
	.long pi_reg_tbl_update
	.long pi_reg_tbl_nothing
	.long pi_reg_tbl_nothing
	.long pi_reg_tbl_nothing
	.long pi_reg_tbl_nothing
	.long pi_reg_tbl_nothing
	.long pi_reg_tbl_nothing
	.long pi_reg_tbl_nothing
	.long pi_reg_tbl_nothing
pi_reg_tbl_pi_dma:
	ORL	$64,mem_todo
	JMP	pi_reg_tbl_nothing
pi_reg_tbl_update:
	ORL	$2,mem_todo
	ANDL	$3,%EAX
	ADDL	$pi_reg_status_w,%EAX
	RET
pi_reg_tbl_nothing:
	ADDL	pi_reg,%EAX
	RET
a_phys_write_addr15:
	SUBL	$0x00100000,%EAX
	JNC	a_phys_write_addr16
	ADDL	$0x000fffe0,%EAX
	JC	a_phys_write_addr_null_mem
	ADDL	ri_reg,%EAX
	ADDL	$0x20,%EAX
	RET
a_phys_write_addr16:
	SUBL	$0x00100000,%EAX
	JNC	a_phys_write_addr_null_mem
	ADDL	$0x000fffe0,%EAX
	JC	a_phys_write_addr_null_mem
	ADDL	$0x20,%EAX
	MOVL	%EAX,%EBX
	ANDL	$0xfffffffc,%EBX
	ADDL	$si_reg_tbl,%EBX
	JMP	(%EBX)
si_reg_tbl:
	.long si_reg_tbl_nothing
	.long si_reg_tbl_pif2dramdma
	.long si_reg_tbl_null_mem
	.long si_reg_tbl_null_mem
	.long si_reg_tbl_dram2pifdma
	.long si_reg_tbl_null_mem
	.long si_reg_tbl_mintrsi
	.long si_reg_tbl_nothing

si_reg_tbl_pif2dramdma:
	ORL	$16,mem_todo
	JMP	si_reg_tbl_nothing
si_reg_tbl_dram2pifdma:
	ORL	$32,mem_todo
	JMP	si_reg_tbl_nothing
si_reg_tbl_null_mem:
	MOVL	null_mem,%EAX
	RET
si_reg_tbl_mintrsi:
	MOVL	mi_reg,%EBX
	ANDL	$0xfffffffb,8(%EBX)
	MOVL	si_reg,%EBX
	ANDL	$0xfffffffe,24(%EBX)
	MOVL	null_mem,%EAX
	RET
si_reg_tbl_nothing:
	ADDL	si_reg,%EAX
	RET

a_phys_write_addr_null_mem:
	MOVL	null_mem,%EAX
	RET

