/**
 * TrueReality - endian.c
 * Copyright (C) 1998, 1999 Niki W. Waibel
 *
 * This program is free software; you can redistribute it and/
 * or modify it under the terms of the GNU General Public Li-
 * cence as published by the Free Software Foundation; either
 * version 2 of the Licence, or any later version.
 *
 * This program is distributed in the hope that it will be use-
 * ful, but WITHOUT ANY WARRANTY; without even the implied war-
 * ranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public Licence for more details.
 *
 * You should have received a copy of the GNU General Public
 * Licence along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 * USA.
 *
 * Information about me (the author):
 *   Niki W. Waibel, Reichenau 20, 6890 Lustenau, Austria - EUROPE
 *   niki.waibel@gmx.net
**/





#include <stdio.h>
#include <stdlib.h>

#include <config.h>
#include <type_sizes.h>

#include "endian.h"



#ifdef DEBUG
        #include <debug.h>
#endif





/*******************************************************************************
*                                                                              *
* if endianness is wrong defined in 'Makefile',      print error message and   *
* return -1 otherwise print cpu endian and return 0                            *
*                                                                              *
*******************************************************************************/

int check_machine_endian()
{
        int result;
   
   
   
        result = get_machine_endianness();



#ifdef ENDIAN_IS_LITTLE
        if(result == BIG_ENDIAN)
        {
                fputs("\n\n"
                      "You are on a big endian cpu but you defined ENDIAN_IS_LITTLE in Makefile.\n"
                      "Change that to ENDIAN_IS_BIG and compile again."
                      "\n\n", stderr);
                return(-1);
        }
        else
                puts("Machine: little endian");
#endif

#ifdef ENDIAN_IS_BIG
        if(result == LITTLE_ENDIAN)
        {
                fputs("\n\n"
                      "You are on a little endian cpu but you defined ENDIAN_IS_BIG in Makefile.\n"
                      "Change that to ENDIAN_IS_LITTLE and compile again."
                      "\n\n", stderr);
                return(-1);
        }
        else
                puts("Machine: big endian");
#endif


        puts("-------------------------------------------------------------------------------"); fflush(stdout);
   
        return(0);
   
} /* int check_machine_endian() */








/*
** check machine endian and return endianness (defined in local endian.h)
*/
   
int get_machine_endianness()
{
        uHWORD endian;


   
        endian = 0x8037;
   
        if( ((uBYTE *)(&endian))[0] == 0x80 && ((uBYTE *)(&endian))[1] == 0x37 )
        {
                return(BIG_ENDIAN);
        }
        else
        if( ((uBYTE *)(&endian))[0] == 0x37 && ((uBYTE *)(&endian))[1] == 0x80 )
        {
                return(LITTLE_ENDIAN);
        }
        else
        {
                /*
                **  this should never happen;
                **  if it happens, it might be that 'sizeof(uHWORD)' is wrong ... check 'type_sizes.h'!!!
                */
         
                fputs("\n\n"
                      "get_endianness: Could not get endianness.\n"
                      "                This should not happen. Check 'type_sizes.h'!"
                      "\n\n", stderr);
                exit(1);
        }
   
} /* int get_machine_endianness() */






