/**
 * TrueReality - include/parser_extern.h
 * Copyright (C) 1998, 1999 Niki W. Waibel
 *
 * This program is free software; you can redistribute it and/
 * or modify it under the terms of the GNU General Public Li-
 * cence as published by the Free Software Foundation; either
 * version 2 of the Licence, or any later version.
 *
 * This program is distributed in the hope that it will be use-
 * ful, but WITHOUT ANY WARRANTY; without even the implied war-
 * ranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public Licence for more details.
 *
 * You should have received a copy of the GNU General Public
 * Licence along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 * USA.
 *
 * Information about me (the author):
 *   Niki W. Waibel, Reichenau 20, 6890 Lustenau, Austria - EUROPE
 *   niki.waibel@gmx.net
**/





#define __PARSER_EXTERN_H





#include <type_sizes.h>





/*** struct section ***/



typedef struct
{
        uWORD    addr;
        char    *name;

}
tr_stab; /* typedef struct */





struct tr_prefs
{
        /* :) ... a developer need this */
#ifndef ANSI_64
        unsigned long long int  debug;
#else
        unsigned long debug_h;
        unsigned long debug_l;
#endif
        
        
        /* N64 image or DMEM/IMEM binaries */

        int     load_what;



        /* Symbols */
        
        int     symbols;
        char    *symbol_file;
        int     n_symbols;      /* there are n_symbols symbols in table */
        tr_stab *symtab;
        
        

        /* Display */
        
        int     display;
        
        

        /* Controller/Cartridge */
        
        int     plugged_controller;
        int     eeprom;
        
        

        /* X11 */
        
        int     UseSHM;
        


        /* Joystick */
        
        int     joystick;
        

        
        /* Sound */
        
        int     sound;
        char    *sound_file;
        int     sound_bits;
        int     sound_freq;
        int     sound_stereo;



        /* this is for displays in which the depth can be specified */

#ifdef COMMAND_DEPTH
        int     depth;
#endif



        /* more should come */

        uDWORD r22;
        
}; /* struct tr_prefs */





/*** extern section ***/

extern struct tr_prefs prefs;





/*** define section ***/

/* tr_prefs.load_what */

#define N64_NOTHING     0       /* These 4 defines are used to    */
#define N64_IMAGE       1       /* determinate                    */
#define N64_IMEM        2       /* if an image or dmem/imem       */
#define N64_DMEM        4       /* is/are to load.                */



/* tr_prefs.symbols */

#define NO_SYMBOLS              0
#define USE_SYMBOLS             1



/* tr_prefs.sound */

#define NO_SOUND                0
#define NORMAL_SOUND            1
#define FILE_SOUND              2



/* tr_prefs.joystick */

#define NO_JOYSTICK             0
#define USE_JOYSTICK            1



/* tr_prefs.display */

#define NO_DISPLAY              0
#define LOCAL_DISPLAY           1
#define LOCAL_DISPLAY_PROCESS   2
#define REMOTE_DISPLAY          3



/* tr_prefs.debug */

#ifndef ANSI_64

        #define DBG_CPU_INSTR           0x0000000000000001ULL
        #define DBG_CPU_COUNT           0x0000000000000002ULL
        #define DBG_CPU_REG             0x0000000000000004ULL
        #define DBG_NATIVE              0x0000000000000008ULL
        #define DBG_RSP_INSTR           0x0000000000000010ULL
        #define DBG_RSP_COUNT           0x0000000000000020ULL
        #define DBG_RSP_REG             0x0000000000000040ULL
        #define DBG_RDP                 0x0000000000000080ULL

        #define DBG_INTR_COMP           0x0000000000000100ULL
        #define DBG_INTR_VI             0x0000000000000200ULL
        #define DBG_INTR_SI             0x0000000000000400ULL
        #define DBG_INTR_SP             0x0000000000000800ULL
        #define DBG_INTR_PI             0x0000000000001000ULL
        #define DBG_INTR_AI             0x0000000000002000ULL
        #define DBG_INTR_DP             0x0000000000004000ULL

        #define DBG_MEMORY              0x0000000000010000ULL
        #define DBG_RD_RAM              0x0000000000020000ULL
        #define DBG_RD_REG              0x0000000000040000ULL
        #define DBG_SP_DMEM             0x0000000000080000ULL
        #define DBG_SP_IMEM             0x0000000000100000ULL
        #define DBG_SP_REG              0x0000000000200000ULL
        #define DBG_SP_PC_REG           0x0000000000400000ULL
        #define DBG_SP_IBIST_REG        0x0000000000800000ULL
        #define DBG_DPC_REG             0x0000000001000000ULL
        #define DBG_DPS_REG             0x0000000002000000ULL
        #define DBG_MI_REG              0x0000000004000000ULL
        #define DBG_VI_REG              0x0000000008000000ULL

        #define DBG_AI_REG              0x0000000010000000ULL
        #define DBG_PI_REG              0x0000000020000000ULL
        #define DBG_RI_REG              0x0000000040000000ULL
        #define DBG_SI_REG              0x0000000080000000ULL
        #define DBG_PI_DOM2_ADDR1       0x0000000100000000ULL
        #define DBG_PI_DOM1_ADDR1       0x0000000200000000ULL
        #define DBG_PI_DOM2_ADDR2       0x0000000400000000ULL
        #define DBG_PI_DOM1_ADDR2       0x0000000800000000ULL
        #define DBG_PI_ROM              0x0000001000000000ULL
        #define DBG_PI_RAM              0x0000002000000000ULL
        #define DBG_PI_DOM1_ADDR3       0x0000004000000000ULL

        #define DBG_DISP_REFRESH        0x0000100000000000ULL
        #define DBG_DISP_MODE           0x0000200000000000ULL

        #define DBG_CONTR_PIF           0x0001000000000000ULL
        #define DBG_CONTR_BUTTONS       0x0002000000000000ULL

        #define DBG_AUDIO               0x0008000000000000ULL

        #define DBG_EXC_OV              0x0010000000000000ULL
        #define DBG_EXC_FPE             0x0020000000000000ULL
        #define DBG_EXC_TR              0x0040000000000000ULL
        #define DBG_EXC_BP              0x0080000000000000ULL
        #define DBG_EXC_SYS             0x0100000000000000ULL
        #define DBG_EXC_RI              0x0200000000000000ULL
        #define DBG_EXC_TLB             0x0400000000000000ULL

#else /* #ifndef ANSI_64 */

        #define DBG_CPU_INSTR           0x00000001
        #define DBG_CPU_COUNT           0x00000002
        #define DBG_CPU_REG             0x00000004
        #define DBG_NATIVE              0x00000008
        #define DBG_RSP_INSTR           0x00000010
        #define DBG_RSP_COUNT           0x00000020
        #define DBG_RSP_REG             0x00000040
        #define DBG_RDP                 0x00000080

        #define DBG_INTR_COMP           0x00000100
        #define DBG_INTR_VI             0x00000200
        #define DBG_INTR_SI             0x00000400
        #define DBG_INTR_SP             0x00000800
        #define DBG_INTR_PI             0x00001000
        #define DBG_INTR_AI             0x00002000
        #define DBG_INTR_DP             0x00004000

        #define DBG_MEMORY              0x00010000
        #define DBG_RD_RAM              0x00020000
        #define DBG_RD_REG              0x00040000
        #define DBG_SP_DMEM             0x00080000
        #define DBG_SP_IMEM             0x00100000
        #define DBG_SP_REG              0x00200000
        #define DBG_SP_PC_REG           0x00400000
        #define DBG_SP_IBIST_REG        0x00800000
        #define DBG_DPC_REG             0x01000000
        #define DBG_DPS_REG             0x02000000
        #define DBG_MI_REG              0x04000000
        #define DBG_VI_REG              0x08000000
        #define DBG_AI_REG              0x10000000
        #define DBG_PI_REG              0x20000000
        #define DBG_RI_REG              0x40000000
        #define DBG_SI_REG              0x80000000

        #define DBG_PI_DOM2_ADDR1       0x00000001
        #define DBG_PI_DOM1_ADDR1       0x00000002
        #define DBG_PI_DOM2_ADDR2       0x00000004
        #define DBG_PI_DOM1_ADDR2       0x00000008
        #define DBG_PI_ROM              0x00000010
        #define DBG_PI_RAM              0x00000020
        #define DBG_PI_DOM1_ADDR3       0x00000040

        #define DBG_DISP_REFRESH        0x00001000
        #define DBG_DISP_MODE           0x00002000

        #define DBG_CONTR_PIF           0x00010000
        #define DBG_CONTR_BUTTONS       0x00020000

        #define DBG_AUDIO               0x00080000

        #define DBG_EXC_OV              0x00100000
        #define DBG_EXC_FPE             0x00200000
        #define DBG_EXC_TR              0x00400000
        #define DBG_EXC_BP              0x00800000
        #define DBG_EXC_SYS             0x01000000
        #define DBG_EXC_RI              0x02000000

#endif /* #else #ifndef ANSI_64 */

/* tr_prefs.plugged_controller */

#define MAX_CONTR       4      /* BUG: maybe 6 ???*/

#define NOCONTR         0x00
#define CONTR_1         0x01
#define CONTR_2         0x02
#define CONTR_3         0x04
#define CONTR_4         0x08
#define CONTR_5         0x10    /* BUG: ??? */
#define CONTR_6         0x20    /* BUG: ??? */


