/**
 * TrueReality - input/joystick/amigados.c
 * Copyright (C) 1998, 1999 Niki W. Waibel
 *
 * This program is free software; you can redistribute it and/
 * or modify it under the terms of the GNU General Public Li-
 * cence as published by the Free Software Foundation; either
 * version 2 of the Licence, or any later version.
 *
 * This program is distributed in the hope that it will be use-
 * ful, but WITHOUT ANY WARRANTY; without even the implied war-
 * ranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public Licence for more details.
 *
 * You should have received a copy of the GNU General Public
 * Licence along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 * USA.
 *
 * Information about me (the author):
 *   Niki W. Waibel, Reichenau 20, 6890 Lustenau, Austria - EUROPE
 *   niki.waibel@gmx.net
**/


/**
 *
 * All of this stuff comes from AmiDog - great work (niki)
 *
**/

#include <stdio.h>
#include <string.h>
#include <fcntl.h>

#include <unistd.h>
#include <sys/ioctl.h>

#include <exec/types.h>
#include <exec/nodes.h>
#include <exec/lists.h>
#include <exec/memory.h>
#include <powerup/ppcproto/intuition.h>
#include <powerup/ppclib/interface.h>
#include <powerup/ppclib/time.h>
#include <powerup/gcclib/powerup_protos.h>
#include <powerup/ppcproto/exec.h>
#include <powerup/ppcinline/graphics.h>
#include <powerup/ppcproto/dos.h>
#include <powerup/ppcproto/lowlevel.h>

struct Library *LowLevelBase=NULL;

#include <config.h>
#include <parser_extern.h>
#include <type_sizes.h>
#include <memory_extern.h>
#include <dispatch.h>

#include "joystick.h"


int OpenJoystick() /* Always successfull */
{
  LowLevelBase = OpenLibrary("lowlevel.library", 39);

  /* Force CD32 pad (if required) */
  if (LowLevelBase != NULL)
    SetJoyPortAttrs(1, SJA_Type, SJA_TYPE_GAMECTLR, TAG_END);
  else
    prefs.joystick = NO_JOYSTICK;

  return 0;
}


int CloseJoystick() /* Always successfull */
{
  if (LowLevelBase != NULL)
    CloseLibrary(LowLevelBase);

  return 0;
}


void ReadJoystick(signed char *x, signed char *y, int *a, int *b)
{
  static unsigned long portstate;
  int jx = 0, jy = 0, multiplyer = 1;

  /* Return if library wasn't present, this function should however not be called then! */
  if (LowLevelBase == NULL)
    return;

  /* Read port 1 (joyport) */
  portstate=ReadJoyPort(1);

  /* Was any buttons pressed? */
  if (portstate&JPF_JOY_LEFT) /* Left */
    jx = -32;

  if (portstate&JPF_JOY_RIGHT) /* Right */
    jx = 32;

  if (portstate&JPF_JOY_UP) /* Up */
    jy = 32;

  if (portstate&JPF_JOY_DOWN) /* Down */
    jy = -32;

  if (portstate&JPF_BUTTON_REVERSE) /* Charcoal Reverse = Double movements */
    multiplyer = 2;

  if (portstate&JPF_BUTTON_FORWARD) /* Charcoal Forward = Quadra movements */
    multiplyer = 4;

  /* Calculate positions */
  jx *= multiplyer;
  jy *= multiplyer;

  /* Limit the values, signed 8bit (char) is -128 to 127 */
  if (jx > 127)
    jx = 127;

  if (jy > 127)
    jy = 127;

  /* Return joystick values */
  *a = 0; /* Not used */
  *b = 0; /* Not used */
  *x = (signed char)jx;
  *y = (signed char)jy;

/*
JPF_BUTTON_PLAY   - Play/Pause Button
JPF_BUTTON_GREEN  - Green Button
JPF_BUTTON_YELLOW - Yellow Button
JPF_BUTTON_BLUE   - Blue Button
JPF_BUTTON_RED    - Red Button
*/
}
