/**
 * TrueReality - input/joystick/linux21x.c
 * Copyright (C) 1998, 1999 Niki W. Waibel
 *
 * This program is free software; you can redistribute it and/
 * or modify it under the terms of the GNU General Public Li-
 * cence as published by the Free Software Foundation; either
 * version 2 of the Licence, or any later version.
 *
 * This program is distributed in the hope that it will be use-
 * ful, but WITHOUT ANY WARRANTY; without even the implied war-
 * ranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public Licence for more details.
 *
 * You should have received a copy of the GNU General Public
 * Licence along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 * USA.
 *
 * Information about me (the author):
 *   Niki W. Waibel, Reichenau 20, 6890 Lustenau, Austria - EUROPE
 *   niki.waibel@gmx.net
**/





#include <stdio.h>
#include <string.h>
#include <fcntl.h>

#include <unistd.h>
#include <sys/ioctl.h>
#include <linux/joystick.h>

#include "joystick.h"





static int              joy_fd;





int OpenJoystick()
{
        char            number_of_axes;
        char            number_of_buttons;
        int             joy_driver_version;
        char            joy_name[255];



        fputs("Opening Linux Joystick device /dev/js0 ...", stdout); fflush(stdout);
        
        joy_fd = open("/dev/js0", O_RDONLY | O_NONBLOCK);
        if(joy_fd == -1)
        {
                perror(" FAIL");
                fputs("Maybe you should 'insmod' the joystick driver(s)!?\n\n", stderr);
                return(-1);
        }

        fputs(" OK\nGetting Joystick info ...", stdout); fflush(stdout);

        ioctl(joy_fd, JSIOCGAXES,    &number_of_axes);
        ioctl(joy_fd, JSIOCGBUTTONS, &number_of_buttons);
        ioctl(joy_fd, JSIOCGVERSION, &joy_driver_version);
        if(ioctl(joy_fd, JSIOCGNAME(sizeof(joy_name)), joy_name) < 0)
                strncpy(joy_name, "Unknown", sizeof(joy_name));

        fputs(" OK\nJoystick info:\n", stdout); fflush(stdout);

        printf("   axes    = %d\n", number_of_axes);
        printf("   buttons = %d\n", number_of_buttons);
        printf("   version = %x\n", joy_driver_version);
        printf("   name    = %s\n", joy_name);


        puts("-------------------------------------------------------------------------------"); fflush(stdout);

        return(0);

} /* int OpenJoystick() */





int CloseJoystick()
{
        fputs("Closing Linux Joystick device /dev/js0 ...", stdout); fflush(stdout);
        
        close(joy_fd);
        
        fputs(" OK\n", stdout); fflush(stdout);

        puts("-------------------------------------------------------------------------------"); fflush(stdout);

        return(0);

} /* int CloseJoystick() */





void ReadJoystick(signed char *x, signed char *y, int *a, int *b)
{
        static int              button[8]       = {0, 0, 0, 0, 0, 0, 0, 0};
        static signed char      axis[2]         = {0, 0};

        struct js_event e;



        while(read(joy_fd, &e, sizeof(struct js_event)) > 0)
        {
                switch(e.type & ~JS_EVENT_INIT)
                {
                    case JS_EVENT_BUTTON:
                        if(e.number < 2)
                        {
                                button[e.number] = e.value;
                        }
                        break;
                        
                    case JS_EVENT_AXIS:
                        if(e.number < 2)
                        {
                                axis[e.number] = e.value >> 8;
                        }
                        break;
                        
                } /* switch(e.type & ~JS_EVENT_INIT) */

        } /* while(read(joy_fd, &e, sizeof(struct js_event)) > 0) */

        
        
        *a = button[0];

        *b = button[1];



        *x = axis[0];
        
        if(axis[1] == -128)       /* this value (0x80) cannot be negated! */
                axis[1] = -127;   /* hope that's ok */
                                  /* otherwise we would have to add 1 but then there would be an offset problem */
        *y = -axis[1];

} /* void ReadJoystick(signed char *x, signed char *y, int *a, int *b) */





