/**
 * TrueReality - input/keyboard/amigados.c
 * Copyright (C) 1998, 1999 Mathias "AmiDog" Roslund
 *
 * This program is free software; you can redistribute it and/
 * or modify it under the terms of the GNU General Public Li-
 * cence as published by the Free Software Foundation; either
 * version 2 of the Licence, or any later version.
 *
 * This program is distributed in the hope that it will be use-
 * ful, but WITHOUT ANY WARRANTY; without even the implied war-
 * ranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public Licence for more details.
 *
 * You should have received a copy of the GNU General Public
 * Licence along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 * USA.
 *
 * Information about me (the author):
 *   Niki W. Waibel, Reichenau 20, 6890 Lustenau, Austria - EUROPE
 *   niki.waibel@gmx.net
**/





/**
 *
 * All of this stuff comes from AmiDog - great work (niki)
 *
**/





#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <signal.h>
#include <unistd.h>

#include <exec/types.h>
#include <exec/nodes.h>
#include <exec/lists.h>
#include <exec/memory.h>
#include <powerup/ppcproto/intuition.h>
#include <powerup/ppclib/interface.h>
#include <powerup/ppclib/time.h>
#include <powerup/gcclib/powerup_protos.h>
#include <powerup/ppcproto/exec.h>
#include <powerup/ppcinline/graphics.h>
#include <powerup/ppcproto/dos.h>
#include <math.h>

#include <config.h>
#include <parser_extern.h>
#include <type_sizes.h>
#include <memory_extern.h>
#include <dispatch.h>

#include "keyboard.h"

#ifdef DISPLAY
  #include "../../output/display/display.h"
#endif
#ifdef JOYSTICK
  #include "../joystick/joystick.h"
#endif


extern struct Window *DispWindow;


void CheckKeyboardSystem()
{
  CheckVisual();
}


uWORD GetControllerData(int n)
{
    /**
    *  for now we have just one controller implemented
    *  so we ignore n
    **/

    return(CheckVisual());

} /* WORD GetControllerData(int n) */


uWORD CheckVisual()
{
  static uHWORD keystate = 0;
  static sBYTE stick_x = 0;
  static sBYTE stick_y = 0;

#ifdef JOYSTICK
  int a, b;
#endif

  struct IntuiMessage *imsg = NULL;
  ULONG imCode,imClass;

#ifdef JOYSTICK
  if(prefs.joystick == USE_JOYSTICK)
     ReadJoystick(&stick_x, &stick_y, &a, &b);
#endif

     /* Check for IDCMP messages */
     if ((imsg=(struct IntuiMessage *)GetMsg(DispWindow->UserPort)))
     {
        imClass=imsg->Class;
        imCode=imsg->Code;

        ReplyMsg((struct Message *)imsg);

        /* KeyPress and KeyReleases */

        if(imClass==IDCMP_RAWKEY)
        {
                if(imCode<128) /* Key is being pressed */
                {
                        switch(imCode)
                        {
                            case 16: /* Q, Start Button */
                                keystate |= 0x1000;
                                break;

                            case 17: /* W, A-Button */
                                keystate |= 0x8000;
                                break;

                            case 18: /* E, B-Button */
                                keystate |= 0x4000;
                                break;

                            case 62: /* 8 (num), C-Button UP */
                                keystate |= 0x0008;
                                break;

                            case 30: /* 2 (num), C-Button DOWN */
                                keystate |= 0x0004;
                                break;

                            case 45: /* 4 (num), C-Button LEFT */
                                keystate |= 0x0002;
                                break;

                            case 47: /* 6 (num), C-Button RIGHT */
                                keystate |= 0x0001;
                                break;

                            case 76: /* UP (arrow), D-Button UP */
                                keystate |= 0x0800;
                                break;

                            case 77: /* DOWN (arrow), D-Button DOWN */
                                keystate |= 0x0400;
                                break;

                            case 79: /* LEFT (arrow), D-Button LEFT */
                                keystate |= 0x0200;
                                break;

                            case 78: /* RIGTH (arrow), D-Button RIGHT */
                                keystate |= 0x0100;
                                break;

                            case 32: /* A, L-Button */
                                keystate |= 0x0020;
                                break;

                            case 33: /* S, R-Button */
                                keystate |= 0x0010;
                                break;

                            case 34: /* D, Z-Button */
                                keystate |= 0x2000;
                                break;
#ifdef JOYSTICK
                            case 39: /* K, Stick Up */
                                if(prefs.joystick != USE_JOYSTICK)
                                    if(stick_y < 120)
                                        stick_y += 0x08;
                                break;

                            case 38: /* J, Stick Down */
                                if(prefs.joystick != USE_JOYSTICK)
                                    if(stick_y > -120)
                                        stick_y -= 0x08;
                                break;

                            case 40: /* L, Stick Right */
                                if(prefs.joystick != USE_JOYSTICK)
                                    if(stick_x < 120)
                                        stick_x += 0x08;
                                break;

                            case 37: /* H, Stick Left */
                                if(prefs.joystick != USE_JOYSTICK)
                                    if(stick_x > -120)
                                        stick_x -= 0x08;
                                break;
#else
                            case 39: /* K, Stick Up */
                                if(stick_y < 120)
                                        stick_y += 0x08;
                                break;

                            case 38: /* J, Stick Down */
                                if(stick_y > -120)
                                        stick_y -= 0x08;
                                break;

                            case 40: /* L, Stick Right */
                                if(stick_x < 120)
                                        stick_x += 0x08;
                                break;

                            case 37: /* H, Stick Left */
                                if(stick_x > -120)
                                        stick_x -= 0x08;
                                break;
#endif
                            case 23: /* I, Stick Up (75%) */
                                stick_y = 96;
                                break;

                            case 22: /* U, Stick Down (75%) */
                                stick_y = -96;
                                break;

                            case 24: /* O, Stick Right (75%) */
                                stick_x = 96;
                                break;

                            case 21: /* Y, Stick Left (75%) */
                                stick_x = -96;
                                break;

                            case 56: /* ,, Stick Up (25%) */
                                stick_y = 32;
                                break;

                            case 55: /* M, Stick Down (25%) */
                                stick_y = -32;
                                break;

                            case 57: /* ., Stick Right (25%) */
                                stick_x = 32;
                                break;

                            case 54: /* N, Stick Left (25%) */
                                stick_x = -32;
                                break;

                            case 69: /* ESC */
                                dispatch(SHUTDOWN, 0);
                                break;
#ifdef OPENGL
                            case 84: /* F5, ALPHA Off */
                                DisableVisualAlpha();
                                break;

                            case 85: /* F6, ALPHA On */
                                EnableVisualAlpha();
                                break;

                            case 88: /* F9, FLAT Shading */
                                VisualFlat();
                                break;

                            case 89: /* F10, SMOOTH Shading */
                                VisualSmooth();
                                break;
#endif
                        } /* switch(imCode) */
                }
                else /* (imCode>127), Key released */
                {
                imCode-=128; /* Remove "Key released" bit */
                        switch(imCode)
                        {
                            case 16: /* Q, Start Button */
                                keystate &= ~0x1000;
                                break;

                            case 17: /* W, A-Button */
                                keystate &= ~0x8000;
                                break;

                            case 18: /* E, B-Button */
                                keystate &= ~0x4000;
                                break;

                            case 62: /* 8 (num), C-Button UP */
                                keystate &= ~0x0008;
                                break;

                            case 30: /* 2 (num), C-Button DOWN */
                                keystate &= ~0x0004;
                                break;

                            case 45: /* 4 (num), C-Button LEFT */
                                keystate &= ~0x0002;
                                break;

                            case 47: /* 6 (num), C-Button RIGHT */
                                keystate &= ~0x0001;
                                break;

                            case 76: /* UP (arrow), D-Button UP */
                                keystate &= ~0x0800;
                                break;

                            case 77: /* DOWN (arrow), D-Button DOWN */
                                keystate &= ~0x0400;
                                break;

                            case 79: /* LEFT (arrow), D-Button LEFT */
                                keystate &= ~0x0200;
                                break;

                            case 78: /* RIGHT (arrow), D-Button RIGHT */
                                keystate &= ~0x0100;
                                break;

                            case 32: /* A, L-Button */
                                keystate &= ~0x0020;
                                break;

                            case 33: /* S, R-Button */
                                keystate &= ~0x0010;
                                break;

                            case 34: /* D, Z-Button */
                                keystate &= ~0x2000;
                                break;

                            case 23: /* I, Stick Up (0%) */
                            case 56: /* ,, Stick Up (0%) */
                            case 22: /* U, Stick Down (0%) */
                            case 55: /* M, Stick Down (0%) */
                                stick_y = 0;
                                break;

                            case 24: /* O, Stick Right (0%) */
                            case 57: /* ., Stick Right (0%) */
                            case 21: /* Y, Stick Left (0%) */
                            case 54: /* N, Stick Left (0%) */
                                stick_x = 0;
                                break;

                        } /* switch(imCode) */

                } /* (imCode>127), Key released */

        } /* imClass=IDCMP_RAWKEY */

     } /* Check for IDCMP messages */

  return(((uWORD)(keystate << 16)) | ((uWORD)(uHWORD)(stick_x << 8)) | ((uWORD)(BYTE)stick_y));

} /* WORD CheckVisual() */
