/**
 * TrueReality - input/keyboard/x11.c
 * Copyright (C) 1998, 1999 Niki W. Waibel
 *
 * This program is free software; you can redistribute it and/
 * or modify it under the terms of the GNU General Public Li-
 * cence as published by the Free Software Foundation; either
 * version 2 of the Licence, or any later version.
 *
 * This program is distributed in the hope that it will be use-
 * ful, but WITHOUT ANY WARRANTY; without even the implied war-
 * ranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public Licence for more details.
 *
 * You should have received a copy of the GNU General Public
 * Licence along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 * USA.
 *
 * Information about me (the author):
 *   Niki W. Waibel, Reichenau 20, 6890 Lustenau, Austria - EUROPE
 *   niki.waibel@gmx.net
**/





#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <signal.h>
#include <unistd.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>

#include <config.h>
#include <parser_extern.h>
#include <type_sizes.h>
#include <memory_extern.h>
#include <dispatch.h>
#include <x11_extern.h>

#include "keyboard.h"

#ifdef DISPLAY
        #include "../../output/display/display.h"
#endif
#ifdef JOYSTICK
        #include "../joystick/joystick.h"
#endif
#ifdef DEBUG
        #include <debug.h>
#endif

XKeyboardState     xkeyboardstate;
static  XEvent          event;



static void SystemSwitch(KeySym key);










void CheckKeyboardSystem()
{
        KeySym          key;



    /* exit if there is no display */
        if(prefs.display == NO_DISPLAY)
                return;



        /* FocusChange */

        if(XCheckWindowEvent(dsp, win, FocusChangeMask, &event))
        {
                if(event.type == FocusIn)
                {
                        puts("***   F O C U S   I N");
                        
                        if(xkeyboardstate.global_auto_repeat == AutoRepeatModeOn)
                                XAutoRepeatOff(dsp);
                        XSelectInput(dsp, win, ButtonPressMask | FocusChangeMask | ExposureMask | KeyPressMask | KeyReleaseMask);
                }
                else
                {
                        puts("***   F O C U S   O U T");

                        XSelectInput(dsp, win, FocusChangeMask | ExposureMask);
                        if(xkeyboardstate.global_auto_repeat == AutoRepeatModeOn)
                                XAutoRepeatOn(dsp);
                }
        }



        /* ButtonPress / ButtonRelease */ 

        if(XCheckWindowEvent(dsp, win, ButtonPressMask | ButtonReleaseMask, &event))
        {
                if(event.type == ButtonPress)
                {
                        puts("***   B U T T O N   P R E S S");
                }
                else
                {
                        puts("***   B U T T O N   R E L E A S E");
                }
        }



        /* KeyPress and KeyReleases */

        if(XCheckWindowEvent(dsp, win, KeyPressMask, &event))
        {
                key = XLookupKeysym((XKeyEvent *)&event, 0);

                SystemSwitch(key);

        } /* if(XCheckWindowEvent(dsp, win, KeyPressMask, &event)) */



        /* Exposure */

        if(XCheckWindowEvent(dsp, win, ExposureMask, &event))
        {
                puts("***   E X P O S U R E");
                dispatch(DISPLAY_REFRESH, 0);
        }

} /* void CheckKeyboardSystem() */





uWORD GetControllerData(int n)
{
    /**
    *  for now we have just one controller implemented
    *  so we ignore n
    **/

        return(CheckVisual());

} /* uWORD GetControllerData(int n) */





uWORD CheckVisual()
{
        KeySym          key;

#ifdef JOYSTICK
        int             a, b;
#endif
        static sBYTE    stick_x = 0;
        static sBYTE    stick_y = 0;

        static uHWORD    keystate = 0;






#ifdef JOYSTICK
        if(prefs.joystick == USE_JOYSTICK)
                ReadJoystick(&stick_x, &stick_y, &a, &b);
#endif


        /* KeyPress and KeyReleases */

        if(XCheckWindowEvent(dsp, win, KeyPressMask | KeyReleaseMask, &event))
        {
                do
                {
                        key = XLookupKeysym((XKeyEvent *)&event, 0);

                        if(event.type == KeyPress)
                        {
                                switch(key)
                                {
                                    case XK_q:           /* Start Button */
                                        keystate |= 0x1000;
                                        break;

                                    case XK_w:                  /* A-Button */
                                        keystate |= 0x8000;
                                        break;

                                    case XK_e:                  /* B-Button */
                                        keystate |= 0x4000;
                                        break;

                                    case XK_Home:               /* C-Button UP */
                                        keystate |= 0x0008;
                                        break;

                                    case XK_End:                /* C-Button DOWN */
                                        keystate |= 0x0004;
                                        break;

                                    case XK_Delete:          /* C-Button LEFT */
                                        keystate |= 0x0002;
                                        break;

                                    case XK_Next:          /* C-Button RIGHT */
                                        keystate |= 0x0001;
                                        break;

                                    case XK_Up:              /* D-Button UP */
                                        keystate |= 0x0800;
                                        break;

                                    case XK_Down:            /* D-Button DOWN */
                                        keystate |= 0x0400;
                                        break;

                                    case XK_Left:            /* D-Button LEFT */
                                        keystate |= 0x0200;
                                        break;

                                    case XK_Right:           /* D-Button RIGHT */
                                        keystate |= 0x0100;
                                        break;

                                    case XK_s:          /* R-Button */
                                        keystate |= 0x0010;
                                        break;

                                    case XK_a:            /* L-Button */
                                        keystate |= 0x0020;
                                        break;

                                    case XK_d:                  /* Z-Button */
                                        keystate |= 0x2000;
                                        break;

#ifndef JOYSTICK
                                    case XK_k:   /* Stick Up */
                                        if(prefs.joystick != USE_JOYSTICK)
                                                if(stick_y < 120)
                                                        stick_y += 0x08;
                                        break;

                                    case XK_j:   /* Stick Down */
                                        if(prefs.joystick != USE_JOYSTICK)
                                                if(stick_y > -120)
                                                        stick_y -= 0x08;
                                        break;

                                    case XK_l:   /* Stick Right */
                                        if(prefs.joystick != USE_JOYSTICK)
                                                if(stick_x < 120)
                                                        stick_x += 0x08;
                                        break;

                                    case XK_h:   /* Stick Left */
                                        if(prefs.joystick != USE_JOYSTICK)
                                                if(stick_x > -120)
                                                        stick_x -= 0x08;
                                        break;
#endif /* #ifndef JOYSTICK */
                                    case XK_i:   /* Stick Up (75%) */
                                        stick_y = 96;
                                        break;

                                    case XK_u:   /* Stick Down (75%) */
                                        stick_y = -96;
                                        break;

                                    case XK_o:   /* Stick Right (75%) */
                                        stick_x = 96;
                                        break;

                                    case XK_y:   /* Stick Left (75%) */
                                        stick_x = -96;
                                        break;

                                    case XK_comma:   /* Stick Up (25%) */
                                        stick_y = 32;
                                        break;

                                    case XK_m:   /* Stick Down (25%) */
                                        stick_y = -32;
                                        break;

                                    case XK_period:   /* Stick Right (25%) */
                                        stick_x = 32;
                                        break;

                                    case XK_n:   /* Stick Left (25%) */
                                        stick_x = -32;
                                        break;
                    
                                    case XK_Escape:
#ifdef DISPLAY
                                    case XK_F2:
                                    case XK_F3:
                                    case XK_F8:
#endif
                                    case XK_F4:
#ifdef OPENGL
                                    case XK_F5:
                                    case XK_F6:
                                    case XK_F9:
                                    case XK_F10:
#endif
                                    case XK_F1:
                                        SystemSwitch(key);
                                        break;

                                } /* switch(key) */
                        }
                        else /* (event.type == KeyRelease) */
                        {
                                switch(key)
                                {
                                    case XK_q:           /* Start Button */
                                        keystate &= ~0x1000;
                                        break;

                                    case XK_w:                  /* A-Button */
                                        keystate &= ~0x8000;
                                        break;

                                    case XK_e:                  /* B-Button */
                                        keystate &= ~0x4000;
                                        break;

                                    case XK_Home:               /* C-Button UP */
                                        keystate &= ~0x0008;
                                        break;

                                    case XK_End:                /* C-Button DOWN */
                                        keystate &= ~0x0004;
                                        break;

                                    case XK_Delete:          /* C-Button LEFT */
                                        keystate &= ~0x0002;
                                        break;

                                    case XK_Next:          /* C-Button RIGHT */
                                        keystate &= ~0x0001;
                                        break;

                                    case XK_Up:              /* D-Button UP */
                                        keystate &= ~0x0800;
                                        break;

                                    case XK_Down:            /* D-Button DOWN */
                                        keystate &= ~0x0400;
                                        break;

                                    case XK_Left:            /* D-Button LEFT */
                                        keystate &= ~0x0200;
                                        break;

                                    case XK_Right:           /* D-Button RIGHT */
                                        keystate &= ~0x0100;
                                        break;

                                    case XK_s:          /* R-Button */
                                        keystate &= ~0x0010;
                                        break;

                                    case XK_a:            /* L-Button */
                                        keystate &= ~0x0020;
                                        break;

                                    case XK_d:                  /* Z-Button */
                                        keystate &= ~0x2000;
                                        break;

                                    case XK_i:   /* Stick Up (0%) */
                                    case XK_comma:   /* Stick Up (0%) */
                                    case XK_u:   /* Stick Down (0%) */
                                    case XK_m:   /* Stick Down (0%) */
                                        stick_y = 0;
                                        break;

                                    case XK_o:   /* Stick Right (0%) */
                                    case XK_period:   /* Stick Right (0%) */
                                    case XK_y:   /* Stick Left (0%) */
                                    case XK_n:   /* Stick Left (0%) */
                                        stick_x = 0;
                                        break;

                                } /* switch(key) */

                        } /* (event.type == KeyRelease) */
                }
                while(XCheckWindowEvent(dsp, win, KeyPressMask | KeyReleaseMask, &event));
        }




#ifdef DEBUG
        if(prefs.debug & DBG_CONTR_BUTTONS)
                printf
                (
                        "Controller: "
                        "0x%04hx %+4d %+4d (0x%08lx) (0xBBBBXXYY) "
                        "%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s\n",

                        keystate, stick_x, stick_y,
                        (uWORD)(keystate << 16) | ((uWORD)(uHWORD)(stick_x << 8)) | ((uWORD)(uBYTE)stick_y),
                        (keystate & 0x8000) ? "A "      : "",
                        (keystate & 0x4000) ? "B "      : "",
                        (keystate & 0x2000) ? "Z "      : "",
                        (keystate & 0x1000) ? "START "  : "",
                        (keystate & 0x0800) ? "Dup "    : "",
                        (keystate & 0x0400) ? "Ddown "  : "",
                        (keystate & 0x0200) ? "Dleft "  : "",
                        (keystate & 0x0100) ? "Dright " : "",
                        (keystate & 0x0080) ? ""        : "",
                        (keystate & 0x0040) ? ""        : "",
                        (keystate & 0x0020) ? "L "      : "",
                        (keystate & 0x0010) ? "R "      : "",
                        (keystate & 0x0008) ? "Cup "    : "",
                        (keystate & 0x0004) ? "Cdown "  : "",
                        (keystate & 0x0002) ? "Cleft "  : "",
                        (keystate & 0x0001) ? "Cright " : "" 
                );
#endif



        return(((uWORD)(keystate << 16)) | ((uWORD)(uHWORD)(stick_x << 8)) | ((uWORD)(uBYTE)stick_y));

} /* uWORD CheckVisual() */









static void SystemSwitch(KeySym key)
{
        switch(key)
        {
            case XK_Escape:
                puts("***   K E Y   P R E S S   (esc: exit emulation)");
                dispatch(SHUTDOWN, 0);

            case XK_F8:
                puts("***   K E Y   P R E S S   (f8: display refresh)");
#ifdef DISPLAY
                RefreshVisual();
#endif
                break;

            case XK_F4:
                puts("***   K E Y   P R E S S   (f4: halt cpu)");
                XSelectInput(dsp, win, FocusChangeMask | ExposureMask);
                if(xkeyboardstate.global_auto_repeat == AutoRepeatModeOn)
                        XAutoRepeatOn(dsp);
                dispatch(CPU_STOP, 0);
                break;

#ifdef OPENGL
            case XK_F6:
                puts("***   K E Y   P R E S S   (f6: ALPHA on)");
                EnableVisualAlpha();
                break;

            case XK_F5:
                puts("***   K E Y   P R E S S   (f5: ALPHA off)");
                DisableVisualAlpha();
                break;

            case XK_F10:
                puts("***   K E Y   P R E S S   (f10: SMOOTH shading)");
                VisualSmooth();
                break;

            case XK_F9:
                puts("***   K E Y   P R E S S   (f9: FLAT shading)");
                VisualFlat();
                break;
#endif

            case XK_F2:
                puts("***   K E Y   P R E S S   (f2: -> 320x240)");
                SetVisualSize(320, 240);
                break;

            case XK_F3:
                puts("***   K E Y   P R E S S   (f3: -> 640x480)");
                SetVisualSize(640, 480);
                break;

            case XK_F1:
                puts("***   K E Y   P R E S S   (f1: HELP - not implemented yet)");
                break;

            default:
                XPutBackEvent(dsp, &event);

        } /* switch(key) */

} /* static void SystemSwitch(KeySym key) */
