
// GameLoad.c

#include <Navigation.h>
#include <Sound.h>
#include <stdlib.h>

#include "GameLoad.h"
#include "StandardDisplay.h"
#include "UtilityFunctions.h"
#include "Alerts.h"
#include "MacOS.h"
#include "Menu.h"

#include "file_loader.h"
#include "file_loader_extern.h"
#include "parser_extern.h"
#include "Memory.h"
#include "DISPATCH.H"
#include "Display.h"


void OpenRomImage(FSSpec* theFile)
{
	Str255	tempPStr;
	int 	err;
	char*	tempBuffer;
	
	if (gRomOpened)
		CloseRom();
	
	// Select Romimage to open
	if (!theFile)
	{
		StandardFileReply reply;
		theFile = &reply.sfFile;
		if(NavServicesAvailable())
		{
			reply.sfGood = NavOpen(&reply.sfFile);
		}
		else
		{
			StandardGetFile(nil, -1, nil, &reply);
		}
		if(!reply.sfGood)
			return;
	}
	
	(void *)rominfo.real_name = malloc(256);
	FullPathname(theFile, tempPStr);
	PascalToC(tempPStr, rominfo.real_name);
	
	prefs.load_what = N64_IMAGE;

	if((err = load_files()) < 0)
	{
		free(rominfo.real_name);
		if(err == -3)
		{
			MyStandardStopAlert(MemTooLow);
		}
		else
			SysBeep(1);
		dispatch(MEMORY_CLEAR, NULL);
		return;
	}
	
	ChangeTypeCreator(theFile);
	
	if(copy_code() < 0)
	{
		free(rom.image);
		free(rominfo.real_name);
		dispatch(MEMORY_CLEAR, NULL);
		SysBeep(1);
		return;
	}
	
#ifdef AUDIO
	dispatch(AUDIO_OPEN, NULL);
	/* Pause the audio */
	AsgardESS_Pause(true);
#endif
	
	dispatch(CPU_RESET, NULL);
	
	EnableItem( GetMenuHandle( mN64 ), 0 );
	InvalMenuBar();
	
	gRomOpened = true;
	
#ifdef NATIVE
	Create_Native_Table();
#endif
	dispatch(CONTROLLER_INIT, &prefs.plugged_controller);
	OpenVisual();
}

void CloseRom()
{
	if(!gRomOpened)
		return;
	
#ifdef AUDIO
	dispatch(AUDIO_CLOSE, NULL);
#endif

	CloseVisual();
	
	free(rom.image);
	free(rominfo.real_name);
	dispatch(MEMORY_CLEAR, NULL);
	
	DisableItem( GetMenuHandle( mN64 ), 0 );
	InvalMenuBar();
	
	gRomOpened = false;
}

pascal void myNavEventProc(	const NavEventCallbackMessage /*callBackSelector*/, 
							NavCBRecPtr /*callBackParms*/, 
							NavCallBackUserData /*callBackUD*/)
{}

pascal Boolean myNavFilterProc(	AEDesc* theItem, 
								void* info, 
								NavCallBackUserData /*callBackUD*/, 
								NavFilterModes /*filterMode*/)
{
	OSErr theErr = noErr;
	Boolean display = true;
	NavFileOrFolderInfo* theInfo = (NavFileOrFolderInfo*)info;
	
	if ( theItem->descriptorType == typeFSS )
		if ( !theInfo->isFolder )
			{
				// Nothing yet
			}
	return display;
}

#define kSelectFilePrefKey		3
bool NavOpen(FSSpec *finalFSSpec)
{
	NavReplyRecord		theReply;
	NavDialogOptions	dialogOptions;
	OSErr				theErr = noErr;
	NavEventUPP			eventUPP  = NewNavEventProc( myNavEventProc );
	NavObjectFilterUPP  filterUPP = NewNavObjectFilterProc( myNavFilterProc );
	bool 				returnValue;
	
	theErr = NavGetDefaultDialogOptions(&dialogOptions);
	dialogOptions.dialogOptionFlags &= ~kNavAllowPreviews & ~kNavAllowMultipleFiles;
	
	dialogOptions.preferenceKey = kSelectFilePrefKey;
	CopyPString( dialogOptions.windowTitle, "\pMac TrueReality" );
	
	theErr = NavChooseFile(	nil,
							&theReply,
							&dialogOptions,
							eventUPP,
							nil,
							nil /*filterUPP*/,
							nil,
							nil );

	DisposeRoutineDescriptor( eventUPP  );
	DisposeRoutineDescriptor( filterUPP );

	returnValue = false;
	if( theReply.validRecord && theErr == noErr )
	{
		AEDesc 	resultDesc;

		theErr = AECoerceDesc( &theReply.selection, typeFSS, &resultDesc );
		if (theErr == noErr)
		{
			BlockMoveData( *resultDesc.dataHandle, finalFSSpec, sizeof(FSSpec) );
			AEDisposeDesc( &resultDesc );
			
			returnValue = true;
		}
		
		theErr |= NavDisposeReply(&theReply);
	}

	return returnValue;
}

