
// ISp.c

#include "ISp.h"
#include "MacOS.h"
#include "Display.h"
#include "StandardDisplay.h"
#include "Blitters.h"
#include "Alerts.h"

#include "registers.h"
#include "dispatch.h"

#include "QuickDraw.h"
#include <math.h>

const OSType kSubCreatorCode	= '0001';
const OSType kCreatorCode		= 'McTR';
ISpElementListReference gVirtualList = NULL;
ISpElementReference gVirtualElements[kNumNeeds];
bool ISpEnabled = false;

static ISpNeed gNeeds[kNumNeeds] = {
	{ "\p1) Analog", 			kIconSuiteID_Analog,	1, 0, kISpElementKind_Movement,	kISpElementLabel_None, 0, 0, 0, 0 },
	{ "\p1) Analog Center",		kIconSuiteID_ACenter,	1, 0, kISpElementKind_Button,	kISpElementLabel_None, 0, 0, 0, 0 },
	{ "\p1) C",					kIconSuiteID_C,			1, 0, kISpElementKind_DPad,		kISpElementLabel_Pad_Move, 0, 0, 0, 0 },
	{ "\p1) D",					kIconSuiteID_D,			1, 0, kISpElementKind_DPad,		kISpElementLabel_Pad_Move, 0, 0, 0, 0 },
	{ "\p1) START",				kIconSuiteID_Start,		1, 0, kISpElementKind_Button,	kISpElementLabel_None, 0, 0, 0, 0 },
	{ "\p1) A",					kIconSuiteID_ButtonA,	1, 0, kISpElementKind_Button,	kISpElementLabel_None, 0, 0, 0, 0 },
	{ "\p1) B",					kIconSuiteID_ButtonB,	1, 0, kISpElementKind_Button,	kISpElementLabel_None, 0, 0, 0, 0 },
	{ "\p1) L",					kIconSuiteID_ButtonL,	1, 0, kISpElementKind_Button,	kISpElementLabel_None, 0, 0, 0, 0 },
	{ "\p1) R",					kIconSuiteID_ButtonR,	1, 0, kISpElementKind_Button,	kISpElementLabel_None, 0, 0, 0, 0 },
	{ "\p1) Z",					kIconSuiteID_ButtonZ,	1, 0, kISpElementKind_Button,	kISpElementLabel_None, 0, 0, 0, 0 },

	{ "\pRefresh Screen",		kIconSuiteID_Button,	0, 0, kISpElementKind_Button,	kISpElementLabel_None, 0, 0, 0, 0 },		
	{ "\pCapture Snapshot",		kIconSuiteID_Button,	0, 0, kISpElementKind_Button,	kISpElementLabel_None, 0, 0, 0, 0 },		
	{ "\pPause",				kIconSuiteID_Button,	0, 0, kISpElementKind_Button,	kISpElementLabel_None, 0, 0, 0, 0 },		
	{ "\pBreak",				kIconSuiteID_Button,	0, 0, kISpElementKind_Button,	kISpElementLabel_Btn_Quit, 0, 0, 0, 0 }	
};

unsigned int InitInputSprockets()
{
	OSStatus			status;
	unsigned long		deviceCount;
	ISpDeviceReference	buffer[100];
	int					i;
	NumVersion			version;
	
	if(!(((Ptr) ISpInit != (Ptr) kUnresolvedCFragSymbolAddress)))
		return ISpNotAvail;
	
	version = ISpGetVersion();
	if ((version.majorRev <= 1) && ((version.minorAndBugRev >> 4) < 2))
	{
		return ISpTooOld;
	}
	
	for(i=0; i<kNumNeeds; i++)
		gVirtualElements[i] = nil;
	
	status = ISpElement_NewVirtualFromNeeds(kNumNeeds, gNeeds, gVirtualElements, 0);
	if (status != noErr)
		return status;
	BuildMyElementList();
	
	status = ISpInit(kNumNeeds, 
					 gNeeds, 
					 gVirtualElements, 
					 kCreatorCode, 
					 kSubCreatorCode, 
					 0, 
					 128, 
					 0);
	if (status != noErr)
		return status;
	
	// Startup the devices, that is going to be used.
	status = ISpDevices_ExtractByClass(kISpDeviceClass_Keyboard, 100, &deviceCount, buffer);
	if( status == noErr )
	{
		if(status = ISpDevices_Activate(deviceCount, buffer) != noErr)
			return status;
	}
	status = ISpDevices_ExtractByClass(kISpDeviceClass_Mouse, 100, &deviceCount, buffer);
	if( status == noErr)
	{
		if(status = ISpDevices_Activate(deviceCount, buffer) != noErr)
			return status;
	}
	
	ISpSuspend();
	
	return noErr;
}

void DeinitInputSprockets()
{
	if (ISpEnabled)
		ISpSuspend();
	ISpStop();
}

void ConfigureInputSprockets()
{
	OSStatus err = 0;
	
	err = ISpResume();
	if(err == noErr) err = ISpConfigure(nil);
	if(err == noErr) err = ISpSuspend();
}

void BuildMyElementList()
{
	gVirtualList = nil;
	ISpElementList_New(0, nil, &gVirtualList, 0);
	
	ISpElementList_AddElements(gVirtualList, kRefresh, 1, &(gVirtualElements[kRefresh]));
	ISpElementList_AddElements(gVirtualList, kCapture, 1, &(gVirtualElements[kCapture]));
	ISpElementList_AddElements(gVirtualList, kPause, 1, &(gVirtualElements[kPause]));
	ISpElementList_AddElements(gVirtualList, kBreak, 1, &(gVirtualElements[kBreak]));
}

struct coordsStruct
{
	double x;
	double y;
};

unsigned long ProcessKeys()
{
	static signed char		stick1[2];
	ISpMovementData			movementData;
	double					x, y;
//	static coordsStruct		absPos = {0.0, 0.0};
	signed short			analog_stick = 0;
	unsigned long			keystate = 0;
	unsigned long			state;
	
	ISpElement_GetComplexState(gVirtualElements[kJoy1Analog], sizeof(ISpMovementData), &movementData);
	x = movementData.xAxis;
	y = movementData.yAxis;
	switch (movementData.direction)
	{
		case kISpPadIdle:	// If no direction key is pressed, then check the mouse position.
			if (x != kISpAxisMiddle)
			{
				x -= kISpAxisMiddle;
				x = x/(kISpAxisMaximum-kISpAxisMiddle);
				x *= 240;
				x = x > 120 ? 120 : x;
//				x *= 120;
				stick1[0] = rinttol(x);
/*				x *= 30;
				absPos.x += x;
				if(absPos.x > 15.0) absPos.x = 15.0;
				else if(absPos.x < -15.0) absPos.x = -15.0;
				stick1[0] = rinttol(absPos.x)*8;*/
			}
				
			if (y != kISpAxisMiddle)
			{
				y -= kISpAxisMiddle;
				y = y/(kISpAxisMaximum-kISpAxisMiddle);
				y *= 240;
				x = y > 120 ? 120 : y;
//				x *= 120;
				stick1[1] = rinttol(y);
/*				y *= 30;
				absPos.y += y;
				if(absPos.y > 15.0) absPos.y = 15.0;
				else if(absPos.y < -15.0) absPos.y = -15.0;
				stick1[1] = rinttol(absPos.y)*8;*/
			}				
			break;
		
		case kISpPadDownLeft:
			if(stick1[1] > -120)
				stick1[1] -= 0x08;
		case kISpPadLeft:
			if(stick1[0] > -120)
				stick1[0] -= 0x08;
			break;
		
		case kISpPadUpLeft:
			if(stick1[0] > -120)
				stick1[0] -= 0x08;				
		case kISpPadUp:
			if(stick1[1] < 120)
				stick1[1] += 0x08;
			break;
		
		case kISpPadUpRight:
			if(stick1[1] < 120)
				stick1[1] += 0x08;
		case kISpPadRight:
			if(stick1[0] < 120)
				stick1[0] += 0x08;
			break;
		
		case kISpPadDownRight:
			if(stick1[0] < 120)
				stick1[0] += 0x08;
		case kISpPadDown:
			if(stick1[1] > -120)
				stick1[1] -= 0x08;
			break;
	}
	ISpElement_GetSimpleState(gVirtualElements[kKey1AnalogCenter], &state);
	if (state == kISpButtonDown) {stick1[0] = stick1[1] = 0; /*absPos.x = absPos.y = 0.0;*/}
	
	analog_stick += (stick1[1] +(stick1[0] << 8));
	
	keystate = 0; // Reset keys
	
	ISpElement_GetSimpleState(gVirtualElements[kKey1C], &state);
	switch (state)
	{
		case kISpPadDownLeft:
			keystate |= 0x0006;
			break;
		case kISpPadLeft:
			keystate |= 0x0002;
			break;
		case kISpPadUpLeft:
			keystate |= 0x000A;
			break;
		case kISpPadUp:
			keystate |= 0x0008;
			break;
		case kISpPadUpRight:
			keystate |= 0x0009;
			break;
		case kISpPadRight:
			keystate |= 0x0001;
			break;
		case kISpPadDownRight:
			keystate |= 0x0005;
			break;
		case kISpPadDown:
			keystate |= 0x0004;
			break;
	}
	ISpElement_GetSimpleState(gVirtualElements[kKey1D], &state);
	switch (state)
	{
		case kISpPadDownLeft:
			keystate |= 0x0600;
			break;
		case kISpPadLeft:
			keystate |= 0x0200;
			break;
		case kISpPadUpLeft:
			keystate |= 0x0A00;
			break;
		case kISpPadUp:
			keystate |= 0x0800;
			break;
		case kISpPadUpRight:
			keystate |= 0x0900;
			break;
		case kISpPadRight:
			keystate |= 0x0100;
			break;
		case kISpPadDownRight:
			keystate |= 0x0500;
			break;
		case kISpPadDown:
			keystate |= 0x0400;
			break;
	}
	ISpElement_GetSimpleState(gVirtualElements[kKey1Start], &state);
	if (state == kISpButtonDown) keystate |= 0x1000UL;
	ISpElement_GetSimpleState(gVirtualElements[kKey1A], &state);
	if (state == kISpButtonDown) keystate |= 0x8000UL;
	ISpElement_GetSimpleState(gVirtualElements[kKey1B], &state);
	if (state == kISpButtonDown) keystate |= 0x4000UL;
	ISpElement_GetSimpleState(gVirtualElements[kKey1L], &state);
	if (state == kISpButtonDown) keystate |= 0x0020UL;
	ISpElement_GetSimpleState(gVirtualElements[kKey1R], &state);
	if (state == kISpButtonDown) keystate |= 0x0010UL;
	ISpElement_GetSimpleState(gVirtualElements[kKey1Z], &state);
	if (state == kISpButtonDown) keystate |= 0x2000UL;
	
	return (unsigned long)(keystate << 16) | (unsigned long)(unsigned short)analog_stick;
}

void ISpProcessSpecialKeys()
{
	Boolean wasEvent;
	ISpElementEvent theEvent;
	while(1)
	{
		ISpElementList_GetNextEvent(gVirtualList, sizeof(ISpElementEvent), &theEvent, &wasEvent);
		if (!wasEvent)
		{
			return;
		}
		switch(theEvent.refCon)
		{
			case kRefresh:
				if (theEvent.data == kISpButtonDown)
				{
					dispatch(DISPLAY_REFRESH, NULL);
				}
				break;
			case kCapture:
				if (theEvent.data == kISpButtonDown)
				{
					SaveScreen(nil);
				}
				break;
			case kPause:
				if (theEvent.data == kISpButtonDown)
				{
					GWorldPtr	tempWorld;
					PicHandle	thePicture;
					GrafPtr		oldPort;
					GDHandle	oldGD;
					GWorldPtr	oldGW;
					unsigned long	state;
					bool		pausing = true;
					long		theWidth, theHeight;
					
					thePicture = (PicHandle)GetResource('PICT', 129);
					NewGWorld(	&tempWorld, 16, &(*thePicture)->picFrame, nil, nil, 0L );
					LockPixels( tempWorld->portPixMap );

					GetPort( &oldPort );
					GetGWorld( &oldGW, &oldGD );
					SetGWorld( tempWorld, nil );

					DrawPicture(thePicture, &(*thePicture)->picFrame);

					ClosePicture( );
					SetGWorld( oldGW, oldGD );
					SetPort( oldPort );
					ReleaseResource( (Handle)thePicture );
					
					theWidth    = tempWorld->portRect.right-tempWorld->portRect.left;
					theHeight   = tempWorld->portRect.bottom-tempWorld->portRect.top;
					BlitPause(	(unsigned short *)(**tempWorld->portPixMap).baseAddr, 
								theWidth, theHeight, (**tempWorld->portPixMap).rowBytes & 0x3fff);
					
					DisposeGWorld( tempWorld );
					
					while(pausing)
					{
						ISpElementList_GetNextEvent(gVirtualList, sizeof(ISpElementEvent), &theEvent, &wasEvent);
						if (wasEvent && (theEvent.refCon == kPause) && theEvent.data == kISpButtonDown)
						{
							pausing = false;
						}
					}
					
					if (!UnBlitPause(theWidth, theHeight))
						dispatch(DISPLAY_REFRESH, NULL);
				}
				break;
			case kBreak:
				if (theEvent.data == kISpButtonDown)
				{
					reg.halt = 2;
					return;
				}
				break;
		}
	}
}