
// MacOS.c

#include <stdlib.h>
#include <string.h>

#include <Dialogs.h>
#include <Events.h>
#include <Gestalt.h>
#include <Appearance.h>
#include <DrawSprocket.h>
#include "Navigation.h"

#include "MacOS.h"
#include "Menu.h"
#include "StandardDisplay.h"
#include "UtilityFunctions.h"
#include "GameLoad.h"
#include "AE.h"
#include "ISp.h"
#include "Alerts.h"

#include "CPU.h"
#include "registers.h"
#include "rsp_registers.h"
#include "dispatch.h"
#include "display.h"
#include "config.h"
#include "parser_extern.h"

//#define TEST_DEBUG

#ifdef TEST_DEBUG
#include <stdio.h>
	int		begun_test = false;
	FILE	*test_file;
#endif

// Uncomment the instuction below, 
// add MSL SIOUX.PPC.Lib to the project file 
// and do not compile console.stubs.c
// to enable the text output window

#ifdef DEBUG
	#include "debug.h"
	#include <SIOUX.h>
#endif

void DoMouseWindowMove();

static bool gIgnoreAll;
static uBYTE gIgnoreListSize;
static long gIgnoreList[256];
unsigned long gIgnoreAddr;

bool hasRAVE;
int gRunning = false;
int gFinished = false;
int gRomOpened = false;
int gFullScreen = false;

static short		systime=0;					//Hog the processor

char gHaltInstrStr[256];

void main()
{
#if __profile__
	if (ProfilerInit(collectSummary, bestTimeBase, 200, 20)!=noErr)
		ExitToShell();
#endif

	Initialize();

	MainLoop();
	
	Finish();

#if __profile__
	ProfilerDump("\pFractal.prof");
	ProfilerTerm();
#endif
}

void Initialize()
{
	long returnVal;
	int temp;
	unsigned int err;
	
#ifdef __SIOUX__
	SIOUXSettings.autocloseonquit = true;
	SIOUXSettings.asktosaveonclose = 
	SIOUXSettings.setupmenus = 
	SIOUXSettings.standalone = 
	SIOUXSettings.initializeTB = false;
#endif	
	InitGraf(&qd.thePort);
	InitFonts();
	FlushEvents(everyEvent, 0L);
	InitWindows();
	InitMenus();
	TEInit();
	InitDialogs(nil);
	InitCursor();
	
	MaxApplZone();
	
	// My Initializations
	if(InitDisplay() < 0)
		ExitToShell();

    prefs.load_what = N64_NOTHING;
	prefs.plugged_controller = CONTR_1;	/* Only 1 controller */
	prefs.symbols = NO_SYMBOLS;
#ifdef AUDIO
	prefs.sound = NORMAL_SOUND;
	prefs.sound_bits = 16;
	prefs.sound_freq = 44100;
	prefs.sound_stereo = true;
#endif
	prefs.eeprom = false;	/* Eeprom defaults to false */
	prefs.r22 = 0x3f;		/* default r22 init value */
#ifdef FRAMESKIP
	prefs.frames_to_skip = 0;
#endif
	
#ifdef DEBUG
//	prefs.debug |= DBG_CPU_INSTR;
	prefs.debug |= DBG_RDP;
	prefs.debug |= DBG_RSP_INSTR;
#endif

	if(dispatch(MEMORY_ALLOC, NULL) < 0) 
		ExitToShell();

	MenuBarInit();
	err = InitInputSprockets();
	if(err != noErr)
	{
		InputSprocketStopAlert(err);
		ExitToShell();
	}
	
	CheckItem(GetMenuHandle(mOption), iEeprom, prefs.eeprom);
	DisableItem(GetMenuHandle(mN64), 0);
	
	InitAppleEvents();
}

void Finish()
{
	CloseRom();
	dispatch(DISPLAY_CLOSE, nil);
	dispatch(MEMORY_FREE, nil);
	DeinitInputSprockets();
	DSpShutdown();
}

inline void MainLoop()
{
	while(!gFinished)
	{
#ifdef DEBUG
        if(prefs.load_what == N64_IMAGE)
                n64_cpu_debugger();
        else
        {
                rsp_reg.halt = 0;
                n64_rsp_debugger();
        }
#else
		while(gRunning)
		{
#ifdef HALT
			cpu_step();
			if (reg.halt == 1)
			{
				short button;
				
				if (!gIgnoreAll)
				{
					PauseDisplay();
					FlushEvents(everyEvent, 0L);
					
					if (gHaltInstrStr[0] != 0)
						button = HaltWarning(KnownHalt);
					else
						button = HaltWarning(UnknownHalt);

					gHaltInstrStr[0] = 0;
					reg.halt = 0;
					
					if (button != 1)
					{
						if (button == 3)
							gIgnoreAll = true;
						
						gIgnoreList[gIgnoreListSize++] = gIgnoreAddr;

						OpenVisual();
					}
				}
				else
					reg.halt = 0;
			}
			else
			{
				PauseDisplay();
				FlushEvents(everyEvent, 0L);
			}
#else
			cpu_step();
#endif
		}
#endif		
		HandleEvents();
	}
}

void HandleEvents()
{
	EventRecord theEvent;
	
	WaitNextEvent(everyEvent, &theEvent, systime, nil);
	
#ifdef __SIOUX__	
	SIOUXHandleOneEvent(&theEvent);
#endif
	
	switch (theEvent.what)
	{
		case mouseDown:
			HandleMouseDown(&theEvent);
			break;
		case keyDown:
			if( theEvent.modifiers & cmdKey )
			{
				HandleMenus( MenuKey(theEvent.message & charCodeMask) );
			}
			break;
		
		case updateEvt:
			//Windowed code is not quite ready yet. You can try uncommenting and swapping DSp files for regular.
			
//			BeginUpdate(theWindow);		//It'll be updated in less than a second anyway (hopefully),
//			EndUpdate(theWindow);		//so just ignore it and have the system stop bothering us.
			break;
			
		case kHighLevelEvent:
			AEProcessAppleEvent(&theEvent);
			break;
		
		default:
			break;
	}
}

void HandleMouseDown( EventRecord *eventPtr )
{
	WindowPtr		whichWindow;
	short			thePart;
	
	thePart = FindWindow( eventPtr->where, &whichWindow );
	switch ( thePart )
	{
		case inMenuBar:
			HandleMenus(MenuSelect(eventPtr->where));
			break;
		case inDrag:
			DragWindow(whichWindow, eventPtr->where, &qd.screenBits.bounds);
		default:
			break;
	}
}

// Empty function that isn't used in macos port
void CatchSignals()
{}


void CpuResetMac()
{
	gIgnoreAll = false;
	gIgnoreListSize = 0;
	ResetGraphics();
}	

bool IsInIgnoreList(unsigned long addr)
{
	int i;
	if (gIgnoreAll) return true;
	
	for (i = 0; i < gIgnoreListSize; i++)
		if (addr == gIgnoreList[i])
			return true;
	
	return false;
}