
// Menu.c

#include <Menus.h>
#include "Navigation.h"
#include "Menu.h"
#include "MacOS.h"
#include "StandardDisplay.h"
#include "UtilityFunctions.h"
#include "GameLoad.h"
#include "ISp.h"
#include "Alerts.h"

#include <stdlib.h>

#include "Memory.h"
#include "file_loader.h"
#include "parser_extern.h"
#include "CPU.h"
#include "DISPATCH.H"
#include "Display.h"

void MenuBarInit()
{
	Handle		menuBar;
	MenuHandle	menu;
	
	// Install Menus
	menuBar = GetNewMBar(128);
	if(menuBar == nil)
		ExitToShell();
	
	SetMenuBar(menuBar);
	DrawMenuBar();

	// Install Apple Menu
	menu = GetMenuHandle( mApple );
	AppendResMenu( menu, 'DRVR' );
}

extern int framesToSkip;
extern bool speedLimit;
extern bool fpsCounter;

void HandleMenus(long menuChoice)
{
	short menu, item;
	
	if ( menuChoice != 0 )
	{
		menu = HiWord( menuChoice );
		item = LoWord( menuChoice );

		switch ( menu )
		{
			case mApple:
				HandleAppleChoice( item );
				break;
				
			case mFile:
				HandleFileChoice( item );
				break;

			case mN64:
				HandleN64Choice( item );
				break;

			case mOption: // Prefs
				switch( item )
				{
					case 1:
						speedLimit = !speedLimit;
						CheckItem(GetMenuHandle(menu), item, speedLimit);
						break;
					case 2:
						fpsCounter = !fpsCounter;
						CheckItem(GetMenuHandle(menu), item, fpsCounter);
						break;

					case 4:
					case 5:
					case 6:
					case 7:
					case 8:
					case 9:
#ifdef FRAMESKIP
						CheckItem(GetMenuHandle(menu), prefs.frames_to_skip+4, false);
						prefs.frames_to_skip = item - 4;
						skipframe = 0;
						CheckItem(GetMenuHandle(menu), item, true);
#endif
						break;
					
					case iEeprom:
						prefs.eeprom = !prefs.eeprom;
						CheckItem(GetMenuHandle(menu), item, prefs.eeprom);
						break;
				}
				break;
		}
		HiliteMenu( 0 );
	}
}

void HandleAppleChoice( short item )
{
	MenuHandle	appleMenu;
	Str255		accName;
	short		accNumber;
	
	switch ( item )
	{
		case 1:		// About
			ShowAboutMeDialog();
			break;
		default:
			appleMenu = GetMenuHandle( mApple );
			GetMenuItemText( appleMenu, item, accName );
			accNumber = OpenDeskAcc( accName );
			break;
	}
}

void HandleFileChoice( short item )
{
	switch ( item )
	{
		case iOpen:
			OpenRomImage(nil);
			break;
		case iPrefs:
			ConfigureInputSprockets();
			break;
		case iQuit:
			gFinished = true;
			break;
	}
}

void HandleN64Choice( short item )
{
	if (gRomOpened)
	{
		switch ( item )
		{
			case iReset:
				OpenVisual();
				dispatch(EMU_RESET, NULL);
				break;
			case iRun:
				OpenVisual();
				break;
		}
	}
}
