
// UtilityFunctions.c

#include "UtilityFunctions.h"
#include <string.h>
#include <stdlib.h>

/* Converts a Pascal string to C-style. */

void PascalToC( const unsigned char *pS, char *cS )
{
	short size = *pS++;
	
	while( size-- ) *cS++ = *pS++;
	*cS = 0;
}

/* Converts a C string to Pascal style. */

void CToPascal( const char *cS, unsigned char *pS )
{
	unsigned char *pP = ++pS;
	
	while( *cS ) *pP++ = *cS++;
	*(pS-1) = pP - pS;
}

/* Copies a Pascal-style string. */

StringPtr CopyPString(StringPtr str1, StringPtr str2)
{
	unsigned char	*cp = str1;
	unsigned		len = *(unsigned char *)str2;

	do *cp++ = *str2++;
	while( len-- );
	
	return str1;
}

/* Concatenates two Pascal-style strings. */

StringPtr ConcatPString(StringPtr str1, StringPtr str2)
{
	unsigned char	*cp = str1;
	unsigned		len1 = *(unsigned char *)str1,
					len2 = *(unsigned char *)str2;
	if ((len1 + len2) >= 255) return 0;
	
	cp += len1 + 1;						
	*(unsigned char *)str1 += len2;		
	++str2;								
	while( len2-- ) *cp++ = *str2++;
	
	return str1;
}


short FullPathname( FSSpec *in, Str255 out )
{
	Str255		dirName;
	DirInfo		dirInfo;
	OSErr		err;

	out[0] = dirName[0] = 0;

	dirInfo.ioNamePtr = dirName;
	dirInfo.ioDrParID = in->parID;

	do
	{
		dirInfo.ioVRefNum = in->vRefNum;
		dirInfo.ioFDirIndex = -1; 				// -1 means use ioDrDirID
		dirInfo.ioDrDirID = dirInfo.ioDrParID;

		err = PBGetCatInfoSync( (CInfoPBPtr) &dirInfo );

		if( err == noErr )
		{
			dirName[ ++dirName[0] ] = ':';

			if( dirName[0] + out[0] > 255 )
			{
				err = bdNamErr;
			}
			else
			{
				ConcatPString( dirName, out );
				CopyPString( out, dirName );
			}
		}
	} 
	while( (dirInfo.ioDrDirID != fsRtDirID) && (err == noErr) );
	
	if( dirName[0] + in->name[0] > 255 )
	{
		err = bdNamErr;
	}
	else
	{
		ConcatPString( out, in->name );
	}
	
	return err;
}

void ChangeTypeCreator(FSSpec *theFile)
{
	OSErr	error = noErr;
	FInfo	fileInfo;
	
	error = FSpGetFInfo(theFile, &fileInfo);
	if (error != noErr)
		return;
	
	if (fileInfo.fdType != 'APPL')
	{
		fileInfo.fdType = 'CART';
		fileInfo.fdCreator = 'McTR';
		
		FSpSetFInfo(theFile, &fileInfo);
	}
}

char *strdup(const char *str)	// Somehow, MSL managed to not have strdup for MacOS (only). I hope this is what it does.
{
	char *returnstr;
	returnstr=(char *)malloc(strlen(str));
	strcpy(returnstr, str);
	return returnstr;
}