/**
 * TrueReality - output/audio/amigados.c
 * Copyright (C) 1998, 1999 Niki W. Waibel
 *
 * This program is free software; you can redistribute it and/
 * or modify it under the terms of the GNU General Public Li-
 * cence as published by the Free Software Foundation; either
 * version 2 of the Licence, or any later version.
 *
 * This program is distributed in the hope that it will be use-
 * ful, but WITHOUT ANY WARRANTY; without even the implied war-
 * ranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public Licence for more details.
 *
 * You should have received a copy of the GNU General Public
 * Licence along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 * USA.
 *
 * Information about me (the author):
 *   Niki W. Waibel, Reichenau 20, 6890 Lustenau, Austria - EUROPE
 *   niki.waibel@gmx.net
**/





/**
 *
 * All of this stuff comes from AmiDog - great work (niki)
 *
**/





/* DOWNSAMPLE: use 0, 1, 2, 3, etc for /1, /2, /4, /8 etc */

#define DOWN_SAMPLE 0

#include <unistd.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <stdio.h>

#include <math.h>
#include "audio.h"

#include <type_sizes.h>
#include <memory_extern.h>
#include <parser_extern.h>
#include "../../N64/interrupts.h"   

FILE *ahifile;
static int audio_is_open;
static int audiofd;


void open_ahi(long channels, long frequency)
{
  char buffer[256];

  sprintf(buffer,"AUDIO:B 16 C %ld F %ld",channels,frequency);
  ahifile = fopen(buffer,"w");

  if (ahifile==NULL)
  {
    printf("AUDIO: could not be opened !!!\n");
    prefs.sound=NO_SOUND;
  }
  else
    audio_is_open=1;
}


void close_ahi()
{
  fclose(ahifile);

  audio_is_open=0;
}


int OpenAudio()
{
    /* open audio device - exit if NO_SOUND and later if FILE_SOUND */

        switch(prefs.sound)
        {
            case NO_SOUND:
                puts("No Sound ... OK\n-------------------------------------------------------------------------------");
                fflush(stdout);

                return(0);   /* Exit here. There is nothing else to do! */

            case NORMAL_SOUND:
                fputs("AUDIO: will be used ...", stdout);
                  audio_is_open=0;
                break;

            case FILE_SOUND:
                printf("Opening audio file (%s) (ideas by Niki W. Waibel) ...", prefs.sound_file);
                break;

        } /* switch(prefs.sound) */

        fflush(stdout);

        if(prefs.sound == FILE_SOUND)
        {
            audiofd = open(prefs.sound_file, O_WRONLY|O_CREAT|O_TRUNC, S_IRUSR|S_IWUSR);

            if(audiofd < 0)
            {
                puts(" FAIL\n"
                     "Turn off audio!!!"); fflush(stdout);
                prefs.sound = NO_SOUND;
                return(-1);
            };
        };

        if(prefs.sound == FILE_SOUND)
        {
                puts(" OK\n"
                     "-------------------------------------------------------------------------------");
                fflush(stdout);

                return(0);   /* Exit here. There is nothing else to do! */

        } /* if(prefs.sound == FILE_SOUND) */

        puts(" OK\n-------------------------------------------------------------------------------"); fflush(stdout);

        return(0);

} /* int OpenAudio() */


int CloseAudio()
{                                                
        switch(prefs.sound)
        {
            case NO_SOUND:
                fputs("No Sound ...", stdout);
                fflush(stdout);
                break;

            case NORMAL_SOUND:
                fputs("Closing audio device ...", stdout);
                fflush(stdout);
                if (audio_is_open==1)
                  close_ahi();
                break;

            case FILE_SOUND:
                printf("Closing audio file (%s) (ideas by Niki W. Waibel) ...", prefs.sound_file);
                fflush(stdout);
                close(audiofd);
                break;
        
        } /* switch(prefs.sound) */


    puts(" OK\n-------------------------------------------------------------------------------"); fflush(stdout);

    return(0);

} /* int CloseAudio() */


void UpdateAudio()
{
    static uWORD     audiofreq;
    static char     *audioptr;
    static uWORD     audiolen;

    static uWORD     frequency; /* Added by AmiDog */

    if(prefs.sound == NO_SOUND)
        return;

    if(audiofreq != (( (93750000 >> 1) / (((uWORD *)mem.ai_reg)[4] & 0x3ffff) + 1) >> DOWN_SAMPLE))
    {
        audiofreq = ((93750000 >> 1) / (((uWORD *) mem.ai_reg)[4] & 0x3ffff) + 1) >> DOWN_SAMPLE;
        if (audio_is_open==1)
          close_ahi();
    }

    if(audioptr != (char *)(mem.rd_ram + (((uWORD *)mem.ai_reg)[0] & 0xfffff8)))
    {
        audioptr = (((uWORD *)mem.ai_reg)[0] & 0xfffff8) + mem.rd_ram;
    }

    if(audiolen != (((uWORD *)mem.ai_reg)[1] & 0x3fff8))
    {
        audiolen = ((uWORD *)mem.ai_reg)[1] & 0x3fff8;
    }

    if(prefs.sound == FILE_SOUND)
        write(audiofd, audioptr, audiolen);

    if(prefs.sound == NORMAL_SOUND)
    {
        if (audio_is_open==0)
        {
            frequency=(((uWORD *)mem.ai_reg)[4] & 0x3ffff)+1; /* True frequency!? /AmiDog */

            open_ahi(2, frequency); /* Always stereo (2 channels) */

            printf("AmigaOS: audio frequency change to: %ld Hz\n", frequency);
        }

        if (audio_is_open==1)
            fwrite(audioptr,2,audiolen,ahifile);
    }
} /* void UpdateAudio() */


int ReadyAudio()
{
    puts("NO DMA or LEN == 0");
    return(AUDIO_DMA_READY);
} /* int ReadyAudio() */

