/**
 * TrueReality - output/audio/msdos.c
 * Copyright (C) 1998 Niki W. Waibel
 *
 * This program is free software; you can redistribute it and/
 * or modify it under the terms of the GNU General Public Li-
 * cence as published by the Free Software Foundation; either
 * version 2 of the Licence, or any later version.
 *
 * This program is distributed in the hope that it will be use-
 * ful, but WITHOUT ANY WARRANTY; without even the implied war-
 * ranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public Licence for more details.
 *
 * You should have received a copy of the GNU General Public
 * Licence along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 * USA.
 *
 * Information about me (the author):
 *   Niki W. Waibel, Reichenau 20, 6890 Lustenau, Austria - EUROPE
 *   niki.waibel@gmx.net
**/





/**
 *
 * This msdos port comes from:
 * "Boca Juniors (Diego Rozensztejn)" <boca_jrs@hotmail.com>
 *
 * DJGPP and Allegro are needed!
 *
**/





/**
 *
 * This file is based on bsd_and_linux.c in this directory
 *
**/





/* DOWNSAMPLE: use 0, 1, 2, 3, etc for /1, /2, /4, /8 etc */

#define DOWN_SAMPLE 0

#define NUM_OF_VOICES 16




#include <unistd.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <stdio.h>

#include <math.h>   

#include <allegro.h>
#include "audio.h"
#include "msdos.h"

#include <type_sizes.h>
#include <memory_extern.h>
#include <parser_extern.h>
#include "../../N64/interrupts.h"   






static int      audiofd;
static int      buffer_location = 0;
static int      BUFFER_SIZE = 4096;

#define         INIT_BUFFER_SIZE  128

SAMPLE          *audiosample[NUM_OF_VOICES];        // Pointers to audio samples
int             voice_allocated[NUM_OF_VOICES];     // -1 not allocated else voice number


int OpenAudio()
{
        int             status;
        int             i;

    /* open audio device - exit if NO_SOUND and later if FILE_SOUND */

        switch(prefs.sound)
        {
            case NO_SOUND:
                puts("No Sound ... OK\n"
                     "-------------------------------------------------------------------------------");
                fflush(stdout);

                return(0);   /* Exit here. There is nothing else to do! */

            case NORMAL_SOUND:
                fputs("Opening DOS audio device ...", stdout);
                break;
        
            case FILE_SOUND:
                printf("Opening audio file (%s) (ideas by Niki W. Waibel) ...", prefs.sound_file);
                break;
        
        } /* switch(prefs.sound) */

        fflush(stdout);



        if(prefs.sound == NORMAL_SOUND)
        {                                                                            
            reserve_voices(NUM_OF_VOICES,0);

            for(i=0; i < NUM_OF_VOICES; i++)
            {   
                audiosample[i] = NULL;
                voice_allocated[i] = -1;
            };

            status = install_sound(DIGI_AUTODETECT, MIDI_NONE, "");
            if(status != 0)
            {
                puts(" FAIL\n"
                     "Turn off audio!!!"); fflush(stdout);
                prefs.sound = NO_SOUND;
                return(-1);
            };
        };

        if(prefs.sound == FILE_SOUND)
        {
            audiofd = open(prefs.sound_file, O_WRONLY | O_CREAT | O_TRUNC | O_BINARY, S_IRUSR | S_IWUSR);

            if(audiofd < 0)
            {
                puts(" FAIL\n"
                     "Turn off audio!!!"); fflush(stdout);
                prefs.sound = NO_SOUND;
                return(-1);
            };
        };


        if(prefs.sound == FILE_SOUND)
        {
                puts(" OK\n"
                     "-------------------------------------------------------------------------------");
                fflush(stdout);

                return(0);   /* Exit here. There is nothing else to do! */

        } /* if(prefs.sound == FILE_SOUND) */




    
        buffer_location = 0;



        puts(" OK\n"
             "-------------------------------------------------------------------------------"); fflush(stdout);

        return(0);

} /* int OpenAudio() */





int CloseAudio()
{                                                
        int i;

        switch(prefs.sound)
        {
            case NO_SOUND:
                fputs("No Sound ...", stdout); fflush(stdout);
                break;

            case NORMAL_SOUND:
                fputs("Closing audio device ...", stdout);
                fflush(stdout);

                for(i=0; i < NUM_OF_VOICES; i++)
                {
                    if(voice_allocated[i] != -1)
                    {
                        deallocate_voice(voice_allocated[i]);   
                        destroy_sample(audiosample[i]);
                    };
                };
                remove_sound();
                break;
        
            case FILE_SOUND:
                printf("Closing audio file (%s) (ideas by Niki W. Waibel) ...", prefs.sound_file);
                fflush(stdout);
                close(audiofd);
                break;
        
        } /* switch(prefs.sound) */


    puts(" OK\n"  
             "-------------------------------------------------------------------------------"); fflush(stdout);


         return(0);

} /* int CloseAudio() */





void UpdateAudio()
{
        static uWORD     audiofreq;
        static char     *audioptr;
        static uWORD     audiolen;

        SAMPLE          *sampleptr;
        int             i;


        if(prefs.sound == NO_SOUND)
                return;


    if(audiofreq != (( (93750000 >> 1) / (((uWORD *)mem.ai_reg)[4] & 0x3ffff) + 1) >> DOWN_SAMPLE))
    {
        audiofreq = ((93750000 >> 1) / (((uWORD *) mem.ai_reg)[4] & 0x3ffff) + 1) >> DOWN_SAMPLE;
    }



    if(audioptr != (char *)(mem.rd_ram + (((uWORD *)mem.ai_reg)[0] & 0xfffff8)))
    {
        audioptr = (((uWORD *)mem.ai_reg)[0] & 0xfffff8) + mem.rd_ram;
    }



        if(audiolen != (((uWORD *)mem.ai_reg)[1] & 0x3fff8))
        {
        audiolen = ((uWORD *)mem.ai_reg)[1] & 0x3fff8;
        }



        /* BUG: ??? */
        /*
        ((uWORD *)mem.ai_reg)[3] = 0xc0000001;
        */

#ifdef DEBUG
        if(prefs.debug & DBG_AUDIO)
        {
                printf("AUDIO: Frequency: %ld\n", audiofreq);
                printf("AUDIO: Address:   %08lx\n", ((uWORD *)mem.ai_reg)[0]);
                printf("AUDIO: Length:    %ld\n", audiolen);
        }
#endif


        if(prefs.sound == FILE_SOUND)
            write(audiofd, audioptr, audiolen);


        if(prefs.sound == NORMAL_SOUND)
        {
            
            /* Destroy allocated samples not playing */

            for(i=0; i < NUM_OF_VOICES; i++)
            {
                if(voice_allocated[i] != -1)
                {
                    if(voice_get_position(voice_allocated[i]) == -1) // Stopped playing
                    {
                        release_voice(voice_allocated[i]);  
                        voice_allocated[i] = -1;    // Mark voice as available
                        destroy_sample(audiosample[i]);
                        audiosample[i] = NULL;
                    };
                };
            };

                
            /* Create sample and allocate first voice available */
    
            sampleptr = create_sample(INIT_AUDIO_BITS, INIT_AUDIO_STEREO,
                                        audiofreq, audiolen);
  
            if(sampleptr == NULL)
            {
                puts(" FAIL\n"
                     "Turn off audio!!!"); fflush(stdout);
                prefs.sound = NO_SOUND;
            }
            

            // Copy audio data
            memcpy(sampleptr->data, audioptr, audiolen);
                        
            for(i=0; i < NUM_OF_VOICES; i++)
            {
                if(voice_allocated[i] == -1)
                {
                    voice_allocated[i] = allocate_voice(sampleptr); // Mark voice as in use
                    audiosample[i] = sampleptr;
                    voice_start(voice_allocated[i]);    // Start playing
                    break;
                };
            };
        
    
        };


    /* not used yet */
        buffer_location += audiolen;
 
} /* void UpdateAudio() */





int ReadyAudio()
{
        {
                puts("NO DMA or LEN == 0");
                return(AUDIO_DMA_READY);
        }

} /* int ReadyAudio() */





