/**
 * TrueReality - output/display/amigados/amigados2D.c
 * Copyright (C) 1998, 1999 Mathias "AmiDog" Roslund
 *
 * This program is free software; you can redistribute it and/
 * or modify it under the terms of the GNU General Public Li-
 * cence as published by the Free Software Foundation; either
 * version 2 of the Licence, or any later version.
 *
 * This program is distributed in the hope that it will be use-
 * ful, but WITHOUT ANY WARRANTY; without even the implied war-
 * ranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public Licence for more details.
 *
 * You should have received a copy of the GNU General Public
 * Licence along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 * USA.
 *
 * Information about me (the author):
 *   Niki W. Waibel, Reichenau 20, 6890 Lustenau, Austria - EUROPE
 *   niki.waibel@gmx.net
**/





/**
 *
 * All of this stuff comes from AmiDog - great work (niki)
 *
**/





// #define AMIGA_UNIMP
// #define AMIGA_DEBUG

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <signal.h>
#include <unistd.h>
#include <math.h>

#include <exec/types.h>
#include <exec/nodes.h>
#include <exec/lists.h>
#include <exec/memory.h>
#include <powerup/ppcproto/intuition.h>
#include <powerup/ppclib/interface.h>
#include <powerup/ppclib/time.h>
#include <powerup/gcclib/powerup_protos.h>
#include <powerup/ppcproto/exec.h>
#include <powerup/ppcinline/graphics.h>
#include <powerup/ppcproto/dos.h>

#include <cybergraphx/cybergraphics.h>
#include <powerup/ppcinline/cybergraphics.h>

#include <config.h>
#include <parser_extern.h>
#include <type_sizes.h>
#include <memory_extern.h>
#include <dispatch.h>

#include "../display.h"
#include "../sw_renderer.h"
#include "../../../N64/registers.h"
#include "../../../N64/draw2d_sw.h"
#include "amigados2D.h"
#include "c2p.c"

#ifdef DEBUG
  #include <debug.h>
#endif

#ifdef FPS
  #include <sys/time.h>
  #include <unistd.h>
#endif

static int valid_back_pix = 0;
unsigned short drawcolor=0x0000;

/* Original Variables */
static uWORD width=0;
static uWORD height=0;
static int WIDTH=640;
static int HEIGHT=480;
#undef DEPTH
static int DEPTH;
static uWORD flags=0;

/* Local functions */
static int InitDisplay();
static int DestroyDisplay();
static void PrintText(char *text, int x, int y);
static void PutImage(int x, int y, int w, int h);
#ifdef DEBUG_DISPLAY
  static void PrintDisplayMode(uWORD flags);
#endif

/* Amiga specific variables */
int _width_, _height_, _depth_;
struct IntuitionBase *IntuitionBase;
struct GfxBase *GfxBase;
struct Library *CyberGfxBase;
struct BitMap agabitmap;
struct Screen *DispScreen;
struct Window *DispWindow;
amigaprefstype amigaprefs;

unsigned char *img_data, *ham_buf, *back_pix;

/* CGFX variables */
unsigned long pixfmt;
unsigned char *baseaddress;
int bpr, bpp;

int load_prefs()
{
  unsigned char type[]={AGA, AGA, AGAHAM, AGAHAMPLUS, AGAHAM, AGAHAMPLUS};
  unsigned char depth[]={6, 8, 6, 6, 8, 8};
  FILE *file=NULL;
  char str[50];

  /* Open file for reading */
  file=fopen("TrueReality.cfg", "r");

  if (file!=NULL)
  {
    while (fscanf(file, "%s", str)!=EOF)
    {
      /* General */
      if (strcmp(str, "#DEF_GRAPHICS")==0)
        fscanf(file, "%d", &amigaprefs.def_graphics);
      if (strcmp(str, "#DEF_FPS")==0)
        fscanf(file, "%d", &amigaprefs.def_fps);

      /* Screenmode IDs */
      if (strcmp(str, "#DEF_SM_LO")==0)
        fscanf(file, "%lx", &amigaprefs.def_sm_lo);
      if (strcmp(str, "#DEF_SM_HI")==0)
        fscanf(file, "%lx", &amigaprefs.def_sm_hi);
    }

    /* Close the file */
    fclose(file);
    file=NULL;

    amigaprefs.def_type=type[amigaprefs.def_graphics];
    amigaprefs.def_depth=depth[amigaprefs.def_graphics];

    /* Use CyberGraphics? */
    if(CyberGfxBase != NULL)
      if(IsCyberModeID(amigaprefs.def_sm_lo))
        if(IsCyberModeID(amigaprefs.def_sm_hi))
        {
          amigaprefs.def_type=GFX; /* CGFX */
          amigaprefs.def_depth=8;
        }
        else
        {
          printf("\nYou must not mix AGA and CGFX screenmodes!\n");
          return -1;
        }

    /* Depth */
    _depth_=amigaprefs.def_depth;

    /* Force PAL for HAM modes */
    if (amigaprefs.def_type==AGAHAM)
    {
      amigaprefs.def_sm_lo=0x00029800;
      amigaprefs.def_sm_hi=0x00029824;
    }
    else if (amigaprefs.def_type==AGAHAMPLUS)
    {
      amigaprefs.def_sm_lo=0x00029820;
      amigaprefs.def_sm_hi=0x00029824;
    }

    return 0;
  }
  else
    return -1;
}

void set_color(int number, int red, int green, int blue)
{
  unsigned long lred=0, lgreen=0, lblue=0;
  
  lred+=red;
  lred=lred & 0x000000ff;
  lred=lred << 24;

  lgreen+=green;
  lgreen=lgreen & 0x000000ff;
  lgreen=lgreen << 24;
  
  lblue+=blue;
  lblue=lblue & 0x000000ff;
  lblue=lblue << 24;

  SetRGB32(&DispScreen->ViewPort, number, lred, lgreen, lblue);
}

int open_libraries()
{
  IntuitionBase = (struct IntuitionBase *) OpenLibrary("intuition.library", 39);

  if(IntuitionBase==NULL)
    return -1;

  GfxBase = (struct GfxBase *) OpenLibrary("graphics.library", 39);

  if(GfxBase==NULL)
  {
    CloseLibrary(IntuitionBase);
    return -1;
  }

  CyberGfxBase = OpenLibrary("cybergraphics.library", 40);
  return 0;
}

void close_libraries()
{
  if (CyberGfxBase!=NULL)
    CloseLibrary(CyberGfxBase);
  CloseLibrary(GfxBase);
  CloseLibrary(IntuitionBase);
}

void close_screen()
{
  CloseWindow(DispWindow);
  CloseScreen(DispScreen);
  if (amigaprefs.def_type==AGA)
    FreeRaster(agabitmap.Planes[0], _width_, _height_*_depth_);
  if (amigaprefs.def_type==AGAHAM)
    FreeRaster(agabitmap.Planes[0], _width_*2, _height_*_depth_);
  if (amigaprefs.def_type==AGAHAMPLUS)
    if (_width_==320)
      FreeRaster(agabitmap.Planes[0], _width_*4, _height_*_depth_);
    else
      FreeRaster(agabitmap.Planes[0], _width_*2, _height_*_depth_);
}

int open_screen(int width, int height)
{
  int dispid, i;
  int red, green, blue;
  unsigned char *temp, *ham_init, *ptr;

  /* Choose ID */
  if (width==320)
    dispid=amigaprefs.def_sm_lo;
  else
    dispid=amigaprefs.def_sm_hi;

  /* Init */
  _width_=width;
  _height_=height;

  /* Screendepth */
  DEPTH=8;

  if (amigaprefs.def_type==AGAHAM)
  {
    /* 32bit for HAM */
    DEPTH=32;

    /* Allocate HAM buffer */
    width*=2;
    ham_buf=(unsigned char*)malloc(width*_height_);
    memset(ham_buf, 0, width*_height_);
  }

  if (amigaprefs.def_type==AGAHAMPLUS)
  {
    /* 32bit for HAM */
    DEPTH=32;

    /* Allocate HAM buffer */
    width*=2;
    
    if (_width_==320)
      width*=2; /* True 12/18bit (4xwidth) */

    ham_buf=(unsigned char*)malloc(width*_height_);
    memset(ham_buf, 0, width*_height_);
  }

  if (amigaprefs.def_type!=GFX)
  {
    /* Init screen resources */
    temp = (unsigned char*)AllocRaster(width, _height_*_depth_);
    memset(temp, 0, width*_height_*_depth_/8);

    InitBitMap(&agabitmap, _depth_, width, _height_);

    for(i=0; i<_depth_; i++)
      agabitmap.Planes[i] = temp + i * RASSIZE(width, _height_);

    /* Open screen */
    DispScreen = OpenScreenTags(NULL,
             SA_BitMap,(int)&agabitmap,
             SA_Width,width,
             SA_Height,_height_,
             SA_Depth,_depth_,
             SA_Quiet,TRUE,
             SA_ShowTitle,FALSE,
             SA_Type,CUSTOMSCREEN,
             SA_DisplayID,dispid,
             TAG_DONE);
  }
  else if (amigaprefs.def_type==GFX)
  {
    /* Get depth */
    _depth_=GetCyberIDAttr(CYBRIDATTR_DEPTH, dispid);

    /* Open screen */
    DispScreen = OpenScreenTags(NULL,
             SA_Width,_width_,
             SA_Height,_height_,
             SA_Depth,_depth_,
             SA_Quiet,TRUE,
             SA_ShowTitle,FALSE,
             SA_Type,CUSTOMSCREEN,
             SA_DisplayID,dispid,
             TAG_DONE);
  }

  /* Could the screen be opened? */
  if(DispScreen == NULL)
  {
    close_libraries();
    printf("The screen could NOT be opened!\n");
    return -1;
  }

  if (amigaprefs.def_type==GFX)
  {
    /* Get screenmode information */
    UnLockBitMap(LockBitMapTags(DispScreen->RastPort.BitMap,
                                LBMI_PIXFMT, (ULONG)&pixfmt,
                                LBMI_BASEADDRESS, (ULONG)&baseaddress,
                                LBMI_BYTESPERROW, (ULONG)&bpr,
                                LBMI_BYTESPERPIX, (ULONG)&bpp,
                                TAG_END));

    /* Set depth */
    if (pixfmt==PIXFMT_LUT8)
      DEPTH=8;
    else if (pixfmt==PIXFMT_RGB16)
      DEPTH=16;
    else if (pixfmt==PIXFMT_ARGB32)
      DEPTH=32;
    else
      DEPTH=32; /* 32bit as default since it will work on any 15bit+ screen using CGFX call */
  }

  /* Allocate chunky buffer */
  img_data=(unsigned char*)malloc(_width_*_height_*DEPTH/8);
  memset(img_data, 0, _width_*_height_*DEPTH/8);
  back_pix=(unsigned char*)malloc(_width_*_height_*DEPTH/8);
  memset(back_pix, 0, _width_*_height_*DEPTH/8);

  /* Open window */
  DispWindow=OpenWindowTags(NULL,
         WA_CustomScreen,(int)DispScreen,
         WA_Width,width,
         WA_Height,_height_,
         WA_IDCMP,IDCMP_RAWKEY,
         WA_Backdrop,TRUE,
         WA_Borderless,TRUE,
         WA_Activate,TRUE,
         TAG_DONE);

  /* Could the window be opened? */
  if(DispWindow == NULL)
  {    
    CloseScreen(DispScreen);
    if (amigaprefs.def_type!=GFX)
      FreeRaster(agabitmap.Planes[0], width, _height_*_depth_);
    close_libraries();
    printf("The window could NOT be opened!\n");
    return -1;
  }

  /* Clear the screen */
  if (bpp>1)
  {
    ptr=baseaddress;
    memset(ptr, 0, _width_*_height_*bpp);
  }

  /* HAM init required? */
  if (amigaprefs.def_type==AGAHAM || amigaprefs.def_type==AGAHAMPLUS)
  { 
    /* Set all colors to black */
    if (amigaprefs.def_depth==8)
      for(i=0; i<64; i++)
        set_color(i,0,0,0);
    else
      for(i=0; i<16; i++)
        set_color(i,0,0,0);

    /* Allocate init buff init buffer */
    ham_init=(unsigned char*)malloc(width*_height_);
    memset(ham_init, 0, width*_height_);
    ptr=ham_init;

    /* Init the init buffer */
    if (amigaprefs.def_type==AGAHAM)
    {
      for (i=0; i<width*_height_/4; i++)
      {
        *ptr++=0x40;
        *ptr++=0xc0;
        *ptr++=0x80;
        *ptr++=0xc0;
      }
    }
    else
    {
      for (i=0; i<width*_height_/4; i++)
      {
        *ptr++=0x80;
        *ptr++=0xc0;
        *ptr++=0x40;
        *ptr++=0x40;
      }
    }

    /* Redraw screen */
    if (amigaprefs.def_depth==8)
      c2p8((void *)ham_init, (void *)agabitmap.Planes[0], width, _height_, 0, 0, RASSIZE(width, _height_));
    else
      c2p6((void *)ham_init, (void *)agabitmap.Planes[0], width, _height_, 0, 0, RASSIZE(width, _height_));

    free(ham_init);
  }

  /* Set the palette if required */
  if (amigaprefs.def_type==AGA || (amigaprefs.def_type==GFX && bpp==1))
  {
    if (amigaprefs.def_depth==8)
    {
      for(i=0; i<256; i++)
      {
        blue  = (i & 0x03) << 6;
        green = (i & 0x1C) << 3;
        red   = (i & 0xE0);

        set_color(i,red,green,blue);
      }
      SetAPen(&DispScreen->RastPort, 255);
      SetBPen(&DispScreen->RastPort, 0);
    }
    else
    {
      for(i=0; i<64; i++)
      {
        blue  = (i & 0x03) << 6;
        green = (i & 0x0C) << 4;
        red   = (i & 0x30) << 2;

        set_color(i,red,green,blue);
      }
      SetAPen(&DispScreen->RastPort, 63);
      SetBPen(&DispScreen->RastPort, 0);
    }
  }

  if (amigaprefs.def_type==GFX && _depth_>8)
  {
    SetAPen(&DispScreen->RastPort, 0);
    SetBPen(&DispScreen->RastPort, 1);
  }

  return 0;
}

/*******************************************************************************
*                                                                              *
* int OpenVisual()                                                             *
* int CloseVisual()                                                            *
*                                                                              *
********************************************************************************
*                                                                              *
* These routines are used by the emulator to open and to close a visual.       *
*                                                                              *
* Both return a negative int on error and zero on success.                     *
*                                                                              *
*******************************************************************************/

int OpenVisual()
{
  if(InitDisplay() < 0)
    exit(20);
  /*  return(-1);  */

  prefs.display = LOCAL_DISPLAY;

  return(0);

} /* int OpenVisual() */

int CloseVisual()
{
  DestroyDisplay();

  prefs.display = NO_DISPLAY;

  return(0);

} /* int CloseVisual() */

/*******************************************************************************
*                                                                              *
* static int InitDisplay()                                                     *
*                                                                              *
********************************************************************************
*                                                                              *
* This routine is used to initialize the visual.                               *
* It should print serious information of the current status.                   *
* It must NOT be called by the emulator. It's called from 'int OpenVisual()'.  *
* The name is not standard but it should be used!                              *
*                                                                              *
* The routine returns a negative int on error and zero on success.             *
*                                                                              *
*******************************************************************************/

static int InitDisplay()
{
  printf("Initializing AmigaOS display: (by Mathias \"AmiDog\" Roslund)\n");

  printf("Opening libraries ... ");
  if (open_libraries()<0)
    return -1;
  else
    printf("OK\n");

  printf("Reading AmigaOS prefs ... ");
  if (load_prefs()<0)
  {
    close_libraries();
    return -1;
  }
  else
    printf("OK\n");

  printf("Opening screen ... ");
  if (open_screen(640,480)<0)
    return -1;
  else
    printf("OK\n");

  printf("-------------------------------------------------------------------------------\n");

  RefreshVisual();

  return(0);

} /* static int InitDisplay() */

/*******************************************************************************
*                                                                              *
* static int DestroyDisplay()                                                  *
*                                                                              *
********************************************************************************
*                                                                              *
* This routine is used to destroy the visual.                                  *
* It should print serious information of the current status.                   *
* It must NOT be called by the emulator. It's called from 'int CloseVisual()'. *
* The name is not standard but it should be used!                              *
*                                                                              *
* The routine returns a negative int on error and zero on success.             *
*                                                                              *
*******************************************************************************/

static int DestroyDisplay()
{
  printf("-------------------------------------------------------------------------------\n");
  printf("Destroying AmigaOS display:\n");

  printf("Closing screen ... ");
  close_screen();
  printf("OK\n");

  printf("Closing libraries ... ");
  close_libraries();
  printf("OK\n");

  printf("-------------------------------------------------------------------------------\n");

  return(0);

} /* static int DestroyDisplay() */


static void PrintText(char *text, int x, int y)
{
  /* Move text start */
  Move(&DispScreen->RastPort, x, y);

  /* Print text */
  Text(&DispScreen->RastPort, text, strlen(text));

} /* static void PrintText(char *text, int x, int y) */


void RefreshVisual()
{
        uWORD    new_width;
        int     i;
        uBYTE    *Imsrc1, *Imdst1, c;
        uHWORD   *Imsrc2, *Imdst2, ui;
        uWORD    *Imsrc4, *Imdst4, ul;

#ifdef FPS
        static char             fps_text[40];
        int                     fps_update_sec = 1;   /* min update frequency in sec */
                                                      /* otherwise you might not read it properly */
        static long             fps_update = 0;   /* stores the next time for update */
        static int              fps_count = 0;   /* to make fps average */
        static float            fps_rate = 0.0;
        static struct timeval   fps_tv_old  = { 0, 0 };
        struct timeval          fps_tv;
#endif

        /* video flags */
        flags = ((uWORD *)mem.vi_reg)[0];

        /* video width */
        /* if width in VI_WIDTH_REG (0x04400008) has changed set windows width */

        new_width = ((uWORD *)mem.vi_reg)[2];
        if(new_width != width)
        {
                if(new_width < 320 || new_width > 640)
                        width = 640;
                else
                        width = new_width;

                height = (width * 3) >> 2;

               /* Change screensize (AmiDog) */
               close_screen();
               open_screen(width, height);
               printf("AmigaOS: screensize changed to: %lu, %lu\n", width, height);
        }

        /* display it! */

        if(new_width >= 320 && new_width <= 640)
        {
                if(flags & 0x0003)
                {                       
                        /* Compare screensize with WIDTH/HEIGHT */
                        if(WIDTH != _width_ || HEIGHT != _height_)
                        {

                                WIDTH  = _width_;
                                HEIGHT = _height_;

                        } /* if(WIDTH != _width_ || HEIGHT != _height_) */

                        /* source video origin */
                        Imsrc4 = (uWORD  *) (mem.rd_ram + ((uWORD *)mem.vi_reg)[1]);
                        Imsrc2 = (uHWORD *) Imsrc4;
                        Imsrc1 = (uBYTE  *) Imsrc2;
         
                        /* destination video origin */
                        Imdst4 = (uWORD  *) img_data;
                        Imdst2 = (uHWORD *) Imdst4;
                        Imdst1 = (uBYTE  *) Imdst2;

                        switch(DEPTH)
                        {
                            case 8:
                                switch(flags & 0x0003)
                                {
                                    case 0:
                                    case 1:
                                        break;

                                    case 2:   /* local: 2/2/2 (6bit) or 3/3/2 (8bit) n64: 5/5/5/3 (16bit) */
                                        if (amigaprefs.def_depth==8)
                                        {
                                          for(i=0; i<width*height; i++)
                                          {
                                            ui = *Imsrc2++;
                                            c = (((ui & 0x30) >> 4) | ((ui & 0x700) >> 6) | (( ui & 0xe000) >> 8));
                                            *Imdst1++ = c;
                                          }
                                        }
                                        else
                                        {
                                          for(i=0; i<width*height; i++)
                                          {
                                            ui = *Imsrc2++;
                                            c = (((ui & 0x30) >> 2) | ((ui & 0x600) >> 5) | (( ui & 0xc000) >> 8));
                                            *Imdst1++ = c;
                                          }                                        
                                        }
                                        break;

                                    case 3:   /* local: 2/2/2 (6bit) or 3/3/2 (8bit) n64: 8/8/8/8 (32bit) */
                                        if (amigaprefs.def_depth==8)
                                        {
                                          for(i=0; i<width*height; i++)
                                          {
                                            ul = *Imsrc4++;
                                            c = (((ul & 0xc0) >> 6) | ((ul & 0xe000) >> 11) | (( ul & 0xe00000) >> 16));
                                            *Imdst1++ = c;
                                          }
                                        }
                                        else
                                        {
                                          for(i=0; i<width*height; i++)
                                          {
                                            ul = *Imsrc4++;
                                            c = (((ul & 0xc0) >> 4) | ((ul & 0xc000) >> 10) | (( ul & 0xc00000) >> 16));
                                            *Imdst1++ = c;
                                          }
                                        }
                                        break;

                                } /* switch(flags & 0x0003) */
                                break;

                            case 16:
                                switch(flags & 0x0003)
                                {
                                    case 0:
                                    case 1:
                                        break;

                                    case 2:   /* local: 5/6/5 (16bit)   n64: 5/5/5/3 (16bit) */
                                        Imdst2=(unsigned short*)baseaddress;
                                        for(i=0; i<width*height; i++)
                                        {
                                          ui = *Imsrc2++;
                                          *Imdst2++ = (((ui >> 1) & 0x001f) | (ui & 0xffc0));
                                        }
                                        break;

                                    case 3:   /* local: 5/6/5 (16bit)   n64: 8/8/8/8 (32bit) */
                                        Imdst2=(unsigned short*)baseaddress;
                                        for(i=0; i<width*height; i++)
                                        {
                                          ul = *Imsrc4++;
                                          ui = ((ul & 0xf8) >> 3) | ((ul & 0xfc00) >> 5) | ((ul & 0xf80000) >> 8);
                                          *Imdst2++ = ui;
                                        }
                                        break;

                                } /* switch(flags & 0x0003) */
                                break;

                            case 32:
                                switch(flags & 0x0003)
                                {
                                    case 0:
                                    case 1:
                                        break;

                                    case 2:   /* local: 8/8/8 (32bit)   n64: 5/5/5/3 (16bit) */
                                        for(i=0; i<width*height; i++)
                                        {
                                          ui = *Imsrc2++;

                                          *Imdst4++ = ((ui << 8) & 0xf80000) | ((ui <<  5) & 0x00f800) | ((ui <<  2) & 0x0000f8);
                                        }
                                        break;

                                    case 3:   /* local: 8/8/8 (32bit)   n64: 8/8/8/8 (32bit) */
                                        for(i=0; i<width*height; i++)
                                        {
                                          ul = *Imsrc4++;

                                          *Imdst4++ = (ul >> 8);
                                        }
                                        break;

                                } /* switch(flags & 0x0003) */
                                break;

                            default:
                                fprintf(stderr, "Display depth of %d bit is not supported! (yet ;) )\n", DEPTH);

                        } /* switch(DEPTH) */

                        PutImage(0, 0, WIDTH, HEIGHT);
                }
                else
                {
                        PrintText("DISPLAY: blank (no data, no sync)", 20, 40);
                }
        }
        else
        {
                if(!(flags & 0x0003))
                        PrintText("DISPLAY: blank (no data, no sync)", 20, 40);

                PrintText("DISPLAY: wrong width (640 will be used!)", 20, 60);
        }

#ifdef FPS
        if (amigaprefs.def_fps==0)
        {

        GetSysTimePPC(&fps_tv);

        fps_rate = fps_count * fps_rate
                   + 1 / ( ((float)fps_tv.tv_sec     + (float)fps_tv.tv_usec     / 1000000)
                         - ((float)fps_tv_old.tv_sec + (float)fps_tv_old.tv_usec / 1000000) );

        fps_count++;

        fps_rate /= fps_count;

        if(fps_update <= fps_tv.tv_sec)
        {
                fps_update = fps_tv.tv_sec + fps_update_sec;
                fps_count  = 0;

                if(fps_rate > 99.9)
                        fps_rate = 99.99;

                sprintf(fps_text, "%05.2f", fps_rate);

                fps_rate = 0.0;
        }

        fps_tv_old.tv_sec = fps_tv.tv_sec;
        fps_tv_old.tv_usec = fps_tv.tv_usec;

        PrintText(fps_text, 40, 20);

        }
#endif

} /* void RefreshVisual() */


void GetVisualName(char *name, int *n)
{
  strncpy(name, DISPLAYNAME, (size_t) *n);

  *n = strlen(DISPLAYNAME);

} /* void GetVisualName() */


static void PutImage(int X, int Y, int W, int H)
{
  unsigned char *dst, *src, g;
  int i;

  if (amigaprefs.def_type==GFX)
  {
    /* Fullscreen CGFX using direct memory access */
    src=(unsigned char *)img_data;
    dst=baseaddress;

    if (pixfmt==PIXFMT_LUT8)
    {
      for (i=0; i<H; i++)
      {
        memcpy(dst, src, W);
        dst+=bpr;
        src+=W;
      }
    }
    else if (pixfmt==PIXFMT_ARGB32)
    {
      for (i=0; i<H; i++)
      {
        memcpy(dst, src, W*4);
        dst+=bpr;
        src+=W*4;
      }
    }
    else if (pixfmt!=PIXFMT_LUT8 && pixfmt!=PIXFMT_RGB16 && pixfmt!=PIXFMT_ARGB32)
    {
      /* Works on ANY 15bit+ screen! */
      WritePixelArray(img_data, 0, 0, W*4, &DispScreen->RastPort, 0, 0, W, H, RECTFMT_ARGB);
    }
  }
  else if (amigaprefs.def_type==AGA)
  {
    /* Perform a fullscreen c2p */
    if (amigaprefs.def_depth==8)
      c2p8((void *)img_data, (void *)agabitmap.Planes[0], W, H, 0, 0, RASSIZE(_width_, _height_));
    else 
      c2p6((void *)img_data, (void *)agabitmap.Planes[0], W, H, 0, 0, RASSIZE(_width_, _height_));
  }  
  else if (amigaprefs.def_type==AGAHAM || (amigaprefs.def_type==AGAHAMPLUS && _width_!=320))
  {
    src=(unsigned char *)img_data;
    dst=(unsigned char *)ham_buf;

    /* 32bit -> HAM, uses 2*1 pixels, good quality and speed! */
    for (i=0; i<_width_*_height_/2; i++)
    {
        src+=2;         /* Skip "unused byte" and R */
        g=*src++;       /* Get G */
        *dst++=*src++;  /* Store B */
        *dst++=g;       /* Store G */
        src++;          /* Skip "unsused" byte */
        *dst++=*src++;  /* Store R */
        *dst++=*src++;  /* Store G */
        src++;          /* Skip B */
    }

    /* Perform C2P */
    if (amigaprefs.def_depth==8)
      c2p6((void *)ham_buf, (void *)agabitmap.Planes[0], W*2, H, 0, 0, RASSIZE(_width_*2, _height_));
    else
      c2p4((void *)ham_buf, (void *)agabitmap.Planes[0], W*2, H, 0, 0, RASSIZE(_width_*2, _height_)); 
  }
  else if (amigaprefs.def_type==AGAHAMPLUS && _width_==320)
  {
    src=(unsigned char *)img_data;
    dst=(unsigned char *)ham_buf;

    /* 32bit -> HAM, uses 4*1 pixels, very good quality, poor speed. */
    for (i=0; i<_width_*_height_; i++)
    {
        src++;         /* Skip "unused" byte */
        *dst++=*src++; /* Store R */
        *dst++=*src++; /* Store G */
        *dst++=*src;   /* Store B */
        *dst++=*src++; /* Store B */
    }

    /* Perform C2P */
    if (amigaprefs.def_depth==8)
      c2p6((void *)ham_buf, (void *)agabitmap.Planes[0], W*4, H, 0, 0, RASSIZE(_width_*4, _height_));
    else
      c2p4((void *)ham_buf, (void *)agabitmap.Planes[0], W*4, H, 0, 0, RASSIZE(_width_*4, _height_)); 
  }
} /* static void PutImage(int X, int Y, int W, int H) */


void SetVisualTitle(char *title)
{}


void Visual3D_FPS()
{}


static char *rdp_address=NULL;
static int  rdp_format;
static int  rdp_size;
static int  rdp_width;


void SetVisualRenderBuffer(int format, int size, int width, char *address)
{
#ifndef JUST3D
        /* Q:Why reallocate if format hasn't changed? A:Don't! */
        if (rdp_format==format && rdp_size==size && rdp_width==width && rdp_address==address && valid_back_pix==1)
        {
            memset(back_pix, 0, ((rdp_width*rdp_width*3/4)<<1));
            return;
        }

        rdp_format  = format;
        rdp_size    = size;
        rdp_width   = width;
        rdp_address = address;

        if (valid_back_pix)   /* This is 0 after OpenVisual() and after CloseVisual(). */
        {
            /* Free buffer */
            free(back_pix);
            back_pix=NULL;
        }
        else
        {
            valid_back_pix = 1;
        } 

        /* Allocate new buffer (always 16bit currently) */        
        back_pix=(unsigned char*)malloc(((rdp_width*rdp_width*3/4)<<1));
        memset(back_pix, 0, ((rdp_width*rdp_width*3/4)<<1));
#endif

#ifdef AMIGA_DEBUG
        printf("SetVisualRenderBuffer (%d %d %d)...\n", format, size, width);
#endif
}


void FlushVisualRenderBuffer()
{
#ifndef JUST3D
        void    *src;
        void    *dst;
        int     i;

        if(valid_back_pix == 1 && rdp_format == 0 && rdp_size == 2)
        {
                dst = rdp_address;
                src = back_pix;

                for(i=0; i<((rdp_width*rdp_width*3/4)<<1); i++)
                {
//                        if(((uBYTE *)src)[i] & 0x01)
//                        {
                                ((uBYTE *)dst)[i] = ((uBYTE *)src)[i];
//                        }
                }
#ifdef AMIGA_DEBUG
        printf("FlushVisualRenderBuffer...\n");
#endif
        }
#endif
}


void SetVisualColor(int color)
{
  unsigned long temp;
  temp=(unsigned long)color;
  drawcolor=((temp&0x00F80000)>>9 | (temp&0x0000F800)>>6 | (temp&0x000000F8)>>3 );
}


void DrawVisualPoint(int x, int y)
{
  unsigned short *dst;

  dst=(unsigned short*)back_pix;

  if (x>=0 && x<_width_)
    if (y>=0 && y<_height_)
      dst[x+_width_*y]=drawcolor;
}


void DrawVisualLine(int x1, int y1, int x2, int y2)
{
  unsigned short *dst;
  int x, y, deltax, deltay, xchange, ychange, i, length, error=0;

  dst=(unsigned short*)back_pix;

  x=x1;
  y=y1;

  deltax=x2-x1;
  deltay=y2-y1;

  if (deltax<0)
  {
    xchange=-1;
    deltax=-deltax;
  }
  else
    xchange=1;

  if (deltay<0)
  {
    ychange=-1;
    deltay=-deltay;
  }
  else
    ychange=1;

  dst[x+_width_*y]=drawcolor;

  if (deltax<deltay)
  {
    length=deltay;

    for (i=0; i<length; i++)
    {
      y+=ychange;
      error+=deltax;

      if (error>deltay)
      {
        x+=xchange;
        error-=deltay;
      }

      if (x>=0 && x<_width_)
        if (y>=0 && y<_height_)
          dst[x+_width_*y]=drawcolor;
    }
  }
  else
  {
    length=deltax;

    for (i=0; i<length; i++)
    {
      x+=xchange;
      error+=deltay;

      if (error>deltax)
      {
        y+=ychange;
        error-=deltax;
      }

      if (x>=0 && x<_width_)
        if (y>=0 && y<_height_)
          dst[x+_width_*y]=drawcolor;
    }
  }
}


void DrawVisualTriangle(int x1, int y1, int x2, int y2, int x3, int y3)
{
  DrawVisualLine(x1, y1, x2, y2);
  DrawVisualLine(x2, y2, x3, y3);
  DrawVisualLine(x3, y3, x1, y1);
}


void InitVisualRenderBuffer()
{}


void ClearVisual()
{
  memset(back_pix, 0, ((rdp_width*rdp_width*3/4)<<1));
  drawcolor = 0x0000;
}


void DrawVisualRectangle(int ulx, int uly, int lrx, int lry)
{
  DrawVisualLine(ulx, uly, lrx, uly);
  DrawVisualLine(ulx, lry, lrx, lry);
  DrawVisualLine(ulx, uly, ulx, lry);
  DrawVisualLine(lrx, uly, lrx, lry);
}


void DrawVisualFillRect(int ulx, int uly, int lrx, int lry)
{
#ifdef AMIGA_UNIMP
  printf("DrawVisualFillRect...\n");
#endif
}


void SetVisualSize(int w, int h)
{
#ifdef AMIGA_UNIMP
  printf("SetVisualSize (%d %d)...\n", w, h);
#endif
}


void SetVisualClipRectangle(int ulx, int uly, int lrx, int lry)
{
#ifdef AMIGA_UNIMP
  printf("SetVisualClipRectangle (%d %d %d %d)...\n", ulx, uly, lrx, lry);
#endif
}


void DrawVisualTextureRectangle(int ulx, int uly, int lrx, int lry, int tile, float f_s, float f_t, float f_dsdx, float f_dtdy)
{
/*
        int  width;
        int  height;
        XGCValues newgc;

        if(tile_img_used[tile] == 1)
                free(tile_img[tile]);
        else
                tile_img_used[tile] = 1;

        width  = rdp_reg.td[tile].lrs - rdp_reg.td[tile].uls + 1;
        height = rdp_reg.td[tile].lrt - rdp_reg.td[tile].ult + 1;
        tile_img[tile] = XCreateImage(dsp, DefaultVisualOfScreen(scr), DEPTH, ZPixmap, 0,
                (uBYTE *)mem.tmem + rdp_reg.td[tile].tmem, width, height, DEPTH, 0);

        //printf("DrawVisualTextureRectangle: tile=%d\n", tile);

        newgc.function = GXcopy;
        XChangeGC(dsp, gc, GCFunction, &newgc);
        XPutImage(dsp, win, gc, tile_img[tile], 0, 0, ulx, uly, lrx-ulx+10, lry-uly+10);
        newgc.function = GXcopy;
        XChangeGC(dsp, gc, GCFunction, &newgc);
*/
#ifdef AMIGA_UNIMP
        printf("DrawVisualTextureRectangle...\n");
#endif
}


void LoadVisualBlock(int addr, int size, int tmem, int tile, int uls, int ult, int lrs, int dxt)
{
        int  lwidth;                            /* width of src scanline */
        int  o;                                 /* offset of src block */
        int  bytes;                             /* src bytes */
        int  dst;
        int  i;


#define MAKE_BYTES(x) ((x) << size >> 1)                /* pixel2bytes: bytes = pixel * (1|2|4|8) / 2 */

        lwidth = dxt << 2;                              /* scanlinewidth in bytes */
        o      = MAKE_BYTES(uls) + ult * lwidth;
        bytes  = MAKE_BYTES(lrs - uls + 1);             /* bytes = pixel * (1|2|4|8) / 2 */
        dst    = tmem;

#undef MAKE_BYTES(x)
/*
        printf("lwidth: %d\n", lwidth);
        printf("o     : %d\n", o     );
        printf("bytes:  %d\n", bytes);
        printf("tmem:   %d\n", tmem);
*/
        addr += o;

        /* we might have to convert to a usable format for the display driver here. */

        for(i=0; i<bytes; i++)
        {
                ((uBYTE *)mem.tmem)[dst] = ((uBYTE *)mem.rd_ram)[i+addr];
                dst++;
        }

#ifdef AMIGA_UNIMP
        printf("LoadVisualBlock...\n");
#endif
}
/*
 * addr: source addr
 * size: 4bit: 0
 *       8bit: 1
 *      16bit: 2
 *      32bit: 3
 * tmem: address offset of tmem
 * tile: 0..7
 * uls, ult: always 0? maybe also offset - then dxt could be used
 * lrs: lrs - uls + 1 = pixel of tile
 * dxt: ? strange - maybe used with offset.
 *      look at the source in x11/x11.c how i have used it.
 */


