/**
 * TrueReality - output/display/amigados/amigadosJ3D.c
 * Copyright (C) 1998, 1999 Mathias "AmiDog" Roslund
 *
 * This program is free software; you can redistribute it and/
 * or modify it under the terms of the GNU General Public Li-
 * cence as published by the Free Software Foundation; either
 * version 2 of the Licence, or any later version.
 *
 * This program is distributed in the hope that it will be use-
 * ful, but WITHOUT ANY WARRANTY; without even the implied war-
 * ranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public Licence for more details.
 *
 * You should have received a copy of the GNU General Public
 * Licence along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 * USA.
 *
 * Information about me (the author):
 *   Niki W. Waibel, Reichenau 20, 6890 Lustenau, Austria - EUROPE
 *   niki.waibel@gmx.net
**/





/**
 *
 * All of this stuff comes from AmiDog - great work (niki)
 *
**/





#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <signal.h>
#include <unistd.h>
#include <math.h>

#include <exec/types.h>
#include <exec/nodes.h>
#include <exec/lists.h>
#include <exec/memory.h>
#include <powerup/ppcproto/intuition.h>
#include <powerup/ppclib/interface.h>
#include <powerup/ppclib/time.h>
#include <powerup/gcclib/powerup_protos.h>
#include <powerup/ppcproto/exec.h>
#include <powerup/ppcinline/graphics.h>
#include <powerup/ppcproto/dos.h>

#include <cybergraphx/cybergraphics.h>
#include <powerup/ppcinline/cybergraphics.h>

#include <config.h>
#include <parser_extern.h>
#include <type_sizes.h>
#include <memory_extern.h>
#include <dispatch.h>

#include "../display.h"
#include "../sw_renderer.h"
#include "../../../N64/registers.h"
#include "../../../N64/draw2d_sw.h"
#include "amigadosJ3D.h"

/* FPS */
#include <sys/time.h>
#include <unistd.h>

/* Scaling */
static float stretch_x;
static float stretch_y;

/* Original Variables */
static uWORD width=0;
static uWORD screenwidth;
static uWORD screenheight;

static int WIDTH=640;
static int HEIGHT=480;

/* Local functions */
static int InitDisplay();
static int DestroyDisplay();
static void PrintText(char *text, int x, int y);
static void PutImage(int x, int y, int w, int h);

/* Amiga specific variables */
struct IntuitionBase *IntuitionBase;
struct GfxBase *GfxBase;
struct Library *CyberGfxBase;
struct Screen *DispScreen;
struct Window *DispWindow;

/* Render color */
unsigned short drawcolor=0x0000;

/* CGFX variables */
unsigned long pixfmt;
unsigned char *baseaddress;
int bpr;

int open_libraries()
{
  IntuitionBase = (struct IntuitionBase *) OpenLibrary("intuition.library", 39);

  if(IntuitionBase==NULL)
    return -1;

  GfxBase = (struct GfxBase *) OpenLibrary("graphics.library", 39);

  if(GfxBase==NULL)
  {
    CloseLibrary(IntuitionBase);
    return -1;
  }

  CyberGfxBase = OpenLibrary("cybergraphics.library", 40);

  if(CyberGfxBase==NULL)
  {
    CloseLibrary(GfxBase);
    CloseLibrary(IntuitionBase);
    return -1;
  }

  return 0;
}

void close_libraries()
{
  CloseLibrary(CyberGfxBase);
  CloseLibrary(GfxBase);
  CloseLibrary(IntuitionBase);
}

void close_screen()
{
  CloseWindow(DispWindow);
  CloseScreen(DispScreen);
}

int open_screen()
{
  unsigned long VideoMode;
  UWORD models[]={PIXFMT_RGB15};
  int width, height, depth;

  /* Ask the user for a screenmode */
  VideoMode = CModeRequestTags(NULL,
                               CYBRMREQ_MinDepth, 15,
                               CYBRMREQ_MaxDepth, 15,
                               CYBRMREQ_CModelArray, models,
                               TAG_DONE);

  if (VideoMode==0)
  {
    printf("You MUST choose a screenmode!\n");
    close_libraries();
    return -1;
  }

  /* Get VideoMode information */
  width=GetCyberIDAttr(CYBRIDATTR_WIDTH, VideoMode);
  height=GetCyberIDAttr(CYBRIDATTR_HEIGHT, VideoMode);
  depth=GetCyberIDAttr(CYBRIDATTR_DEPTH, VideoMode);

  /* Open screen */
  DispScreen = OpenScreenTags(NULL,
               SA_Width,width,
               SA_Height,height*2,
               SA_Depth,depth,
               SA_Quiet,TRUE,
               SA_ShowTitle,FALSE,
               SA_Type,CUSTOMSCREEN,
               SA_DisplayID,VideoMode,
               TAG_DONE);

  /* Could the screen be opened? */
  if (DispScreen == NULL)
  {
    close_libraries();
    printf("Could NOT open screen!\n");
    return -1;
  }

  /* Open window */
  DispWindow = OpenWindowTags(NULL,
               WA_CustomScreen,(int)DispScreen,
               WA_Width,DispScreen->Width,
               WA_Height,DispScreen->Height,
               WA_IDCMP,IDCMP_RAWKEY,
               WA_Backdrop,TRUE,
               WA_Borderless,TRUE,
               WA_Activate,TRUE,
               TAG_DONE);

  /* Could the window be opened? */
  if(DispWindow == NULL)
  {
    CloseScreen(DispScreen);
    close_libraries();
    printf("Could NOT open window!\n");
    return -1;
  }

  /* Get screenmode information */
  UnLockBitMap(LockBitMapTags(DispScreen->RastPort.BitMap,
                              LBMI_PIXFMT, (ULONG)&pixfmt,
                              LBMI_BASEADDRESS, (ULONG)&baseaddress,
                              LBMI_BYTESPERROW, (ULONG)&bpr,
                              TAG_END));

  /* Check depth */
  if (pixfmt!=PIXFMT_RGB15)
  {
    printf("Could NOT open 15bit screen!\n");
    DestroyDisplay();
  }

  /* Init variables */
  screenwidth=width;
  screenheight=height;
  WIDTH  = screenwidth;
  HEIGHT = screenheight;
  stretch_x = (float)WIDTH/width;
  stretch_y = (float)HEIGHT/((width*3)>>2);

  /* Clear the screen */
  FillPixelArray(&DispScreen->RastPort, 0, 0, screenwidth, screenheight*2, 0x00000000);

  return 0;
}

/*******************************************************************************
*                                                                              *
* int OpenVisual()                                                             *
* int CloseVisual()                                                            *
*                                                                              *
********************************************************************************
*                                                                              *
* These routines are used by the emulator to open and to close a visual.       *
*                                                                              *
* Both return a negative int on error and zero on success.                     *
*                                                                              *
*******************************************************************************/

int OpenVisual()
{
  if(InitDisplay() < 0)
    exit(20);

  prefs.display = LOCAL_DISPLAY;

  return(0);

} /* int OpenVisual() */

int CloseVisual()
{
  DestroyDisplay();

  prefs.display = NO_DISPLAY;

  return(0);

} /* int CloseVisual() */

/*******************************************************************************
*                                                                              *
* static int InitDisplay()                                                     *
*                                                                              *
********************************************************************************
*                                                                              *
* This routine is used to initialize the visual.                               *
* It should print serious information of the current status.                   *
* It must NOT be called by the emulator. It's called from 'int OpenVisual()'.  *
* The name is not standard but it should be used!                              *
*                                                                              *
* The routine returns a negative int on error and zero on success.             *
*                                                                              *
*******************************************************************************/

static int InitDisplay()
{
  printf("Initializing AmigaOS display: (by Mathias \"AmiDog\" Roslund)\n");

  printf("Opening libraries ... ");
  if (open_libraries()<0)
    return -1;
  else
    printf("OK\n");

  printf("Opening screen ... ");
  if (open_screen()<0)
    return -1;
  else
    printf("OK\n");

  printf("-------------------------------------------------------------------------------\n");

  return(0);

} /* static int InitDisplay() */

/*******************************************************************************
*                                                                              *
* static int DestroyDisplay()                                                  *
*                                                                              *
********************************************************************************
*                                                                              *
* This routine is used to destroy the visual.                                  *
* It should print serious information of the current status.                   *
* It must NOT be called by the emulator. It's called from 'int CloseVisual()'. *
* The name is not standard but it should be used!                              *
*                                                                              *
* The routine returns a negative int on error and zero on success.             *
*                                                                              *
*******************************************************************************/

static int DestroyDisplay()
{
  printf("-------------------------------------------------------------------------------\n");
  printf("Destroying AmigaOS display:\n");

  printf("Closing screen ... ");
  close_screen();
  printf("OK\n");

  printf("Closing libraries ... ");
  close_libraries();
  printf("OK\n");

  printf("-------------------------------------------------------------------------------\n");

  return(0);

} /* static int DestroyDisplay() */


static void PrintText(char *text, int x, int y)
{
  Move(&DispScreen->RastPort, x, y+screenheight);
  Text(&DispScreen->RastPort, text, strlen(text));
}


void RefreshVisual()
{}


void GetVisualName(char *name, int *n)
{
  strncpy(name, DISPLAYNAME, (size_t) *n);
  *n = strlen(DISPLAYNAME);
}


static void PutImage(int X, int Y, int W, int H)
{}


void SetVisualTitle(char *title)
{}


void Visual3D_FPS()
{
#ifdef FPS
        static char             fps_text[40];
        int                     fps_update_sec = 1;   /* min update frequency in sec */
                                                      /* otherwise you might not read it properly */
        static long             fps_update = 0;   /* stores the next time for update */
        static int              fps_count = 0;   /* to make fps average */
        static float            fps_rate = 0.0;
        static struct timeval   fps_tv_old  = { 0, 0 };
        struct timeval          fps_tv;

        GetSysTimePPC(&fps_tv);

        fps_rate = fps_count * fps_rate
                   + 1 / ( ((float)fps_tv.tv_sec     + (float)fps_tv.tv_usec     / 1000000)
                         - ((float)fps_tv_old.tv_sec + (float)fps_tv_old.tv_usec / 1000000) );

        fps_count++;

        fps_rate /= fps_count;

        if(fps_update <= fps_tv.tv_sec)
        {
                fps_update = fps_tv.tv_sec + fps_update_sec;
                fps_count  = 0;

                if(fps_rate > 99.9)
                        fps_rate = 99.99;

                sprintf(fps_text, "%05.2f", fps_rate);

                fps_rate = 0.0;
        }

        fps_tv_old.tv_sec = fps_tv.tv_sec;
        fps_tv_old.tv_usec = fps_tv.tv_usec;

        PrintText(fps_text, 40, 20);
#endif
}


void InitVisualRenderBuffer()
{
  /* Copy from buffer do display */
  MovePixelArray(0, screenheight, &DispScreen->RastPort, 0, 0, screenwidth, screenheight);

  width = ((uWORD *)mem.vi_reg)[2];
  WIDTH  = screenwidth;
  HEIGHT = screenheight;
  stretch_x = (float)WIDTH/width;
  stretch_y = (float)HEIGHT/((width*3)>>2);

  /* Clear the buffer */
  drawcolor=0x0000;
  FillPixelArray(&DispScreen->RastPort, 0, screenheight, screenwidth, screenheight, 0x00000000);
}


void SetVisualRenderBuffer(int format, int size, int width, char *address)
{}


void FlushVisualRenderBuffer()
{}


void SetVisualColor(int color)
{
  unsigned long temp;
  temp=(unsigned long)color;
  drawcolor=((temp&0x00F80000)>>9 | (temp&0x0000F800)>>6 | (temp&0x000000F8)>>3 );
}


void DrawVisualPoint(int x, int y)
{
  unsigned short *dst;

  dst=(unsigned short*)baseaddress;
  dst+=(screenwidth*screenheight);

  if (x>=0 && x<screenwidth)
    if (y>=0 && y<screenheight)
      dst[x+screenwidth*y]=drawcolor;
}


void DrawVisualLine(int x1, int y1, int x2, int y2)
{
  unsigned short *dst;
  int x, y, deltax, deltay, xchange, ychange, i, length, error=0;

  dst=(unsigned short*)baseaddress;
  dst+=(screenwidth*screenheight);

  x=x1;
  y=y1;

  deltax=x2-x1;
  deltay=y2-y1;

  if (deltax<0)
  {
    xchange=-1;
    deltax=-deltax;
  }
  else
    xchange=1;

  if (deltay<0)
  {
    ychange=-1;
    deltay=-deltay;
  }
  else
    ychange=1;

  dst[x+screenwidth*y]=drawcolor;

  if (deltax<deltay)
  {
    length=deltay;

    for (i=0; i<length; i++)
    {
      y+=ychange;
      error+=deltax;

      if (error>deltay)
      {
        x+=xchange;
        error-=deltay;
      }

      if (x>=0 && x<screenwidth)
        if (y>=0 && y<screenheight)
          dst[x+screenwidth*y]=drawcolor;
    }
  }
  else
  {
    length=deltax;

    for (i=0; i<length; i++)
    {
      x+=xchange;
      error+=deltay;

      if (error>deltax)
      {
        y+=ychange;
        error-=deltax;
      }

      if (x>=0 && x<screenwidth)
        if (y>=0 && y<screenheight)
          dst[x+screenwidth*y]=drawcolor;
    }
  }
}


void DrawVisualTriangle(int x1, int y1, int x2, int y2, int x3, int y3)
{
  x1 = (int)((float)(x1) * stretch_x);
  y1 = (int)((float)(y1) * stretch_y);
  x2 = (int)((float)(x2) * stretch_x);
  y2 = (int)((float)(y2) * stretch_y);
  x3 = (int)((float)(x3) * stretch_x);
  y3 = (int)((float)(y3) * stretch_y);

  DrawVisualLine(x1, y1, x2, y2);
  DrawVisualLine(x2, y2, x3, y3);
  DrawVisualLine(x3, y3, x1, y1);
}


void ClearVisual()
{
  /* Clear the screen */
  FillPixelArray(&DispScreen->RastPort, 0, 0, screenwidth, screenheight*2, 0x00000000);
}


void DrawVisualRectangle(int ulx, int uly, int lrx, int lry)
{
  int x1 = (int)((float)(ulx) * stretch_x);
  int y1 = (int)((float)(uly) * stretch_y);
  int x2 = (int)((float)(lrx) * stretch_x);
  int y2 = (int)((float)(lry) * stretch_y);

  DrawVisualLine(x1, y1, x2, y1);
  DrawVisualLine(x1, y2, x2, y2);
  DrawVisualLine(x1, y1, x1, y2);
  DrawVisualLine(x2, y1, x2, y2);
}


// ----------- Not finished yet --------------


void DrawVisualFillRect(int ulx, int uly, int lrx, int lry)
{
//  printf("DrawVisualFillRect...(%d %d %d %d)\n", ulx, uly, lrx, lry);
}


void SetVisualSize(int w, int h)
{
//  printf("SetVisualSize (%d %d)...\n", w, h);
}


void SetVisualClipRectangle(int ulx, int uly, int lrx, int lry)
{
//  printf("SetVisualClipRectangle (%d %d %d %d)...\n", ulx, uly, lrx, lry);
}

void LoadVisualBlock(int addr, int size, int tmem, int tile, int uls, int ult, int lrs, int dxt)
{}
/*
 * addr: source addr
 * size: 4bit: 0
 *       8bit: 1
 *      16bit: 2
 *      32bit: 3
 * tmem: address offset of tmem
 * tile: 0..7
 * uls, ult: always 0? maybe also offset - then dxt could be used
 * lrs: lrs - uls + 1 = pixel of tile
 * dxt: ? strange - maybe used with offset.
 *      look at the source in x11/x11.c how i have used it.
 */

