/**
 * TrueReality - output/display/gl_renderer.h
 * Copyright (C) 1998, 1999 by Niki W. Waibel
 *
 * This program is free software; you can redistribute it and/
 * or modify it under the terms of the GNU General Public Li-
 * cence as published by the Free Software Foundation; either
 * version 2 of the Licence, or any later version.
 *
 * This program is distributed in the hope that it will be use-
 * ful, but WITHOUT ANY WARRANTY; without even the implied war-
 * ranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public Licence for more details.
 *
 * You should have received a copy of the GNU General Public
 * Licence along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 * USA.
 *
 * Information about me (the author):
 *   Niki W. Waibel, Reichenau 20, 6890 Lustenau, Austria - EUROPE
 *   niki.waibel@gmx.net
**/





/**
 * most of the following stuff is from FiRES and Gil.
 * some of the ideas are from Jeff.
 *
 * you all did great work!
 * - niki
**/





/* yep! we are using opengl :-) */

#include <GL/gl.h>
#include <GL/glut.h>





/* display list defines for 2d stuff */

#define kListEnd2D          1
#define kListBegin2D160   160
#define kListBegin2D320   320
#define kListBegin2D640   640
#define kListBegin2D800   800
#define kListBegin2D960   960
#define kListBegin2D1024 1024
#define kListBegin2D1200 1200



/**
 * Texture caching was first done by Jeff.
 * FiRES has modified it compleately.
**/

#define CACHE_TEXTURES
#define MAX_CACHED_TEXTURES 1024





/* prototypes */

void CreateDisplayLists();
#ifndef JUST3D
void            DrawScreenGL(unsigned int width, unsigned int height);
#endif

static void     SetActiveTmemTexture(int tile, int mipmap, int width, int height, int xo, int yo);
static void     LoadVisualBlock();
static void     LoadVisualTile();
static uBYTE    *DecompressTexture(t_td *tmpTile);
static int      TestPower2(int width, int height, int *nw, int *nh);

static void     CopyTextureRGBA32(uBYTE *src, uBYTE *dst, uWORD width, uWORD height);
static void     CopyTextureRGBA16(uBYTE *src, uBYTE *dst, uWORD width, uWORD height);
static void     CopyTextureYUV   (uBYTE *src, uBYTE *dst, uWORD width, uWORD height);
static void     CopyTextureCI8   (uBYTE *src, uBYTE *dst, uWORD width, uWORD height);
static void     CopyTextureCI4   (uBYTE *src, uBYTE *dst, uWORD width, uWORD height);
static void     CopyTextureIA16  (uBYTE *src, uBYTE *dst, uWORD width, uWORD height);
static void     CopyTextureIA8   (uBYTE *src, uBYTE *dst, uWORD width, uWORD height);
static void     CopyTextureI16   (uBYTE *src, uBYTE *dst, uWORD width, uWORD height);
static void     CopyTextureI8    (uBYTE *src, uBYTE *dst, uWORD width, uWORD height);
static void     CopyTextureI4    (uBYTE *src, uBYTE *dst, uWORD width, uWORD height);





/* variables */

#ifdef CACHE_TEXTURES
static uWORD            num_cached_texture = 1;
#endif
static unsigned char    tbuffer[4096*4*2];
static unsigned char    tbuffer2[256*256*4];



/* precalculated projection matrixes for use in OpenGL display lists - from gil */

const float proj_matrix[2][4][4] =
{
    /* 320 * 240 */
        {
                { 0.0062500, 0.0000000, 0.0000000, 0.0000000},
                { 0.0000000,-0.0083333, 0.0000000, 0.0000000},
                { 0.0000000, 0.0000000,-1.0000000, 0.0000000},
                {-1.0000000, 1.0000000, 0.0000000, 1.0000000}
        },
    /* 640 * 480 */
        {
                { 0.0031250, 0.0000000, 0.0000000, 0.0000000},
                { 0.0000000,-0.0041667, 0.0000000, 0.0000000},
                { 0.0000000, 0.0000000,-1.0000000, 0.0000000},
                {-1.0000000, 1.0000000, 0.0000000, 1.0000000}
        }
};










/*****************************************************************************/
/* here we start with real stuff                                             */
/*****************************************************************************/





void CreateDisplayLists()
{
    /* Gil: Some precompiled display lists that will make
            the emulation slighty faster. 
            I think we shold use more of these, if possible. */
    /* niki: gil - this is a fantastic idea!!! */

    /* This routine should be called when you init your opengl graphics */
       
    /* BEGIN_2D_320 */
        glNewList(kListBegin2D320, GL_COMPILE);
            glPushAttrib(GL_ENABLE_BIT);
        
            glDisable(GL_DEPTH_TEST);
            glDisable(GL_CULL_FACE);
            glDisable(GL_ALPHA_TEST);
            glEnable(GL_BLEND);

            glMatrixMode(GL_PROJECTION);
            glPushMatrix();
               /* I use precalculated matrixes for speed */
            glLoadMatrixf((GLfloat *)(&(proj_matrix[0][0][0])));
            glMatrixMode(GL_MODELVIEW);
            glPushMatrix();
            glLoadIdentity();
        glEndList();

   /* BEGIN_2D_640 */
        glNewList(kListBegin2D640, GL_COMPILE);
            glPushAttrib(GL_ENABLE_BIT);

            glDisable(GL_CULL_FACE);
            glDisable(GL_DEPTH_TEST);
            glDisable(GL_ALPHA_TEST);
            glEnable(GL_BLEND);

            glMatrixMode(GL_PROJECTION);
            glPushMatrix();
               /* I use precalculated matrixes for speed */
            glLoadMatrixf((GLfloat *)(&(proj_matrix[1][0][0])));
            glMatrixMode(GL_MODELVIEW);
            glPushMatrix();
            glLoadIdentity();
        glEndList();

    /* END_2D */
        glNewList(kListEnd2D, GL_COMPILE);
            glPopMatrix();
            glMatrixMode(GL_PROJECTION);
            glPopMatrix();
            glMatrixMode(GL_MODELVIEW);
            glPopAttrib();
        glEndList();
}












void Visual3D_FPS()
{
#ifdef FPS

#ifdef LINUX
    /* FPS - I don't know if this works on other unixes than linux - please mail! */
        static char             fps_text[40];
        int                     fps_update_sec = 1;   /* min update frequency in sec */
                                                      /* otherwise you might not read it properly */
        static long             fps_update = 0;   /* stores the next time for update */
        static int              fps_count = 0;   /* to make fps average */
        static float            fps_rate = 0.0;
        static struct timeval   fps_tv_old  = { 0, 0 };
        struct timeval          fps_tv;

        int len, i;



        gettimeofday(&fps_tv, NULL);

        fps_rate = fps_count * fps_rate
                   + 1 / ( ((float)fps_tv.tv_sec     + (float)fps_tv.tv_usec     / 1000000)
                         - ((float)fps_tv_old.tv_sec + (float)fps_tv_old.tv_usec / 1000000) );

        fps_count++;

        fps_rate /= fps_count;

        if(fps_update <= fps_tv.tv_sec)
        {
                fps_update = fps_tv.tv_sec + fps_update_sec;
                fps_count  = 0;

                if(fps_rate > 99.9)
                        fps_rate = 99.99;

                sprintf(fps_text, "%05.2f", fps_rate);

                fps_rate = 0.0;
        }

        fps_tv_old.tv_sec = fps_tv.tv_sec;
        fps_tv_old.tv_usec = fps_tv.tv_usec;

        glCallList(rdp_reg.colorimg_width);

        glColor3f(0.0f, 0.0f, 0.0f);
        glRasterPos2i(11, 11);
        len = (int)strlen(fps_text);
        for(i=0; i<len; i++)
                glutBitmapCharacter(GLUT_BITMAP_HELVETICA_10, fps_text[i]);
        glColor3f(1.0f, 1.0f, 1.0f);
        glRasterPos2i(10, 10);
        len = (int)strlen(fps_text);
        for(i=0; i<len; i++)
                glutBitmapCharacter(GLUT_BITMAP_HELVETICA_10, fps_text[i]);

        glCallList(kListEnd2D);

#endif /* #ifdef LINUX */

#ifdef AMIGADOS
        static char             fps_text[40];
        int                     fps_update_sec = 1;   /* min update frequency in sec */
                                                      /* otherwise you might not read it properly */
        static long             fps_update = 0;   /* stores the next time for update */
        static int              fps_count = 0;   /* to make fps average */
        static float            fps_rate = 0.0;
        static struct timeval   fps_tv_old  = { 0, 0 };
        struct timeval          fps_tv;

        GetSysTimePPC(&fps_tv);

        fps_rate = fps_count * fps_rate
                   + 1 / ( ((float)fps_tv.tv_sec     + (float)fps_tv.tv_usec     / 1000000)
                         - ((float)fps_tv_old.tv_sec + (float)fps_tv_old.tv_usec / 1000000) );

        fps_count++;

        fps_rate /= fps_count;

        if(fps_update <= fps_tv.tv_sec)
        {
                fps_update = fps_tv.tv_sec + fps_update_sec;
                fps_count  = 0;

                if(fps_rate > 99.9)
                        fps_rate = 99.99;

                sprintf(fps_text, "%05.2f", fps_rate);

                fps_rate = 0.0;
        }

        fps_tv_old.tv_sec = fps_tv.tv_sec;
        fps_tv_old.tv_usec = fps_tv.tv_usec;

        PrintText(fps_text, 40, 20);

#endif /* #ifdef AMIGADOS */

#ifdef MACOS
       static const Rect eraseRect = {0, 0, 14, 46};
       if (fpsCounter)
       {
           static int              fps_count = 1;    /* to make fps average */
               static long long                fps_target = 0, fps_last = 0;
               long long                               fps_current;
           float                   fps_rate;
           Fixed                                       fix_rate;

               static char status[256];
               
               Microseconds( (UnsignedWide*) &fps_current );
               
               fps_count++;
               
           if(fps_target <= fps_current)
           {
                   fps_rate = (float)fps_count/((fps_current-fps_last)/1000000.0);

               fps_target = fps_current + 1000000;

#ifdef FRAMESKIP
                       fix_rate = X2Fix(fps_rate*(prefs.frames_to_skip + 1));
#else
                       fix_rate = X2Fix(fps_rate);
#endif
                       
                       sprintf( status, "FPS: %u.%1.1u", fix_rate>>16, (fix_rate & 0xffff) / (0xffff/10));

               fps_count  = 0;
               fps_last = fps_current;
           }
               
               SetPort((GrafPtr)theVideoBuffer);
               ForeColor( whiteColor );
               BackColor( blackColor );
               TextFont( kFontIDGeneva );
               TextSize( 9 );
               EraseRect( &eraseRect );
               MoveTo( 0, 10 );
               DrawText( status, 0, strlen( status ) );
               ForeColor( blackColor );
               BackColor( whiteColor );
       }
#endif /* #ifdef MACOS */

#endif /* #ifdef FPS */

}













void tr_ren_load_modelview(float m[4][4])
{
        glMatrixMode(GL_MODELVIEW);
        glLoadMatrixf((GLfloat *)m);
}

void tr_ren_load_projection(float m[4][4])
{
        glMatrixMode(GL_PROJECTION);
        glLoadMatrixf((GLfloat *)m);
        glMatrixMode(GL_MODELVIEW);
}

void tr_ren_mul_modelview(float m[4][4])
{
        glMatrixMode(GL_MODELVIEW);
        glMultMatrixf((GLfloat *)m);
}

void tr_ren_mul_projection(float m[4][4])
{
        glMatrixMode(GL_PROJECTION);
        glMultMatrixf((GLfloat *)m);
        glMatrixMode(GL_MODELVIEW);
}

void tr_ren_push_load_modelview(float m[4][4])
{
        glMatrixMode(GL_MODELVIEW);
        glPushMatrix();
        glLoadMatrixf((GLfloat *)m);
}

void tr_ren_push_mul_modelview(float m[4][4])
{
        glMatrixMode(GL_MODELVIEW);
        glPushMatrix();
        glMultMatrixf((GLfloat *)m);
}

void tr_ren_pop_modelview()
{
        glMatrixMode(GL_MODELVIEW);
        glPopMatrix();
}

void tr_ren_viewport()
{
}

void tr_ren_lookat_x()
{
}

void tr_ren_lookat_y()
{
}

void tr_ren_light(int n)
{
        glLightModeli(GL_LIGHT_MODEL_LOCAL_VIEWER, GL_TRUE);
        if(rdp_reg.lights == n)
        {
                float col[4] = { 1.0f, 1.0f, 1.0f, 1.0f };
            /* last light is the ambient light */
                //glMaterialf(GL_LIGHT0 + n, GL_AMBIENT, 1.0f);
                //glLightfv(GL_LIGHT0 + n, GL_POSITION, pos);
                //glLightModelfv(GL_LIGHT_MODEL_AMBIENT, col);
                //glLightModelfv(GL_LIGHT_MODEL_AMBIENT, &(rdp_reg.light[n].r));
                glLightfv(GL_LIGHT0 + n, GL_AMBIENT, &(rdp_reg.light[n].r));
                //puts("...ambient");

        }
        else
        {
            /* the first lights are diffuse lights */
                //glMaterialf(GL_LIGHT0 + n, GL_DIFFUSE, 1.0f);
                glLightfv(GL_LIGHT0 + n, GL_SPOT_DIRECTION, &(rdp_reg.light[n].x));
                glLightfv(GL_LIGHT0 + n, GL_AMBIENT, &(rdp_reg.light[n].r));
                glLightfv(GL_LIGHT0 + n, GL_DIFFUSE, &(rdp_reg.light[n].r));
                //glLightfv(GL_LIGHT0 + n, GL_SPECULAR, &(rdp_reg.light[n].r));
                //puts("...diffuse");
        }
}

void tr_ren_numlights()
{
        int n;

        for(n=0; n<=rdp_reg.lights; n++)
        {
                glEnable(GL_LIGHT0 + n);
        }
        for(; n<=7; n++)
        {
                glDisable(GL_LIGHT0 + n);
        }
}

void tr_ren_lightcol(int n)
{
        glLightModeli(GL_LIGHT_MODEL_LOCAL_VIEWER, GL_TRUE);
        if(rdp_reg.lights == n)
        {
                float col[4] = { 1.0f, 1.0f, 1.0f, 1.0f };
            /* last light is the ambient light */
                //glMaterialf(GL_LIGHT0 + n, GL_AMBIENT, 1.0f);
                //glLightfv(GL_LIGHT0 + n, GL_POSITION, pos);
                //glLightModelfv(GL_LIGHT_MODEL_AMBIENT, col);
                //glLightModelfv(GL_LIGHT_MODEL_AMBIENT, &(rdp_reg.light[i].r));
                glLightfv(GL_LIGHT0 + n, GL_AMBIENT, &(rdp_reg.light[n].r));
                puts("...ambient");

        }
        else
        {
            /* the first lights are diffuse lights */
                //glMaterialf(GL_LIGHT0 + i, GL_DIFFUSE, 1.0f);
                glLightfv(GL_LIGHT0 + n, GL_SPOT_DIRECTION, &(rdp_reg.light[n].x));
                glLightfv(GL_LIGHT0 + n, GL_AMBIENT, &(rdp_reg.light[n].r));
                glLightfv(GL_LIGHT0 + n, GL_DIFFUSE, &(rdp_reg.light[n].r));
                //glLightfv(GL_LIGHT0 + n, GL_SPECULAR, &(rdp_reg.light[n].r));
                puts("...diffuse");
        }
}

void tr_ren_line3d(int vn[])
{
        int     i;
        float   v[4];

        glBegin(GL_LINES);

        for(i=0; i<2; i++)
        {
                v[0] = rdp_reg.vtx[vn[i]].x;   /* * perspective_normal; */
                v[1] = rdp_reg.vtx[vn[i]].y;   /* * perspective_normal; */
                v[2] = rdp_reg.vtx[vn[i]].z;   /* * perspective_normal; */
                v[3] = 1.0;   /* * perspective_normal; */

                glColor4ub(0x00, 0x99, 0x00, 0xff);
                glVertex3f(v[0],
                        v[1],
                        v[2]
                        );
        }
        glEnd();
}

void tr_ren_tri1(int vn[])
{
        DrawVisualTexTriangle(vn);
}


void tr_ren_tri2(int vn[])
{
        DrawVisualTexTriangle(vn);
        DrawVisualTexTriangle(vn+3);
}

void tr_ren_geometry_zbuffer(int on)
{
        if(on)
                glEnable(GL_DEPTH_TEST);
        else
                glDisable(GL_DEPTH_TEST);
}

void tr_ren_geometry_cullfront(int on)
{
        if(on)
        {
                glCullFace(GL_FRONT);
                glEnable(GL_CULL_FACE);
        }
        else
        {
                glDisable(GL_CULL_FACE);
        }
}

void tr_ren_geometry_cullback(int on)
{
        if(on)
        {
                glCullFace(GL_BACK);
                glEnable(GL_CULL_FACE);
        }
        else
        {
                glDisable(GL_CULL_FACE);
        }
}

void tr_ren_geometry_cullfrontback(int on)
{
        if(on)
        {
                glCullFace(GL_FRONT_AND_BACK);
                glEnable(GL_CULL_FACE);
        }
        else
        {
                glDisable(GL_CULL_FACE);
        }
}
















void SetVisualColor(int color)
{
#ifndef MACOS

        glColor3ub((color >> 24) & 0xff,
                (color >> 16) & 0xff,
                (color >>  8) & 0xff);
#endif

}





void DrawVisualFillRect(int ulx, int uly, int lrx, int lry)
{
} /* void DrawVisualFillRect() */





void ClearVisual()
{
        glClearColor(0.0, 0.0, 0.0, 1.0);
        glClearDepth(1.0);
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

} /* void ClearVisual() */





void DrawVisualLine(int ulx, int uly, int lrx, int lry)
{
        glCallList(rdp_reg.colorimg_width);
        glBegin(GL_LINE_STRIP);
        glColor3f(0.25f, 0.25f, 1.0f);
        glVertex2f(ulx, uly);
        glVertex2f(lrx, lry);
        glEnd();
        glCallList(kListEnd2D);

} /* void DrawVisualLine(int ulx, int uly, int lrx, int lry) */





void DrawVisualTriangle(int x1, int y1, int x2, int y2, int x3, int y3)
{
    glCallList(rdp_reg.colorimg_width);
        glBegin(GL_LINE_STRIP);

        glColor3f(1.0f, 1.0f, .25f);
        glVertex2f(x1, y1);
        glVertex2f(x2, y2);
        glVertex2f(x3, y3);
        glVertex2f(x1, y1);

        glEnd();
    glCallList(kListEnd2D);

} /* void DrawVisualTriangle(int x1, int y1, int x2, int y2, int x3, int y3) */





void DrawVisualTexTriangle(int vn[3])
{
    /* from FiRES - modified by niki and gil */
        uBYTE flag;
        int i;
        int width  = rdp_reg.td[rdp_reg.tile].sts_lrs - rdp_reg.td[rdp_reg.tile].sts_uls + 1;
        int height = rdp_reg.td[rdp_reg.tile].sts_lrt - rdp_reg.td[rdp_reg.tile].sts_ult + 1;



    /* lightning - originaly from icepir8 - modified by me (niki) */
        if(rdp_reg.geometrymode & 0x20000)
        {
/*
                int i;
                for(i = 0; i < (rdp_reg.lights - 1); i++)
                {
                        //glMaterialf(GL_LIGHT0 + i, GL_DIFFUSE, 1.0f);
                        glLightfv(GL_LIGHT0 + i, GL_POSITION, &(rdp_reg.light[i].x));
                        glLightfv(GL_LIGHT0 + i, GL_DIFFUSE, &(rdp_reg.light[i].r));
                        glEnable(GL_LIGHT0 + i);
                        //putchar(':');
                }
*/
                //glMaterialf(GL_LIGHT0 + i, GL_AMBIENT, 1.0f);
                //glLightfv(GL_LIGHT0 + i, GL_AMBIENT, &(rdp_reg.light[i].r));
                //glEnable(GL_LIGHT0 + i);
                glEnable(GL_LIGHTING);
                //glEnable(GL_COLOR_MATERIAL);
                //putchar(';');
        }
        else
        {
                glDisable(GL_LIGHTING);
                //glDisable(GL_COLOR_MATERIAL);
        }

    /* if i do it like this then ultra1 does not work!!!
       so it must be wrong in my opinion - niki
    */
    /*
        SetActiveTmemTexture(rdp_reg.tile, rdp_reg.texture[rdp_reg.tile].mipmap, width, height, 0, 0);
        if(rdp_reg.geometrymode_textures == 1)

        if( (rdp_reg.mode_h & 0x00300000) != 0x00300000)  // not fill mode
    */

    
        //if( (rdp_reg.texture[rdp_reg.tile].on == 1) )
        //if( 0 )
        if(rdp_reg.do_textures)
        {
                SetActiveTmemTexture(rdp_reg.tile, rdp_reg.texture[rdp_reg.tile].mipmap_level, width, height, 0, 0);

                glEnable(GL_TEXTURE_2D);
                glBegin(GL_TRIANGLES);
                glColor3ub(0xff, 0xff, 0xff);       /* Textures are in modulate-mode... */
                                                    /* NOT ALL! - niki - to be implemented... */

                for(i=0; i<3; i++)
                {
                        t_vtx *vertex = &rdp_reg.vtx[vn[i]];
                        float s = (float)(vertex->s * (rdp_reg.texture[rdp_reg.tile].scale_s));
                        float t = (float)(vertex->t * (rdp_reg.texture[rdp_reg.tile].scale_t));
                        //if(rdp_reg.mode_h & 0x00200000)   /* if we are in copy or fill mode */
                        //        glTexCoord2f(s/32.0/4.0, t/32.0f);
                        //else
                                glTexCoord2f(s/32.0f, t/32.0f);

                        //glColor4ub(vertex->r, vertex->g, vertex->b, vertex->a);
                        glNormal3b(vertex->r, vertex->g, vertex->b);
                        glVertex3f(
                                (float)vertex->x,
                                (float)vertex->y,
                                (float)vertex->z
                                );

                }
                glEnd();

                /* Gil: Textures should only be enabled when they are needed.
                        Otherwise we get into trouble. */
                glDisable(GL_TEXTURE_2D);
        }
        else
        {
                /* Gil: No need to disable textures here, because
                        they should be disabled at all times */
                glDisable(GL_TEXTURE_2D);

                glBegin(GL_TRIANGLES);

                switch(rdp_reg.geometrymode & 0x00000204)
                {
                    case 0x000:   /* No shading. Use the primcolor. */
                    puts("primcolor shading");
                        glColor4ub(
                                rdp_reg.primcolor >> 24,
                                rdp_reg.primcolor >> 16,
                                rdp_reg.primcolor >>  8,
                                rdp_reg.primcolor >>  0
                                );
                        for(i=0; i<3; i++)
                        {
                                t_vtx *vertex = &rdp_reg.vtx[vn[i]];
                                glVertex3f(
                                        (float)vertex->x,
                                        (float)vertex->y,
                                        (float)vertex->z
                                        );
                        }
                        break;

                    case 0x004:   /* Simple shading only. Use the color from vn[flag] */
                    puts("flag shading");
                        flag = (rdp_reg.cmd1 >> 24) & 0xff;
                        glColor4ub(
                                rdp_reg.vtx[vn[flag]].r,
                                rdp_reg.vtx[vn[flag]].g,
                                rdp_reg.vtx[vn[flag]].b,
                                rdp_reg.vtx[vn[flag]].a
                                );
                        for(i=0; i<3; i++)
                        {
                                t_vtx *vertex = &rdp_reg.vtx[vn[i]];
                                glVertex3f(
                                        (float)vertex->x,
                                        (float)vertex->y,
                                        (float)vertex->z
                                        );
                        }
                        break;

                    case 0x200:   /* Gouraud shading without shading. */
                        puts("textri: this cannot happen");
                        break;   /* Shouldn't be possible */

                    case 0x204:   /* Gouraud shading */
                        for(i=0; i<3; i++)
                        {
                                t_vtx *vertex = &rdp_reg.vtx[vn[i]];
                                glColor4ub(vertex->r, vertex->g, vertex->b, vertex->a);
                                glVertex3f(
                                        (float)vertex->x,
                                        (float)vertex->y,
                                        (float)vertex->z
                                        );
                        }
                        break;

                } /* switch(rdp_reg.geometrymode & 0x00000204) */
                glEnd();
        }


    /* we switch off lighting for other stuff */
        if(rdp_reg.geometrymode & 0x20000)
        {
                //glDisable(GL_COLOR_MATERIAL);
                glDisable(GL_LIGHTING);
        }

/* debug in red color */
///*  
        glColor3f(1.0f, 0.25f, 0.25f);
            
        glBegin(GL_LINE_STRIP);
            for(i=0; i<=3; i++)
            {
                    t_vtx *vertex = &rdp_reg.vtx[vn[i%3]];

                    //glColor4ub(vertex->r, vertex->g, vertex->b, vertex->a);
                    glVertex3f(
                            (float)vertex->x,
                            (float)vertex->y,
                            (float)vertex->z
                            );
            }
        glEnd();
//*/

} /* void DrawVisualTexTriangle(int vn[3]) */





void DrawVisualTexRectangle(
        int ulx, int uly,
        int lrx, int lry,
        int tile,
        float s, float t, float dsdx, float dtdy
        )
{
    /* from FiRES - modified by niki and gil */
        int w;
        int h;
        float sw;
        float sh;



        glCallList(rdp_reg.colorimg_width);

        w = rdp_reg.td[tile].sts_lrs - rdp_reg.td[tile].sts_uls + 1;
        h = rdp_reg.td[tile].sts_lrt - rdp_reg.td[tile].sts_ult + 1;
        sw = lrx - ulx;
        sh = lry - uly;



        SetActiveTmemTexture(tile, 0, w, h, s, t);


        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);

        glColor3ub(0xff, 0xff, 0xff);       /* Textures are in modulate-mode... */
        glEnable(GL_TEXTURE_2D);

        glBegin(GL_QUADS);
            glTexCoord2f(dsdx*(s/w + sw/w), dtdy*(t/h + sh/h)); glVertex2i(lrx, lry);
            glTexCoord2f(dsdx*(s/w +    0), dtdy*(t/h + sh/h)); glVertex2i(ulx, lry);
            glTexCoord2f(dsdx*(s/w +    0), dtdy*(t/h +    0)); glVertex2i(ulx, uly);
            glTexCoord2f(dsdx*(s/w + sw/w), dtdy*(t/h +    0)); glVertex2i(lrx, uly);
        glEnd();

        glDisable(GL_TEXTURE_2D);
        glCallList(kListEnd2D);

/* debug in blue color */
///*  
        SetVisualColor(0x3f3fcfff);
        DrawVisualRectangle(ulx, uly, lrx, lry);
//*/

} /* void DrawVisualTexRectangle(?) */





void DrawVisualRectangle(int ulx, int uly, int lrx, int lry)
{
        glCallList(rdp_reg.colorimg_width);
            glBegin(GL_LINE_STRIP);
                glVertex2f(ulx, uly);
                glVertex2f(lrx, uly);
                glVertex2f(lrx, lry);
                glVertex2f(ulx, lry);
                glVertex2f(ulx, uly);
           glEnd();
        glCallList(kListEnd2D);

} /* void DrawVisualRectangle(int ulx, int uly, int lrx, int lry) */





void EnableVisualAlpha()
{
        glEnable(GL_BLEND);
}

void DisableVisualAlpha()
{
        glDisable(GL_BLEND);
}

void VisualFlat()
{
        glShadeModel(GL_FLAT);
}

void VisualSmooth()
{
        glShadeModel(GL_SMOOTH);
}


/**
 * The following routine is from Gil Pedersen.
 * Gil you did great work again! (niki)
 * Modified by me (niki) and icepir8.
**/

/* This will draw the current n64 frame buffer to the screen
   At the moment only 15 bit graphics at a resolution of 320*240 is supported.
   It probably only works right with big endian at the momment.
   This routine has been coded by Gil Pedersen */

void DrawScreenGL(unsigned int width, unsigned int height)
{
        //static unsigned char tbuffer[256*256*4];

        int             i, j;
        unsigned short  *src, *base_src_ptr;
        unsigned char   *dest;
//        int             new_width, new_height;
        GLint           ntex[1] = { 1+MAX_CACHED_TEXTURES };
        GLint           hadDepthTest, hadBlending;

      //  return;
puts("DrawScreenGL()");fflush(stdout);
        
        //glCallList(rdp_reg.colorimg_width);



        glGetIntegerv(GL_DEPTH_TEST, &hadDepthTest);
        glGetIntegerv(GL_BLEND, &hadBlending);
        glDisable(GL_DEPTH_TEST);
        glDisable(GL_BLEND);

        glMatrixMode(GL_PROJECTION);
        glPushMatrix();
        glLoadIdentity();
        glOrtho(0, width, height, 0, -1, 1);
        glViewport(0, 0, WIDTH, HEIGHT);
        glMatrixMode(GL_MODELVIEW);
        glPushMatrix();
        glLoadIdentity();

        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_LINEAR);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_LINEAR);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_REPEAT);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_REPEAT);
        glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_DECAL);

        //glColor3ub(0xff, 0xff, 0xff);       /* Textures are in modulate-mode... */
        glEnable(GL_TEXTURE_2D);

        //glPixelStorei(GL_UNPACK_ROW_LENGTH, 0);
        //glPixelStorei(GL_UNPACK_SKIP_PIXELS, 0);
        //glPixelStorei(GL_UNPACK_SKIP_ROWS, 0);
        //glPixelStorei(GL_UNPACK_ALIGNMENT, 4);
        
        //glDeleteTextures(1, ntex);
        //glBindTexture(GL_TEXTURE_2D, 1+MAX_CACHED_TEXTURES);

        base_src_ptr = (uHWORD *)(uWORD *)(mem.rd_ram + ((uWORD *)mem.vi_reg)[1]);
        dest = (unsigned char *)tbuffer2;

    /* 256x240 */
        for(i=0; i<240; i++)
        {
                for(j=0; j<256; j++)
                {
#ifdef ENDIAN_IS_LITTLE
                        src = base_src_ptr + i*width + (j^0x01);   /* thanks Icepir8 */
#endif
#ifdef ENDIAN_IS_BIG
                        src = base_src_ptr + i*width + j;
#endif
                        *(dest)++ = (((*src >> 11) & 0x001f)<<3);
                        *(dest)++ = (((*src >>  6) & 0x001f)<<3);
                        *(dest)++ = (((*src >>  1) & 0x001f)<<3);
                        *(dest)++ = 0xff;   /* alpha */
                        //*(dest)++ = (*src & 0x0001) ? 0xff : 0x00;   /* alpha */
                }
        }

/*
        TestPower2(width, height, &new_width, &new_height);

        gluScaleImage(
                GL_RGBA,
                width, height,
                GL_UNSIGNED_BYTE, (unsigned char*)base_src_ptr,
                new_width, new_height,
                GL_UNSIGNED_BYTE, (unsigned char*)tbuffer2
                );
*/
puts("DrawScreenGL() tex");fflush(stdout);
        glTexImage2D(
                GL_TEXTURE_2D, 0,
                GL_RGB,
                256, 256, 0,
                GL_RGBA,
                GL_UNSIGNED_BYTE,
                (unsigned char*)tbuffer2
                );

puts("DrawScreenGL() quads");fflush(stdout);

        //glCallList(rdp_reg.colorimg_width);

        glBegin(GL_QUADS);
/*
            glTexCoord2f(0.0, 1.0); glVertex2f(  0.0, 256.0);
            glTexCoord2f(0.0, 0.0); glVertex2f(  0.0,   0.0);
            glTexCoord2f(1.0, 1.0); glVertex2f(256.0, 256.0);
            glTexCoord2f(1.0, 0.0); glVertex2f(256.0,   0.0);
*/
            glTexCoord2f(0.0f, 0.0f); glVertex2i(  0,   0);
            glTexCoord2f(0.0f, 1.0f); glVertex2i(  0, 256);
            glTexCoord2f(1.0f, 1.0f); glVertex2i(256, 256);
            glTexCoord2f(1.0f, 0.0f); glVertex2i(256,   0);
        glEnd();

        glDisable(GL_TEXTURE_2D);

        glPopMatrix();
        glMatrixMode(GL_PROJECTION);
        glPopMatrix();
        if (hadBlending)
                glEnable(GL_BLEND);
        if (hadDepthTest)
                glEnable(GL_DEPTH_TEST);

        //glCallList(kListEnd2D);

        DrawVisualRectangle(0, 0, 239, 239);
        DrawVisualRectangle(240, 0, 319, 239);

puts("DrawScreenGL() end");fflush(stdout);

} /* void DrawScreenGL(unsigned int width, unsigned int height) */





static void SetActiveTmemTexture(int tile, int mipmap, int w, int h, int xo, int yo)
{
    /**
    * This is a modified version of FiRESs SetActiveTexture().
    * It was modified by me (niki).
    * I also added MIPMAP stuff.
    **/
        uWORD   *text    = &(((uWORD *)mem.rd_ram)[rdp_reg.textureimg_addr >> 2]);
        t_td    *td_p    = &(rdp_reg.td[tile]);
        t_td    *ctd_p   = &(rdp_reg.td[rdp_reg.loadtile]);
        uBYTE   *tex_ptr = NULL;
        int     i;
        static char *format[]   = { "RGBA", "YUV", "CI", "IA", "I", "?", "?", "?" };
        static char *size[]     = { "4bit", "8bit", "16bit", "32bit" };
        static char *cm[]       = { "NOMIRROR/WARP(NOCLAMP)", "MIRROR", "CLAMP", "MIRROR&CLAMP" };



        int   new_width, new_height;

    /* test, if texture is disabled */
    /* Gil: textured rectangles don't test for this,
            so it should not be here. */
    /* Niki: okay - let's put it into the tri routine! */
/*
        if(rdp_reg.texture[tile].on == 0)
                return(0);
*/


#ifdef CACHE_TEXTURES
    /* test, if texture is cached - great work FiRES ;-) */

        if( (text[0] == 0xff000047) && (text[1] == 0x7d534654) && (text[2] == 0x0000fe01) )
        {
                glBindTexture(GL_TEXTURE_2D, text[3]);
                //puts("<<<<<<<<<<<<<<<<< C A C H E >>>>>>>>>>>>>>>>>>>>>");
                return;
        }
#endif

        puts("<<<<<<<<<<<<<<<<< N O T   I N   C A C H E >>>>>>>>>>>>>>>>>>>>>");
        printf("SETTEXTUREIMG was: ");
        printf("format=%s; size=%s; width=%hu; addr=$%08lx\n",
               format[rdp_reg.textureimg_fmt],
               size[rdp_reg.textureimg_size],
               rdp_reg.textureimg_width,
               rdp_reg.textureimg_addr);
        printf("SETTILE load was:  ");
        printf(
                "format=%s; size=%s; line=%hu (pixel); tmem=%hu; tile=%u; palette=$%x; "
                "cmt=%s; maskt=%u; shiftt=%u; cms=%s; masks=%u; shifts=%u\n",
                format[ctd_p->format],
                size[ctd_p->size],
                ctd_p->line,
                ctd_p->tmem,
                tile,
                ctd_p->palette,
                cm[(ctd_p->clampt << 1) | ctd_p->mirrort],
                ctd_p->maskt,
                ctd_p->shiftt,
                cm[(ctd_p->clamps << 1) | ctd_p->mirrors],
                ctd_p->masks,
                ctd_p->shifts);
        printf("LOADTILESIZE was:  ");
        printf("uls=%d; ult=%d; lrs=%d; lrt=%d; dxt=%d tile=%d\n",
                ctd_p->uls,
                ctd_p->ult,
                ctd_p->lrs,
                ctd_p->lrt,
                ctd_p->dxt,
                tile
                );
        printf("SETTILE show was:  ");
        printf(
                "format=%s; size=%s; line=%hu (pixel); tmem=%hu; tile=%u; palette=$%x; "
                "cmt=%s; maskt=%u; shiftt=%u; cms=%s; masks=%u; shifts=%u\n",
                format[td_p->format],
                size[td_p->size],
                td_p->line,
                td_p->tmem,
                tile,
                td_p->palette,
                cm[(td_p->clampt << 1) | td_p->mirrort],
                td_p->maskt,
                td_p->shiftt,
                cm[(td_p->clamps << 1) | td_p->mirrors],
                td_p->masks,
                td_p->shifts);
        printf("SETTILESIZE show was:  ");
        printf("uls=%6.2f; ult=%6.2f; lrs=%6.2f; lrt=%6.2f; tile=%d\n",
                td_p->sts_uls,
                td_p->sts_ult,
                td_p->sts_lrs,
                td_p->sts_lrt,
                tile
                );




        switch(rdp_reg.td[rdp_reg.loadtile].set_by)
        {
            case RDPTD_LOADBLOCK:
                LoadVisualBlock();
                break;
            case RDPTD_LOADTILE:
                LoadVisualTile();
                break;
            case RDPTD_LOAD_NOT_SET:
                printf("RDP WARNING: TEXRECT/TRI without LOADBLOCK/-TILE\n");
                break;
            default:
                printf("this cannot happen\n");
        }


    /* texture is not cached yet... so we cache it */
#ifdef CACHE_TEXTURES
        glBindTexture(GL_TEXTURE_2D, num_cached_texture);
        text[0] = 0xff000047;
        text[1] = 0x7d534654;
        text[2] = 0x0000fe01;
        text[3] = num_cached_texture;
#else
        glBindTexture(GL_TEXTURE_2D, 1);
#endif




        if(td_p->clampt)
                glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
        else
                glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);

        if(td_p->clamps)
                glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
        else
                glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);

        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
        /* speed
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
        */

        glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

        for(i=0; i<=mipmap; i++)
        {
                tex_ptr = DecompressTexture(&(rdp_reg.td[rdp_reg.tile+i]));

                if(TestPower2(w, h, &new_width, &new_height))
                {
                        glTexImage2D(
                                GL_TEXTURE_2D, i,
                                GL_RGBA,
                                w, h, 0,
                                GL_RGBA,
                                GL_UNSIGNED_BYTE,
                                tex_ptr
                                );
                }
                else
                {
                        gluScaleImage(
                                GL_RGBA,
                                w, h,
                                GL_UNSIGNED_BYTE, (unsigned char*)tex_ptr,
                                new_width, new_height,
                                GL_UNSIGNED_BYTE, (unsigned char*)tbuffer2
                                );

                        glTexImage2D(
                                GL_TEXTURE_2D, i,
                                GL_RGBA,
                                new_width, new_height, 0,
                                GL_RGBA,
                                GL_UNSIGNED_BYTE,
                                tbuffer2
                                );
                }
putchar('.');fflush(stdout);
        }

#ifdef CACHE_TEXTURES
        num_cached_texture++;
        if(num_cached_texture >= MAX_CACHED_TEXTURES)
                num_cached_texture = 1;
#endif

        return;

} /* static void SetActiveTmemTexture(int tile, int mipmap, int w, int h, int xo, int yo) */





static void LoadVisualBlock()
{
        int   lwidth;    /* width of src scanline */
        int   o;         /* offset of src block */
        int   bytes;     /* src bytes */
        int   dst;
        int   i;
        t_td  *tmp_tile = &(rdp_reg.td[rdp_reg.loadtile]);


                                                
#define MAKE_BYTES(x) (((x) << tmp_tile->size) >> 1)   /* pixel2bytes: bytes = pixel * (1|2|4|8) / 2 */

        lwidth = (tmp_tile->dxt) << 3;                                  /* scanlinewidth in bytes */
        o      = MAKE_BYTES(tmp_tile->uls) + (tmp_tile->ult) * lwidth;
    /* it's possible that we don't have to add 1 here - niki */
        bytes  = MAKE_BYTES((tmp_tile->lrs) - (tmp_tile->uls) + 1);                 /* bytes = pixel * (1|2|4|8) / 2 */
        dst    = tmp_tile->tmem;

#undef MAKE_BYTES

/*
        printf("%d %d %d %d\n", tmp_tile->uls, tmp_tile->ult, tmp_tile->lrs, tmp_tile->dxt);
        printf("lwidth: %d\n", lwidth);
        printf("o     : %d\n", o     );
        printf("bytes:  %d\n", bytes);
        printf("tmem:   %d\n", tmp_tile->tmem);
*/

        o += tmp_tile->addr;

        /* we might have to convert to a usable format for the display driver here. */

        for(i=0; i<bytes; i++)
        {
#ifdef ENDIAN_IS_LITTLE
                //((uBYTE *)mem.tmem)[dst] = ((uBYTE *)mem.rd_ram)[(i+o)^3];
                /* strange, but this seems to be correct - niki */
                ((uBYTE *)mem.tmem)[dst] = ((uBYTE *)mem.rd_ram)[(i+o)];
#endif
#ifdef ENDIAN_IS_BIG
                ((uBYTE *)mem.tmem)[dst] = ((uBYTE *)mem.rd_ram)[i+o];
#endif
                dst++;
        }

} /* static void LoadVisualBlock(int tile) */





static void LoadVisualTile()
{
        puts("LoadVisualTile - TO BE IMPLEMENTED...");

} /* static void LoadVisualTile() */





static int TestPower2(int width, int height, int *nw, int *nh)
{
        int NewWidth = *nw;
        int NewHeight = *nh;
        int w;
        int temp;



    /* test if width has the power of 2 */
        temp = width;
        for(w = 11; w >= 0; w--)
        {
                if(temp & 0x800)
                {
                        NewWidth = 1 << w;
                        break;
                }
                temp <<= 1;
        }

    /* test if height has the power of 2 */
        temp = height;
        for(w = 11; w >= 0; w--)
        {
                if(temp & 0x800)
                {
                        NewHeight = 1 << w;
                        break;
                }
                temp <<= 1;
        }

    /* return */
        if( (NewWidth == width) && (NewHeight == height) )
                return(1);
        else
        {
                *nw = NewWidth;
                *nh = NewHeight;
                return(0);
        }

        return(1);

} /* static int TestPower2(int width, int height, int *nw, int *nh) */





static uBYTE *DecompressTexture(t_td *tmpTile)
{
        int         width  = tmpTile->sts_lrs - tmpTile->sts_uls + 1;
        int         height = tmpTile->sts_lrt - tmpTile->sts_ult + 1;
    /* niki
        uBYTE       size   = rdp_reg.textureimg_size;
        uBYTE       format = rdp_reg.textureimg_fmt;
    */
        uBYTE       size   = tmpTile->size;
        uBYTE       format = tmpTile->format;



    /* we could do this with a function pointer for more speed! */
        switch(format)
        {
            case 0:   /* rgba */
                switch(size)
                {
                    case 3:
                        CopyTextureRGBA32(mem.tmem + tmpTile->tmem, tbuffer, width, height);
                        return(mem.tmem + tmpTile->tmem);
                    case 2:
                        CopyTextureRGBA16(mem.tmem + tmpTile->tmem, tbuffer, width, height);
                        break;
                    default:
                        puts("RDP WARNING: textureimage format rgba: unknown size!");
                }
                break;

            case 1:   /* YUV */
                /* to be implemented */
                switch(size)
                {
                    default:
                        puts("RDP WARNING: textureimage format yuv: not correct implemented!");
                        CopyTextureYUV(mem.tmem + tmpTile->tmem, tbuffer, width, height);
                        break;
                }

            case 2:   /* CI */
                switch(size)
                {
                    case 1:
                        CopyTextureCI8(mem.tmem + tmpTile->tmem, tbuffer, width, height);
                        break;
                    case 0:  /* i think this is not used - niki */
                        CopyTextureCI4(mem.tmem + tmpTile->tmem, tbuffer, width, height);
                        break;
                    default:
                        puts("RDP WARNING: textureimage format ci: unknown size!");
                }
                break;

            case 3:   /* IA */
                switch(size)
                {
                    case 2:
                        CopyTextureIA16(mem.tmem + tmpTile->tmem, tbuffer, width, height);
                        break;
                    case 1:
                        CopyTextureIA8(mem.tmem + tmpTile->tmem, tbuffer, width, height);
                        break;
                    default:
                        puts("RDP WARNING: textureimage format ia: unknown size!");
                }
                break;

            case 4:   /* I */
                switch(size)
                {
                    case 2:
                        CopyTextureI16(mem.tmem + tmpTile->tmem, tbuffer, width, height);
                        break;
                    case 1:
                        CopyTextureI8(mem.tmem + tmpTile->tmem, tbuffer, width, height);
                        break;
                    case 0:
                        CopyTextureI4(mem.tmem + tmpTile->tmem, tbuffer, width, height);
                        break;
                    default:
                        puts("RDP WARNING: textureimage format i: unknown size!");
                }
                break;

            default:
                puts("RDP WARNING: texture format not known!");

        } /* switch(format) */

        return(tbuffer);

} /* static uBYTE *DecompressTexture(t_td *tmpTile) */





/*****************************************************************************/




static void CopyTextureRGBA32(uBYTE *src, uBYTE *dst, uWORD width, uWORD height)
{
        uWORD *d = (uWORD *)dst;
        uWORD *s = (uWORD *)src;
        uWORD x, y;

        for(y=0; y < height; y++)
                for(x=0; x < width; x++)
                {
                    /* hope that this works on little endians */
                        *d++ = *s++;
                }
}

static void CopyTextureRGBA16(uBYTE *src, uBYTE *dst, uWORD width, uWORD height)
{
        uBYTE   *d   = dst;
        uHWORD  *s   = (uHWORD*)src;
        uWORD   x, y;
        uHWORD  rgba;   /* color */

        for(y=0; y<height; y++)
                for(x=0; x<width; x++)
                {
#ifdef ENDIAN_IS_LITTLE
                        if((int)x & 1)
                                rgba = *(s-1);
                        else
                                rgba = *(s+1);
#endif
#ifdef ENDIAN_IS_BIG
                        rgba = *s; /* rgba_5_5_5_1 */
#endif

                   /* otherwise NOMM does not work. it uses 320x64 tiles! */

                        if((x+y*width) >= 2047)
                                break;

                        *d++ = ((rgba & 0xf800) >> 11) << 3;
                        *d++ = ((rgba & 0x07c0) >>  6) << 3;
                        *d++ = ((rgba & 0x003e) >>  1) << 3;
                        *d++ =  (rgba & 0x0001) ? 0xff : 0x00;
                        s++;
                }
}

static void CopyTextureCI8(uBYTE *src, uBYTE *dst, uWORD width, uWORD height)
{
   /* done by me (niki) - not tested */
        uBYTE   *d = dst;
        uBYTE   *s = src;
        int     x, y;
        uHWORD  *pal = (uHWORD *)(mem.rd_ram + rdp_reg.tlut_8_addr);   /* palette */
        uHWORD  rgba;   /* color */

        for(y = 0; y < height; y++)
                for(x = 0; x < width; x++)
                {
                    /* 0x800 might be not correct! */
                        rgba = pal[*s]; /* load rgba16 from ci8 */
                        *d++ = ((rgba & 0xf800) >> 11) << 3;
                        *d++ = ((rgba & 0x07c0) >>  6) << 3;
                        *d++ = ((rgba & 0x003e) >>  1) << 3;
                        *d++ =  (rgba & 0x0001) ? 0xff : 0x00;
                        s++;
                }
}

static void CopyTextureCI4(uBYTE *src, uBYTE *dst, uWORD width, uWORD height)
{
    uWORD x, y;
    uBYTE *d = dst;
    uBYTE *s = src;
    uHWORD color;
    uHWORD *palette = (uHWORD*)(mem.rd_ram + (int)rdp_reg.tlut_4_addr);
    uBYTE index;

puts("4bit ci texture - i thought that this is impossible - niki");

    for(y=0; y < height; y++)
        for(x=0; x < (width>>1); x++)
        {
            s = src + ((y * width) + (x ^ 0x03));
            index = *s;

            color = palette[index >> 4];
            *d++ = (((color >> 11 ) & 0x001f)<<3);
            *d++ = (((color >>  6 ) & 0x001f)<<3);
            *d++ = (((color >>  1 ) & 0x001f)<<3);
            *d++ =  ((color & 0x01) * 0xff);

            color = palette[index & 0x0f];
            *d++ = (((color >> 11 ) & 0x001f)<<3);
            *d++ = (((color >>  6 ) & 0x001f)<<3);
            *d++ = (((color >>  1 ) & 0x001f)<<3);
            *d++ =  ((color & 0x01) * 0xff);
        }
}

static void CopyTextureIA16(uBYTE *src, uBYTE *dst, uWORD width, uWORD height)
{
        uBYTE   *d = dst;
        uHWORD  *s = (uHWORD *)src;
        uWORD   x, y;
        uHWORD  ia;

        for(y=0; y < height; y++)
                for(x=0; x < width; x++)
                {
#ifdef ENDIAN_IS_LITTLE
                        if((int)x & 1)
                                ia = *(s-1);
                        else
                                ia = *(s+1);
#endif
#ifdef ENDIAN_IS_BIG
                        ia = *s;
#endif
                        *d++ = ia >> 8;
                        *d++ = ia >> 8;
                        *d++ = ia >> 8;
                        *d++ = ia & 0xff;
                }
}

static void CopyTextureIA8(uBYTE *src, uBYTE *dst, uWORD width, uWORD height)
{
    /* mod by me (niki) */
        uBYTE   *d = dst;
        uBYTE   *s = src;
        uWORD   x, y;
        uBYTE   ia = 0; /* to make compiler happy */

        for(y=0; y < height; y++)
                for(x=0; x < width; x++)
                {
#ifdef ENDIAN_IS_LITTLE
                    /* this is fucking stupid (and maybe wrong)! */
                        switch((int)x & 3)
                        {
                            case 0: ia = *(s+3); break;
                            case 1: ia = *(s+2); break;
                            case 2: ia = *(s+1); break;
                            case 3: ia = *(s+0); break;
                        }
#endif
#ifdef ENDIAN_IS_BIG
                        ia = *s;
#endif
                        *d++ = ia & 0xf0;
                        *d++ = ia & 0xf0;
                        *d++ = ia & 0xf0;
                        *d++ = ia << 4;
                }
}

static void CopyTextureI16(uBYTE *src, uBYTE *dst, uWORD width, uWORD height)
{
    /* modified by me (niki, 19990615) */
        uBYTE  *d = dst;
        uHWORD *s;
        uWORD  x, y;

        for(y=0; y < height; y++)
                for(x=0; x < width; x++)
                {
#ifdef ENDIAN_IS_LITTLE
                        s = &(((uHWORD *)src)[y*width + (x^0x01)]);
#endif
#ifdef ENDIAN_IS_BIG
                        s = &(((uHWORD *)src)[y*width + x]);
#endif
                        *d++ = *s >> 8;
                        *d++ = *s >> 8;
                        *d++ = *s >> 8;
                        *d++ = *s >> 8;
                        s++;
                }
}

static void CopyTextureI8(uBYTE *src, uBYTE *dst, uWORD width, uWORD height)
{
    /* modified by me (niki, 19990615) */
        uWORD  x, y;
        uBYTE  *d = dst;
        uBYTE  *s;

        for(y=0; y < height; y++)
                for(x=0; x < width; x++)
                {
#ifdef ENDIAN_IS_LITTLE
                        s = src + y*width + (x^0x03);
#endif
#ifdef ENDIAN_IS_BIG
                        s = src + y*width + x;
#endif
                        *d++ = *s;
                        *d++ = *s;
                        *d++ = *s;
                        *d++ = *s;
                        s++;
                }
}

static void CopyTextureI4(uBYTE *src, uBYTE *dst, uWORD width, uWORD height)
{
    /* modified by me (niki, 19990615) */
        uWORD x, y;
        uBYTE *d = dst;
        uBYTE *s;
        uBYTE i;

        for(y=0; y < height; y++)
                for(x=0; x < (width>>1); x++)
                {
#ifdef ENDIAN_IS_LITTLE
                        s = src + y*(width>>1) + (x^0x03);
#endif
#ifdef ENDIAN_IS_BIG
                        s = src + y*(width>>1) + x;
#endif
                        i = (*s & 0xf0) | ((*s & 0xf0) >> 4);
                        *d++ = i;
                        *d++ = i;
                        *d++ = i;
                        *d++ = i;

                        i = (*s & 0x0f) | ((*s & 0x0f) << 4);
                        *d++ = i;
                        *d++ = i;
                        *d++ = i;
                        *d++ = i;

                        s++;
                }
}



/*****************************************************************************/
/*** the follow function is  not implemented yet *****************************/
/*****************************************************************************/

static void CopyTextureYUV(uBYTE *src, uBYTE *dst, uWORD width, uWORD height)
{
    uWORD i, j;
    uBYTE *d = dst;
    uHWORD *s = (uHWORD*)src;

    for(i=0; i < height; i++)
        for(j=0; j < width; j++)
        {
            *d++ = 100;
            *d++ = 150;
            *d++ = 200;
            *d++ = 255;
            s++;
        }

}

