// StandardDisplay.c

#include "display.h"
#include "config.h"
#include "parser_extern.h"
#include "memory_extern.h"
#include "type_sizes.h"
#include "dispatch.h"
#include <DrawSprocket.h>
#include "string.h"
#include <stdlib.h>
#include <Events.h>
#include <Palettes.h>
#include <LowMem.h>

#include "StandardDisplay.h"
//#include "RAVESupport.h"
#include "Blitters.h"
#include "ISp.h"
#include "MacOS.h"
#include "Alerts.h"
#include "warp_resolution_and_depth.h"

int StartDSp();
int EndDSp();
int StartWindowed();
int EndWindowed();


DSpContextReference 	theContext;
DSpContextAttributes	theDesiredAttributes;

CGrafPtr		theBackBuffer;
CGrafPtr		theScreenBuffer;
unsigned short	rowBytes;
Point			VidSize;
DSpBlitInfo		theBlit;
WindowPtr		theWindow=nil;
Boolean			gFullscreen=true;
Rect			windRect;
short 			*tmpVidBuffer=nil;
PixMapHandle	offscreen;				//A PixMap structure we use to make CopyBits know what it's doing.
static long		pixflags;
unsigned short	*dstPtr;

uWORD				width  =   0;   /* initial VI_WIDTH_REG (0x4400008) value */
uWORD				height =   0;   /* width*3/4 */
#undef DEPTH
int					DEPTH;
uWORD				flags  = 0;


int OpenVisual()	//The game is starting.  Show the window/grab the screen
{
	if(ISpResume() != 0)
	{
		MyStandardStopAlert(ISpStandardStopError);
		return -1;
	}
	ISpEnabled = true;
	
	if (gFullscreen)
		StartDSp();
	else
		StartWindowed();
		
	gRunning = true;
	
	return 0;
} /* int OpenVisual() */

int CloseVisual()	//Done with the game.  Go back to a waiting position.
{
	if (gFullscreen)
		EndDSp();
	else
		EndWindowed();
	
	ISpSuspend();
	ISpEnabled = false;
	gRunning = false;
	
	return(0);
} /* int CloseVisual() */

int InitDisplay()		//The first initialization done at startup
{
	OSStatus	theError;
	
	DEPTH = 16;
	prefs.display = LOCAL_DISPLAY;
	
	theError = DSpStartup();
	if (theError != noErr) return -1;
	
	//Set initial options.  Also blank out the reserveds/ignoreds
	theDesiredAttributes.reserved1 = 0;
	theDesiredAttributes.reserved2 = 0;
	theDesiredAttributes.reserved3[0] = 0;
	theDesiredAttributes.reserved3[1] = 0;
	theDesiredAttributes.reserved3[2] = 0;
	theDesiredAttributes.reserved3[3] = 0;
	theDesiredAttributes.gameMustConfirmSwitch = false;
	theDesiredAttributes.frequency = 0;
	theDesiredAttributes.colorNeeds = 0;
	theDesiredAttributes.colorTable = NULL;
	theDesiredAttributes.contextOptions = 0;
	theDesiredAttributes.displayWidth = MAX_WIDTH;
	theDesiredAttributes.displayHeight = MAX_HEIGHT;
	theDesiredAttributes.colorNeeds = kDSpColorNeeds_Request;
	theDesiredAttributes.pageCount = 1;

	if (DEPTH == 16)
	{
		theDesiredAttributes.backBufferDepthMask = kDSpDepthMask_16;
		theDesiredAttributes.backBufferBestDepth = 16;
		theDesiredAttributes.displayDepthMask = kDSpDepthMask_16;
		theDesiredAttributes.displayBestDepth = 16;
	}
	else if (DEPTH == 8)	// Not supported yet =(
	{
		theDesiredAttributes.backBufferDepthMask = kDSpDepthMask_8;
		theDesiredAttributes.backBufferBestDepth = 8;
		theDesiredAttributes.displayDepthMask = kDSpDepthMask_8;
		theDesiredAttributes.displayBestDepth = 8;
	}
	theError = DSpFindBestContext(&theDesiredAttributes, &theContext);	
	if (theError != noErr) return -1;
	
	// Enable pageflipping if available
	theDesiredAttributes.contextOptions |= kDSpContextOption_PageFlip;
	
	theError = DSpContext_Reserve(theContext, &theDesiredAttributes);
	if (theError != noErr) return -1;
		
	return 0;
}

int StartDSp()
{
	OSStatus	theError;
	
	theError = DSpContext_FadeGamma(theContext, 0, NULL);				// Gamma fade out
	if (theError != noErr)
	{
		MyStandardNumberStopAlert(DSpStandardNumberStopAlert, theError);
		return -1;
	}
	theError = DSpContext_SetState(theContext, kDSpContextState_Active);// Changes the resolution
	if (theError != noErr)
	{
		DSpContext_FadeGamma(theContext, 100, NULL);
		MyStandardNumberStopAlert(DSpStandardNumberStopAlert, theError);
		return -1;
	}
	HideCursor();
	theError = DSpContext_FadeGamma(theContext, 100, NULL);				// Gamma fade in
	if (theError != noErr)
		return -1;
	
	
	//SetPort((GrafPtr)theScreenBuffer); 	// set the buffer to be the current port
	
	DSpContext_SwapBuffers(theContext, nil, nil);	// Blank Screen
	theError = DSpContext_GetFrontBuffer(theContext, &theScreenBuffer);
	if (theError != noErr)
	{
		MyStandardNumberStopAlert(DSpStandardNumberStopAlert, theError);
		return -1;
	}
	theError = DSpContext_GetBackBuffer(theContext, kDSpBufferKind_Normal, &theBackBuffer);
	if (theError != noErr)
	{
		MyStandardNumberStopAlert(DSpStandardNumberStopAlert, theError);
		return -1;
	}
	
	// Get the address to write the graphics to
/*	theError = DSpContext_GetFrontBuffer(theContext, &theVideoBuffer);
	if (theError != noErr)
	{
		MyStandardNumberStopAlert(DSpStandardNumberStopAlert, theError);
		return -1;
	}*/
	
	rowBytes = (unsigned short)(**theScreenBuffer->portPixMap).rowBytes;
	rowBytes &= 0x3fff;
	dstPtr = (unsigned short *)(**theScreenBuffer->portPixMap).baseAddr;
	
	//PrepareRAVE();
	
	theBlit.completionFlag=true;
	theBlit.completionProc=NULL;
	theBlit.srcContext=theContext;
	theBlit.dstContext=theContext;
	theBlit.srcBuffer=theBackBuffer;
	theBlit.dstBuffer=theScreenBuffer;

	theBlit.srcRect.top=0;
	theBlit.srcRect.left=0;
	theBlit.srcRect.bottom=0;
	theBlit.srcRect.right=0;
	theBlit.dstRect.top=0;
	theBlit.dstRect.left=0;
	theBlit.dstRect.bottom=MAX_HEIGHT;
	theBlit.dstRect.right=MAX_WIDTH;

	theBlit.srcKey=0;
	theBlit.dstKey=0;
	theBlit.mode=kDSpBlitMode_Interpolation;//(DSpBlitMode)0;
	theBlit.reserved[0]=0;
	theBlit.reserved[1]=0;
	theBlit.reserved[2]=0;
	theBlit.reserved[3]=0;
	
	RefreshVisual();

	return 0;
}


void RefreshVisual()
{
	uWORD  new_width;

	/* video width */
	/* if width in VI_WIDTH_REG (0x4400008) has changed set windows width */
   
	if((new_width = ((uWORD *)mem.vi_reg)[2]) != width)
	{
		if(new_width < 320 || new_width > 640)
			width = 640;
		else
			width = new_width;

		height = (width * 3) >> 2;
		theBlit.srcRect.bottom=height;
		theBlit.srcRect.right=width;
	}
   
	/* video flags */   
	flags = ((uWORD *)mem.vi_reg)[0];
	switch(flags & 0x0003){
		case 0:			// Blank: No Video data or sync
		case 1:			// Reserved
			return;
			
		case 2:			// 15-Bit	
			if (gFullscreen)
			{
				if(width == 320)
					BlitLong16(rowBytes);
				else
					Blit16(rowBytes);
				//DrawRAVE(width, height);
			}
			else
			{
				render_image_16_to_15Mac((uHWORD *)(mem.rd_ram + ((uWORD *)mem.vi_reg)[1]), (uHWORD *)tmpVidBuffer, width, height);
				(**offscreen).pixelSize=16;
				(**offscreen).rowBytes=pixflags | (width*2);
				(**offscreen).cmpSize=5;
				BlitOffscreen();
			}
			break;
		case 3:			// 32-Bit
			if (gFullscreen)
			{
				if(width == 320)
					BlitDouble32(rowBytes);
				else
					Blit32(rowBytes);		
				//DrawRAVE(width, height);
			}
			else
			{
				(**offscreen).pixelSize=32;
				(**offscreen).rowBytes=pixflags | (width*4);
				(**offscreen).cmpSize=8;
				BlitOffscreen();
			}
			break;
	}
}

unsigned long CheckVisual()
{
	return ProcessKeys();
}

int EndDSp()
{
	if(!gRunning)
		return -1;
	
	//FinishRAVE();
	
	DSpContext_FadeGamma(theContext, 0, NULL);					// Gamma fade out
	DSpContext_SetState(theContext, kDSpContextState_Inactive);	// Changes the resolution
	ShowCursor();
	DSpContext_FadeGamma(theContext, 100, NULL);				// Gamma fade in
	return 0;
	
}

// Not used
void GetVisualName(char *, int *)
{}

int StartWindowed()
{
	tmpVidBuffer=(short *)malloc(320*240*2);
	theWindow=NewCWindow(nil, &windRect, "\pMac TrueReality", true, documentProc, WindowPtr(-1L), true, nil);	//Create a hidden window, wait until StartWindowed to show

	offscreen=NewPixMap();
	LockPixels(offscreen);
	(**offscreen).bounds.left=0;
	(**offscreen).bounds.top=0;
	(**offscreen).bounds.bottom=height;
	(**offscreen).bounds.right=width;
	pixflags=(**offscreen).rowBytes & 0xC000;		//Save the flags, since we will be messing with rowBytes.
	(**offscreen).rowBytes=pixflags | (width*2);
	(**offscreen).pixelType=RGBDirect;
	(**offscreen).cmpCount=3;
	(**offscreen).baseAddr=(char *)tmpVidBuffer;
	RefreshVisual();
	
	return 0;
}

int EndWindowed()
{
	DisposePixMap(offscreen);
	DisposeWindow(theWindow);
	
	return 0;
}

