/**
 * TrueReality - output/display/sw_renderer.h
 * Copyright (C) 1998, 1999 by Niki W. Waibel
 *
 * This program is free software; you can redistribute it and/
 * or modify it under the terms of the GNU General Public Li-
 * cence as published by the Free Software Foundation; either
 * version 2 of the Licence, or any later version.
 *
 * This program is distributed in the hope that it will be use-
 * ful, but WITHOUT ANY WARRANTY; without even the implied war-
 * ranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public Licence for more details.
 *
 * You should have received a copy of the GNU General Public
 * Licence along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 * USA.
 *
 * Information about me (the author):
 *   Niki W. Waibel, Reichenau 20, 6890 Lustenau, Austria - EUROPE
 *   niki.waibel@gmx.net
**/










/* prototypes */

static void vertex_mult(float r[4], float v[4], float m[4][4]);
static void matrix_mult(float r[4][4], float m1[4][4], float m2[4][4]);





/* variables */

static float mtx_proj[4][4];            /* 1 projection matrix */
static float mtx_model[10][4][4];       /* 10 modelview matrices */










/* real stuff! */





void tr_ren_load_modelview(float m[4][4])
{
    /**
     * we can use memcopy here because we have
     * native float values here -> endian does not matter
    **/
        memcpy(mtx_model[0], m, sizeof(mtx_model[0]));
}

void tr_ren_load_projection(float **m)
{
        memcpy(mtx_proj, m, sizeof(mtx_proj));
}

void tr_ren_mul_modelview(float m[4][4])
{
        int     i;
        float   mtx_copy[4][4];

        for(i = 0; i < 16; i++)
        {
                ((float *)mtx_copy)[i] = ((float *)(mtx_model[0]))[i];
        }
        matrix_mult(mtx_model[0], m, mtx_copy);   /* thanx Gil - i mixed m and mtx_copy! */
}

void tr_ren_mul_projection(float m[4][4])
{
        int     i;
        float   mtx_copy[4][4];

        for(i = 0; i < 16; i++)
        {
                ((float *)mtx_copy)[i] = ((float *)(mtx_model[0]))[i];
        }
        matrix_mult(mtx_proj[0], m, mtx_copy);   /* thanx Gil - i mixed m and mtx_copy! */
}

void tr_ren_push_load_modelview(float m[4][4])
{
        memcpy(mtx_model[1], mtx_model[0], sizeof(mtx_model[0]));
        memcpy(mtx_model[2], mtx_model[1], sizeof(mtx_model[0]));
        memcpy(mtx_model[3], mtx_model[2], sizeof(mtx_model[0]));
        memcpy(mtx_model[4], mtx_model[3], sizeof(mtx_model[0]));
        memcpy(mtx_model[5], mtx_model[4], sizeof(mtx_model[0]));
        memcpy(mtx_model[6], mtx_model[5], sizeof(mtx_model[0]));
        memcpy(mtx_model[7], mtx_model[6], sizeof(mtx_model[0]));
        memcpy(mtx_model[8], mtx_model[7], sizeof(mtx_model[0]));
        memcpy(mtx_model[9], mtx_model[8], sizeof(mtx_model[0]));

        memcpy(mtx_model[0], m, sizeof(mtx_model[0]));
}

void tr_ren_push_mul_modelview(float m[4][4])
{
        float   mtx_copy[4][4];

        memcpy(mtx_model[1], mtx_model[0], sizeof(mtx_model[0]));
        memcpy(mtx_model[2], mtx_model[1], sizeof(mtx_model[0]));
        memcpy(mtx_model[3], mtx_model[2], sizeof(mtx_model[0]));
        memcpy(mtx_model[4], mtx_model[3], sizeof(mtx_model[0]));
        memcpy(mtx_model[5], mtx_model[4], sizeof(mtx_model[0]));
        memcpy(mtx_model[6], mtx_model[5], sizeof(mtx_model[0]));
        memcpy(mtx_model[7], mtx_model[6], sizeof(mtx_model[0]));
        memcpy(mtx_model[8], mtx_model[7], sizeof(mtx_model[0]));
        memcpy(mtx_model[9], mtx_model[8], sizeof(mtx_model[0]));

        for(i = 0; i < 16; i++)
        {
                ((float *)mtx_copy)[i] = ((float *)(mtx_model[0]))[i];
        }
        matrix_mult(mtx_model[0], m, mtx_copy);
}

void tr_ren_pop_modelview()
{
        for(i=0; i<9; i++)
        {
                memcpy(mtx_model[i], mtx_model[i+1], sizeof(mtx_model[0]));
        }
}

void tr_ren_viewport()
{
}

void tr_ren_lookat_x()
{
}

void tr_ren_lookat_y()
{
}

void tr_ren_light(int n)
{
}

void tr_ren_numlights()
{
}

void tr_ren_lightcol(int n)
{
}

void tr_ren_line3d(int vn[])
{
        int     i;
        float   m[4][4];
        float   p[4];
        float   v[4];
        float   x[3], y[3];
        //float   z[3], w[3];

        matrix_mult(m, mtx_model[0], mtx_proj);

        for(i=0; i<2; i++)
        {
                v[0] = rdp_reg.vtx[vn[i]].x;   /* * perspective_normal; */
                v[1] = rdp_reg.vtx[vn[i]].y;   /* * perspective_normal; */
                v[2] = rdp_reg.vtx[vn[i]].z;   /* * perspective_normal; */
                v[3] = 1.0;   /* * perspective_normal; */

                vertex_mult(p, v, m);

                x[i] = p[0]/p[3] * rdp_reg.vp[0] + rdp_reg.vp[4];
                y[i] = -(p[1]/p[3]) * rdp_reg.vp[1] + rdp_reg.vp[5];
                //z[i] = p[2]/p[3] * rdp_reg.vp[2] + rdp_reg.vp[6];
                //w[i] = p[3];

                //printf("nx=%g ny=%g nz=%g (w=%g)\n", x[i], y[i], z[i], w[i]);
        }


#ifdef DISPLAY
        if(prefs.display != NO_DISPLAY)
        {
                SetVisualColor(0x07e007e0);
                DrawVisualLine(x[0], y[0], x[1], y[1]);
        }
#endif

}

void tr_ren_tri1(int vn[])
{
        float m[4][4];
        float v[4];
        float p[4];
        float x[3], y[3];
        //float z[3], w[3];

        matrix_mult(m, mtx_model[0], mtx_proj);
        for(i=0; i<3; i++)
        {
                v[0] = rdp_reg.vtx[vn[i]].x;   /* * perspective_normal; */
                v[1] = rdp_reg.vtx[vn[i]].y;   /* * perspective_normal; */
                v[2] = rdp_reg.vtx[vn[i]].z;   /* * perspective_normal; */
                v[3] = 1.0;   /* * perspective_normal; */

                vertex_mult(p, v, m);

                x[i] = p[0]/p[3] * rdp_reg.vp[0] + rdp_reg.vp[4];
                y[i] = -(p[1]/p[3]) * rdp_reg.vp[1] + rdp_reg.vp[5];
                //z[i] = p[2]/p[3] * rdp_reg.vp[2] + rdp_reg.vp[6];
                //w[i] = p[3];

                //printf("nx=%g ny=%g nz=%g (w=%g)\n", x[i], y[i], z[i], w[i]);
        }

#ifdef DISPLAY
        if(prefs.display != NO_DISPLAY)
        {
                SetVisualColor(0xf800f800);
                DrawVisualTriangle(x[0], y[0], x[1], y[1], x[2], y[2]);
        }
#endif

}

void tr_ren_tri2(int vn[])
{
        float   m[4][4];
        int     i;
        float   v[4];
        float   p[4];
        float   x[6], y[6];
        //float z[3], w[3];

        matrix_mult(m, mtx_model[0], mtx_proj);

        for(i=0; i<6; i++)
        {
                v[0] = rdp_reg.vtx[vn[i]].x;   /* * perspective_normal; */
                v[1] = rdp_reg.vtx[vn[i]].y;   /* * perspective_normal; */
                v[2] = rdp_reg.vtx[vn[i]].z;   /* * perspective_normal; */
                v[3] = 1.0;   /* * perspective_normal; */

                vertex_mult(p, v, m);

                x[i] = p[0]/p[3] * rdp_reg.vp[0] + rdp_reg.vp[4];
                y[i] = -(p[1]/p[3]) * rdp_reg.vp[1] + rdp_reg.vp[5];
                //z[i] = p[2]/p[3] * rdp_reg.vp[2] + rdp_reg.vp[6];
                //w[i] = p[3];
        }

#ifdef DISPLAY
        if(prefs.display != NO_DISPLAY)
        {
                SetVisualColor(0xf800f800);
                DrawVisualTriangle(x[0], y[0], x[1], y[1], x[2], y[2]);
                DrawVisualTriangle(x[3], y[3], x[4], y[4], x[5], y[5]);
        }
#endif

}

void tr_ren_geometry_zbuffer(int on)
{
}

void tr_ren_geometry_cullfront(int on)
{
}

void tr_ren_geometry_cullback(int on)
{
}

void tr_ren_geometry_cullfrontback(int on)
{
}






















/* local (static) routines */

static void vertex_mult(float r[4], float v[4], float m[4][4])
{
    /* see below for faster solution! */
    /*
        int i;

        for(i = 0; i < 4; i++)
        {
                r[i] = v[0]*m[0][i]
                     + v[1]*m[1][i]
                     + v[2]*m[2][i]
                     + v[3]*m[3][i];
        }
    */

    /* the following is the fastest way in ansi c in my opinion */

        *((float *)r)++ = v[0]*m[0][0] + v[1]*m[1][0] + v[2]*m[2][0] + v[3]*m[3][0];
        *((float *)r)++ = v[0]*m[0][1] + v[1]*m[1][1] + v[2]*m[2][1] + v[3]*m[3][1];
        *((float *)r)++ = v[0]*m[0][2] + v[1]*m[1][2] + v[2]*m[2][2] + v[3]*m[3][2];
        *((float *)r)++ = v[0]*m[0][3] + v[1]*m[1][3] + v[2]*m[2][3] + v[3]*m[3][3];

} /* static void vertex_mult(float r[4], float v[4], float m[4][4]) */





/* This calculates the product of 2 4x4 float matrices. */

static void matrix_mult(float r[4][4], float m1[4][4], float m2[4][4])
{
    /* see below for faster solution! */
    /*
        int     i, j;

        for(i = 0; i < 4; i++)
        {
                for(j = 0; j < 4; j++)
                {
                        r[i][j] = m1[i][0]*m2[0][j]
                                + m1[i][1]*m2[1][j]
                                + m1[i][2]*m2[2][j]
                                + m1[i][3]*m2[3][j];
                }
        }
    */

    /* the following is the fastest way in ansi c in my opinion */

        *((float *)r)++ = m1[0][0]*m2[0][0] + m1[0][1]*m2[1][0] + m1[0][2]*m2[2][0] + m1[0][3]*m2[3][0];
        *((float *)r)++ = m1[0][0]*m2[0][1] + m1[0][1]*m2[1][1] + m1[0][2]*m2[2][1] + m1[0][3]*m2[3][1];
        *((float *)r)++ = m1[0][0]*m2[0][2] + m1[0][1]*m2[1][2] + m1[0][2]*m2[2][2] + m1[0][3]*m2[3][2];
        *((float *)r)++ = m1[0][0]*m2[0][3] + m1[0][1]*m2[1][3] + m1[0][2]*m2[2][3] + m1[0][3]*m2[3][3];

        *((float *)r)++ = m1[1][0]*m2[0][0] + m1[1][1]*m2[1][0] + m1[1][2]*m2[2][0] + m1[1][3]*m2[3][0];
        *((float *)r)++ = m1[1][0]*m2[0][1] + m1[1][1]*m2[1][1] + m1[1][2]*m2[2][1] + m1[1][3]*m2[3][1];
        *((float *)r)++ = m1[1][0]*m2[0][2] + m1[1][1]*m2[1][2] + m1[1][2]*m2[2][2] + m1[1][3]*m2[3][2];
        *((float *)r)++ = m1[1][0]*m2[0][3] + m1[1][1]*m2[1][3] + m1[1][2]*m2[2][3] + m1[1][3]*m2[3][3];

        *((float *)r)++ = m1[2][0]*m2[0][0] + m1[2][1]*m2[1][0] + m1[2][2]*m2[2][0] + m1[2][3]*m2[3][0];
        *((float *)r)++ = m1[2][0]*m2[0][1] + m1[2][1]*m2[1][1] + m1[2][2]*m2[2][1] + m1[2][3]*m2[3][1];
        *((float *)r)++ = m1[2][0]*m2[0][2] + m1[2][1]*m2[1][2] + m1[2][2]*m2[2][2] + m1[2][3]*m2[3][2];
        *((float *)r)++ = m1[2][0]*m2[0][3] + m1[2][1]*m2[1][3] + m1[2][2]*m2[2][3] + m1[2][3]*m2[3][3];

        *((float *)r)++ = m1[3][0]*m2[0][0] + m1[3][1]*m2[1][0] + m1[3][2]*m2[2][0] + m1[3][3]*m2[3][0];
        *((float *)r)++ = m1[3][0]*m2[0][1] + m1[3][1]*m2[1][1] + m1[3][2]*m2[2][1] + m1[3][3]*m2[3][1];
        *((float *)r)++ = m1[3][0]*m2[0][2] + m1[3][1]*m2[1][2] + m1[3][2]*m2[2][2] + m1[3][3]*m2[3][2];
        *((float *)r)++ = m1[3][0]*m2[0][3] + m1[3][1]*m2[1][3] + m1[3][2]*m2[2][3] + m1[3][3]*m2[3][3];

} /* static void matrix_mult(float r[4][4], float m1[4][4], float m2[4][4]) */





