/**
 * TrueReality - play.c
 * Copyright (C) 1998, 1999 Niki W. Waibel
 *
 * This program is free software; you can redistribute it and/
 * or modify it under the terms of the GNU General Public Li-
 * cence as published by the Free Software Foundation; either
 * version 2 of the Licence, or any later version.
 *
 * This program is distributed in the hope that it will be use-
 * ful, but WITHOUT ANY WARRANTY; without even the implied war-
 * ranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public Licence for more details.
 *
 * You should have received a copy of the GNU General Public
 * Licence along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 * USA.
 *
 * Information about me (the author):
 *   Niki W. Waibel, Reichenau 20, 6890 Lustenau, Austria - EUROPE
 *   niki.waibel@gmx.net
**/





#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <linux/soundcard.h>





#define PLAY_BUFFER     1024

#define INIT_FREQ       44100

#define DOWN_SAMPLE     0

/* this is the time in seconds for what the sound is replayed if
   the end is reached */
#define LOOP_TIME       10





static int      halt = 0;

static int      audiofd;
static FILE     *filefd;
static int      BUFFER_SIZE;
static int      buffer_location;





static void     ctrl_c__pressed();





int OpenAudio(char *soundfile)
{
        int             status;
        audio_buf_info  abi;





        /* open audio device */

        fputs("Opening Linux audio device /dev/dsp ...", stdout); fflush(stdout);
        
        audiofd = open("/dev/dsp", O_WRONLY | O_SYNC);
	filefd = fopen(soundfile, "rb");
        
        if((audiofd < 0) || (filefd == NULL))
	{
		puts(" FAIL");
		return(-1);
	}



        /* reset */

        fputs(" OK\n"
              "Reset ...", stdout); fflush(stdout);

	if(ioctl(audiofd, SNDCTL_DSP_RESET) == -1)
	{
		puts(" FAIL");
                return(-1);
	}



        /* set non blocking mode */

        //ioctl(audiofd, SNDCTL_DSP_NONBLOCK);



        /* set loop if cpu is too slow (seems not to work) */

        //ioctl(audiofd, SNDCTL_DSP_PROFILE, APF_CPUINTENS);



        /* determine output buffer size */

        fputs(" OK\n"
              "Get audio output buffer size ...", stdout); fflush(stdout);

        status = ioctl(audiofd, SNDCTL_DSP_GETOSPACE, &abi);
        if(status == -1)
        {
                puts(" FAIL");
                return(-1);
        }
        BUFFER_SIZE = abi.bytes;
        buffer_location = 0;



        /* determine possible audio formats */

        fputs(" OK\n"
              "Get audio formats ...", stdout); fflush(stdout);

        if(ioctl(audiofd, SNDCTL_DSP_GETFMTS, &status) == -1)
        {
                puts(" FAIL");
                return(-1);
        }

        puts(" OK");

        if(status & AFMT_MU_LAW)
		puts("   MU_LAW");
        if(status & AFMT_A_LAW)
		puts("   A_LAW");
        if(status & AFMT_IMA_ADPCM)
		puts("   IMA_ADPCM");
        if(status & AFMT_U8)
		puts("   U8");
        if(status & AFMT_S16_LE)
		puts("   S16_LE");
        if(status & AFMT_S16_BE)
		puts("   S16_BE");
        if(status & AFMT_S8)
		puts("   S8");
        if(status & AFMT_U16_LE)
		puts("   U16_LE");
        if(status & AFMT_U16_BE)
		puts("   U16_BE");
        if(status & AFMT_MPEG)
		puts("   MPEG");

        if(status & AFMT_S16_BE)
	{
                fputs("   ... S16_BE is supported\n"
                      "Trying to set audio format to S16_BE ...", stdout); fflush(stdout);

                status = AFMT_S16_BE;
		if(ioctl(audiofd, SNDCTL_DSP_SETFMT, &status) == -1)
		{
			puts(" FAIL");
                        return(-1);
		}
	}
        else if(status & AFMT_S16_LE)
	{
                fputs("   ... S16_LE is supported\n"
                      "Trying to set audio format to S16_LE ...", stdout); fflush(stdout);
	
                status = AFMT_S16_LE;
		if(ioctl(audiofd, SNDCTL_DSP_SETFMT, &status) == -1)
		{
			puts(" FAIL");
                        return(-1);
		}
	}
        else
        {
                puts("   ... S16_BE is not supported ... will disable audio"); fflush(stdout);
                return(-1);
        }



        /* determine possible audio capabilities */

        fputs(" OK\n"
              "Get audio capabilities ...", stdout); fflush(stdout);

        if(ioctl(audiofd, SNDCTL_DSP_GETCAPS, &status) == -1)
        {
                puts(" FAIL");
                return(-1);
        }

        puts(" OK");

        printf("   Revision: $%02x (%03d)\n", status & 0xff, status & 0xff);
        if(status & DSP_CAP_DUPLEX)
		puts("   DUPLEX");
        if(status & DSP_CAP_REALTIME)
		puts("   REALTIME");
        if(status & DSP_CAP_BATCH)
		puts("   BATCH");
        if(status & DSP_CAP_COPROC)
		puts("   COPROC");
        if(status & DSP_CAP_TRIGGER)
		puts("   TRIGGER");
        if(status & DSP_CAP_MMAP)
		puts("   MMAP");



        /* set channels to 2 (stereo)  */

        fputs("Set audio channels to 2 (stereo)  ...", stdout); fflush(stdout);

	status = 2;
	if(ioctl(audiofd, SNDCTL_DSP_CHANNELS, &status) == -1)
	{
		puts(" FAIL");
	}



        /* set frequency to INIT_FREQ >> DOWN_SAMPLE */

        printf(" OK\n"
               "Set speed (frequency) to %d ...", INIT_FREQ >> DOWN_SAMPLE); fflush(stdout);

	status = INIT_FREQ >> DOWN_SAMPLE;
	if(ioctl(audiofd, SNDCTL_DSP_SPEED, &status) == -1)
	{
		puts(" FAIL");
	}



        puts(" OK\n"
             "-------------------------------------------------------------------------------"); fflush(stdout);

        return(0);

} /* int OpenAudio() */





int CloseAudio()
{
        fputs("Closing Linux audio device /dev/dsp ...", stdout); fflush(stdout);

	if(ioctl(audiofd, SNDCTL_DSP_SYNC) == -1)
	{
		puts(" FAIL");
                return(-1);
	}

	close(audiofd);
	fclose(filefd);

	puts(" OK\n"  
             "-------------------------------------------------------------------------------"); fflush(stdout);

        return(0);

} /* int CloseAudio() */





int main(int argc, char *argv[])
{
        char    c[PLAY_BUFFER];
        ssize_t size;
        int     i = 0;



        if(argc != 2)
        {
                fprintf(stderr, "USAGE: %s file\n 'file' is the soundfile to play\n\n", argv[0]);
                exit(1);
        }


        signal(SIGINT, ctrl_c__pressed);



        OpenAudio(argv[1]);



        printf("The file read buffer is: %d bytes\nPlaying ...", PLAY_BUFFER);
        fflush(stdout);



        for(;;)
        {
                do
                {
                        size = fread(&c, 1, PLAY_BUFFER, filefd);
                        write(audiofd, &c, size);
                        i+=size;
                }
                while(size != 0 && halt != 1);

                if(halt == 1)
                        break;

                printf("\n   will go %d sec back - we reached the EOF", LOOP_TIME); fflush(stdout);

                if(fseek(filefd, -1 * (LOOP_TIME * ((INIT_FREQ >> DOWN_SAMPLE) << 2)), SEEK_CUR) != 0)
                        rewind(filefd);
                /* The "<< 2" is because we use 16bit samples
                   and a stereo signal instead of 8bit mono.
                   That means that we have to multiply the
                   frequency with 4 to get one second. */

        } /* for(;;) */



	printf(" done\nI played %d bytes\n", i); fflush(stdout);

        puts("-------------------------------------------------------------------------------"); fflush(stdout);



        CloseAudio();



        return(0);

} /* int main() */










static void ctrl_c__pressed()
{
        signal(SIGINT, SIG_IGN);

        halt = 1;

        signal(SIGINT, ctrl_c__pressed);

} /* static void ctrl_c__pressed() */
