#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <linux/soundcard.h>



void yes_no(int cond)
{
        cond ? fputs("   yes     ", stdout) : fputs("   no      ", stdout);
}

int main(int argc, char *argv[])
{
        int fd, i, level;
        char *device;
        int s;
        int recsrc, devmask, recmask, stereodevs, caps;
        const char *sound_dev_names[] = SOUND_DEVICE_LABELS;


        if(argc == 2)
                device = argv[2];
        else
                device = "/dev/mixer";

        fd = open(device, O_RDONLY);
        if(fd == -1)
        {
                perror("");
                return(1);
        }

        s = ioctl(fd, SOUND_MIXER_READ_RECSRC, &recsrc);
        s = ioctl(fd, SOUND_MIXER_READ_DEVMASK, &devmask);
        s = ioctl(fd, SOUND_MIXER_READ_RECMASK, &recmask);
        s = ioctl(fd, SOUND_MIXER_READ_STEREODEVS, &stereodevs);
        s = ioctl(fd, SOUND_MIXER_READ_CAPS, &caps);

        printf("Status of %s:\n\n"
               "Mixer      Device     Recording  Active     Stereo     Current\n"
               "Channel    Available  Source     Source     Device     Level\n"
               "---------  ---------  ---------  ---------  ---------  ---------\n",
               device
        );

        for(i=0; i<SOUND_MIXER_NRDEVICES; i++)
        {
                printf("%2d %-8s", i, sound_dev_names[i]);
                yes_no((1 << i) & devmask);
                yes_no((1 << i) & recmask);
                yes_no((1 << i) & recsrc);
                yes_no((1 << i) & stereodevs);
                
                if((1 << i) & devmask)
                {
                        s = ioctl(fd, MIXER_READ(i), &level);
                        
                        if((1 << i) & stereodevs)
                                printf("  %d%% %d%%", level & 0xff, (level & 0xff00) >> 8);
                        else
                                printf("  %d%%", level & 0xff);
                }
                puts("");
        }

        fputs("\nNote: Choices for recording source are ", stdout);
        if( !(caps & SOUND_CAP_EXCL_INPUT) )
                fputs("not ", stdout);
        puts("exclusive.");

        close(fd);

        return(0);
}
